import java.io.Serializable;


public class ABIRead implements Comparable, Serializable {

    protected String barcode, code, sequence;
    protected int pos, length, chr, nMismatches, nMatches = 1;
    protected boolean strand;
    protected double qualityScore;

    //////////////////////////////////////////////////////////////////

    public ABIRead(int c, boolean s, int p, int l, double q, int m){
	chr = c;
	strand = s;
	pos = p>0 ? p : 1;
	length = l;
	qualityScore = q;
	nMismatches = m;
    }

    public ABIRead(int c, boolean s, int p, int l){
	chr = c;
	strand = s;
	pos = p>0 ? p : 1;
	length = l;
    }

    public ABIRead(int c, boolean s, int p){
	chr = c;
	strand = s;
	pos = p;
	length = FindNongenicTranscripts.MRNA_READ_LENGTH;
    }

    public ABIRead(int m, int c, boolean s, int p, int l){
	nMatches = m;
	chr = c;
	strand = s;
	pos = p;
	length = l;
    }

    public ABIRead(ABIRead r){
	chr = r.chr;
	strand = r.strand;
	pos = r.pos;
	length = r.length;
	qualityScore = r.qualityScore;
	nMismatches = r.nMismatches;
    }

    ///////////////////////////////////////////////////////

    public int getPosition(){ return pos; }

    public void setPosition(int p) { pos = p; }

    public boolean getStrand(){ return strand; }
    
    public int getLength(){ return length; }

    public void setCode(String c) { code = c; }

    public String getCode(){ return code; }

    public int getChromosome() { return chr; }

    public void setSequence(String s) { sequence = s; }

    public String getSequence(){ return sequence; }

    public int getGCContent() { return AnalyzeGCBias.gcContent(sequence); }

    public double getQualityScore() { return qualityScore; }

    public int numberOfMatches() { return nMatches; }

    public int numberOfMismatches() { return nMismatches; }

    public int getCenter(boolean frag) {
	//Find the center of the read or the extended fragment
	if (frag)
	    return getPosition() + length/2;
	else if (strand)
	    return getPosition() + AnalyzeControls.READ_LENGTH/2;
	else
	    return getPosition() + length - AnalyzeControls.READ_LENGTH/2;
    }

    public boolean overlaps(int start, int end) { return pos + length<=start && pos + length>=end; }

    public boolean overlaps(int s) { return pos<=s && pos + length>=s; }

    public boolean overlaps(ABIRead r) { 
	if (r.chr==chr)
	    return (r.pos + r.length>=pos && r.pos<=pos) || (pos + length>=r.pos && pos<=r.pos);
	return false;
    }

    //////////////////////////////////////////////////////////////////

    public String toString() { return chr + "_" + (strand ? "" : "-") + pos; }

    public boolean equals(Object o) {
	if (!(o instanceof ABIRead))
	    System.err.println("Error comparing reads!");
	ABIRead a = (ABIRead)o;
	return this.pos==a.pos && this.strand==a.strand && this.length==length;
    }

    public int compareTo(Object o) {
	//Sort according to position
	if (!(o instanceof ABIRead))
	    System.err.println("Error comparing reads!");
	ABIRead a = (ABIRead)o;
	if (this.pos<a.pos)
	    return -1;
	else if (this.pos>a.pos)
	    return 1;
	else {
	    if (this.strand && !a.strand)
		return -1;
	    else if (!this.strand && a.strand)
		return 1;
	    else { //Let the ones with few mismatches come first
		if (this.nMismatches<a.nMismatches)
		    return -1;
		else if (this.nMismatches>a.nMismatches)
		    return 1;
		else
		    return 0;
	    }
	}
    }
}
