//Use this method to calculate correlations across the genome. The -f
//flag allows us to restrict ourselves to certain features, at the
//moment, these inclde:
//
//- within a distance of TSS
//
//- coding regions


//Example:

//java -Xmx512m CalculateCorrelations -n Npas4/KCl/Npas4_KCl -w 1000 -f Genome

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class CalculateCorrelations{

    public static String GENOME_LOC;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    //Keep track of all the peaks in this datastructure
    protected static TreeSet[] peaks = new TreeSet[21];


    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option featureFlag = argParser.addStringOption('f', "feature");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //use this flag when we only want to focus on features
	    //which are guaranteed to be isolated from one another.
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('i', "isolated");
	    //The value for this flag should correspond to the value
	    //used for CalculateCriticalWindowDifference.java
	    CmdLineParser.Option windowStepFlag = argParser.addIntegerOption('W', "windowStep");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option mappabilityBinsFlag = argParser.addIntegerOption('m', "mappabilityBins");
	    CmdLineParser.Option gcBinsFlag = argParser.addIntegerOption('g', "GCBins");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    tmp = argParser.getOptionValue(featureFlag);
	    String feature = tmp!=null ? "TSS" : tmp.toString();
	    tmp = argParser.getOptionValue(upstreamFlag);
	    int upstream = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(downstreamFlag);
	    int downstream = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(windowStepFlag);
	    int windowStep = tmp!=null ? ((Integer)tmp).intValue() : 10;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 240;
	    tmp = argParser.getOptionValue(gcBinsFlag);
	    int nGCBins = tmp!=null ? ((Integer)tmp).intValue() : 6;
	    tmp = argParser.getOptionValue(mappabilityBinsFlag);
	    int nMappabilityBins = tmp!=null ? ((Integer)tmp).intValue() : 6;
	    //////////////////////////////////////////////////////////////////
	    if (chr==-1) {
		double hists[][] = new double[21][numberOfBins(downstream, upstream, windowStep)], sumHist[] = new double[numberOfBins(downstream, upstream, windowStep)];
		for (int i=1; i<=21; i++) 
		    hists[i-1] = analyzeChromosome(i, fileName, upstream, downstream, windowSize, windowStep, nMappabilityBins, nGCBins, feature, isolated, orchestra);
		//Add them all up and save to file
		for (int i=1; i<21; i++) {
		    for (int j=0; j<hists[i - 1].length; j++) 
			sumHist[j] += hists[i - 1][j];
		}
		String saveFileName = fileName + "_combined_" + feature + ".txt";
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		for (int i=0; i<sumHist.length; i++)
		    bw.write((-upstream + i*windowStep) + "\t" + sumHist[i]);
		bw.close();
	    }
	    else
		analyzeChromosome(chr, fileName, upstream, downstream, windowSize, windowStep, nMappabilityBins, nGCBins, feature, isolated, orchestra);
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static int numberOfBins(int up, int down, int step) { return (up + down)/step; }

    public static double[] analyzeChromosome(int chr, String fileName, int upstream, int downstream, int windowSize, int windowStep, int nMappabilityBins, int nGCBins, String feature, boolean isolated, boolean orchestra) {
	double hist[] = new double[numberOfBins(downstream, upstream, windowStep)];
	try {
	    if (feature.toLowerCase().equals("tss")) {
		System.out.println("Loading data for " + chr);
		RefSeqGenes geneInfo = new RefSeqGenes(chr, orchestra);
		Iterator it = geneInfo.getGenes().iterator();
		WindowFinder windowFinder = new WindowFinder(fileName, chr, true, windowSize, windowStep, nMappabilityBins, nGCBins);
		while (it.hasNext()) {
		    RefSeqGene g = (RefSeqGene)it.next();
		    Vector tsss = g.getStarts();
		    Iterator tssIt = tsss.iterator();
		    while (tssIt.hasNext()) {
			int tss = ((Integer)tssIt.next()).intValue();
			int ind = windowStep*((tss - upstream)/windowStep), end = windowStep*((tss + downstream)/windowStep), i = g.getStrand() ? 0 : hist.length - 1;
			while (ind<end) {
			    Window win = windowFinder.findReadsInRegion(ind, ind + windowSize);
			    if (win!=null) 
				hist[i] += (double)win.numberOfReads()/geneInfo.size();
			    ind += windowStep;
			    i += g.getStrand() ? 1 : -1;
			}
			//System.out.println("Found gene " + g.toString() + ": " + Arrays.toString(hist));			
		    }
		}
		String saveFileName = fileName + "_combined_" + feature + "_" + chr + ".txt";
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		for (int i=0; i<hist.length; i++)
		    bw.write((-upstream + i*windowStep) + "\t" + hist[i]);
		bw.close();
	    }
	}
	catch (IOException e) {
	    System.err.println("Error extracting features!");
	}
	return hist;
    }
    
    
}
