//Use this method to calculate the mappability at the enhancer
//regions. In particular, we want to compare the core enhancer region
//with the flanking regions as well as the random regions.
//
//Example:
//
//bsub -q shared_2h java CalculateEnhancerRegionMappability -o -c 19 
//

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.list.ListOfTallies;
import umontreal.iro.lecuyer.stat.TallyStore;

public class CalculateEnhancerRegionMappability {

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");	    
	    CmdLineParser.Option incFlag = argParser.addIntegerOption("inc");	    
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option enhancerFileNameFlag = argParser.addStringOption("enhancers");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(enhancerFileNameFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmDataEnhancers") + "/enhancers_b2_null_all.txt";
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(incFlag);
	    int nIncrements = tmp!=null ? ((Integer)tmp).intValue() : 4;
	    ///////////////////////////////////////////////////////
	    int nChr = 21;
	    Vector<Enhancer> enhancers = FindReadsAcrossFeature.parseEnhancerFile(enhancerFileName);
	    Collections.sort(enhancers);
	    ListOfTallies<TallyStore> mappabilityPotential = ListOfTallies.createWithTallyStore(nIncrements) ;
	    int nEnhancers = 0;
	    Vector<TreeMap<Enhancer, Integer>> enhancerMappability = new Vector<TreeMap<Enhancer, Integer>>();
	    for (int j=0; j<nIncrements; j++) 
		enhancerMappability.addElement(new TreeMap<Enhancer, Integer>());
	    //Save the sequences as well
	    BufferedWriter bw = nIncrements==0 ? new BufferedWriter(new FileWriter(enhancerFileName.substring(0, enhancerFileName.lastIndexOf(".")) + "_intergenic_w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".fa")) : null;
	    BufferedWriter bwCore = nIncrements==0 ? new BufferedWriter(new FileWriter(enhancerFileName.substring(0, enhancerFileName.lastIndexOf(".")) + "_core_intergenic_w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".fa")) : null;
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 ||i==chr) {
		    RefSeqGenes geneInfo = new RefSeqGenes(i, orchestra);
		    FastaSequence2 fastaReader = new FastaSequence2(i, orchestra);
		    if (nIncrements>0) {
			for (int j=0; j<nIncrements; j++) {
			    Mappability mappability = new Mappability(orchestra, i, windowSize*(j+1));
			    //Extract the mappability in the region surrounding the enhancers
			    int prevLoc = 0;			
			    for (Enhancer en : enhancers) {
				//System.out.println(en.toString());
				if (en.getChromosome()==i && !geneInfo.overlapsGene(en.getCenter())) {
				    int moveWindowDist = en.getCenter() - prevLoc - windowSize*(j+1)/2;
				    enhancerMappability.elementAt(j).put(en, mappability.numberOfPotentialReadsNextWindow(moveWindowDist));
				    prevLoc = en.getCenter() - windowSize*(j+1)/2;
				}
			    }
			}
		    }
		    else {
			System.out.println("Saving to " + enhancerFileName.substring(0, enhancerFileName.lastIndexOf(".")) + "_intergenic_w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".fa");
			for (Enhancer en : enhancers) {
			    if (en.getChromosome()==i && !geneInfo.overlapsGene(en.getCenter())) {
				bw.write(">chr" + i + ":" + (en.getCenter() - windowSize) + "-" + en.getCenter() + "(-)\n" + FastaSequence2.reverse(FastaSequence2.complement(fastaReader.getSequence(en.getCenter() - windowSize, windowSize))) + "\n>chr" + i + ":" + en.getCenter() + "-" + (en.getCenter() + windowSize) + "(+)\n" + fastaReader.getSequence(en.getCenter(), windowSize) + "\n");
				bwCore.write(">chr" + i + ":" + (en.getCenter() - windowSize) + "-" + (en.getCenter() + windowSize) + "\n" + fastaReader.getSequence(en.getCenter() - windowSize, windowSize*2 + 1) + "\n");
			    }
			}
		    }
		}
	    }
	    if (bw!=null) {
		bw.close();
		bwCore.close();
	    }
	    //Compare the mappability near the enhancer with
	    //the flanking regions
	    System.out.println("Found " + enhancerMappability.elementAt(0).size() + " enhancers.");
	    for (Enhancer en : enhancerMappability.elementAt(0).keySet()) {
		double d[] = new double[nIncrements];
		d[0] = (double)enhancerMappability.elementAt(0).get(en)/(windowSize*2 - 240);
		for (int j=1; j<nIncrements; j++) {
		    d[j] = (double)(enhancerMappability.elementAt(j).get(en) - enhancerMappability.elementAt(j-1).get(en))/(windowSize*2 - 240);
		}
			mappabilityPotential.add(d);
	    }
	    /*System.out.println("Correlation for mappabilities near enhancers are:");
	    for (int j=0; j<nIncrements; j++) {
		for (int k=j+1; k<nIncrements; k++) 
		    System.out.println(j + ", " + k + ": " + mappabilityPotential.correlation(j, k));
		    }*/
	    double r[] = new double[nIncrements], std[] = new double[nIncrements];
	    mappabilityPotential.average(r);
	    mappabilityPotential.standardDeviation(std);
	    System.out.println("Average mappabilities near enhancers are: " + Arrays.toString(r));
	    System.out.println("Std dev mappabilities near enhancers are: " + Arrays.toString(std));
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

}
