//Convert the genome into pseudo-color space. This is more relevant
//for analyzing the SOLID data since it is degenerate and has
//different properties than the original genome

import java.io.*;
import jargs.gnu.CmdLineParser;

public class ConvertGenome2PseudoColor {


    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    //Use this flag to for the location of the .fa-files
	    CmdLineParser.Option saveDirectoryFlag = argParser.addStringOption('d', "saveDirectory");
	    CmdLineParser.Option loadDirectoryFlag = argParser.addStringOption('l', "loadDirectory");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag to decide which species to
	    //investigate. Available species are: 
	    //Mouse (Mm) Default
	    //Yeast (Sc)
	    CmdLineParser.Option speciesFlag = argParser.addStringOption('s', "species");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    String GENOME_LOC = "./";
	    int nChr = 0;
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? tmp.toString() : "Mm";
	    if (species.equals("Mm")) {
		GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
		nChr = 21;
	    }
	    else if (species.equals("Sc")) {
		GENOME_LOC = orchestra ? "/home/mh168/Genomes/Sc/sacCer1/" : "/media/exthd/Genomes/Sc/sacCer1/";
		nChr = 17;
	    }
	    tmp = argParser.getOptionValue(saveDirectoryFlag);
	    String saveDir = tmp!=null ? tmp.toString() : GENOME_LOC + "PseudoColorCode/";
	    tmp = argParser.getOptionValue(loadDirectoryFlag);
	    String loadDir = tmp!=null ? tmp.toString() : GENOME_LOC;
	    //make sure that it ends with '/'
	    if (!saveDir.endsWith("/")) { saveDir += "/"; }
	    if (!loadDir.endsWith("/")) { loadDir += "/"; }
	    //////////////////////////////////////////////////////////////////
	    for (int i=1; i<=nChr; i++) {
		BufferedReader br = new BufferedReader(new FileReader(loadDir + AnalyzeControls.getChromosomeString(i, species)));
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveDir + AnalyzeControls.getChromosomeString(i, species)));
		String line = br.readLine();
		while (line!=null) {
		    if (line.startsWith(">")) //comment
			bw.write(line + "\n");
		    else
			bw.write(FastaSequence2.getPseudoColorCode(line) + "\n");
		    line = br.readLine();
		}
		br.close();
		bw.close();
	    }
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

}
