import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class CreateBedFiles{

    public static final int READS_NORMALIZATION_FACTOR = 1000000;

    public static String GENOME_LOC;

    protected static int nMismatches, fragmentLength, windowSize, downSample;
    protected static double normalizationFactor, qualityScore;
    protected static boolean verbose, unique, orchestra, extend, noHeader;
    protected static String fileName;
    
    public static DecimalFormat format = new DecimalFormat("###.######");

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('m', "mismatches");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    //Use this flag for determining the windowsize when
	    //scanning the genome for various properties
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option singletonFlag = argParser.addBooleanOption('s', "singletons");
	    CmdLineParser.Option uniqueFlag = argParser.addBooleanOption('u', "unique");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag if we want to extend the reads to length fragmentLength
	    CmdLineParser.Option extendFlag = argParser.addBooleanOption('e', "extend");
	    //Avoid writing the header, good for bigBedFiles
	    CmdLineParser.Option noHeaderFlag = argParser.addBooleanOption("noHeader");
	    //Use this flag to normalize the count by the total number
	    //of reads and then mulitply by the number given as
	    //argument here.
	    CmdLineParser.Option normalizeFlag = argParser.addDoubleOption('N', "normalize");
	    //Use this flag when we want to downsample the number of
	    //tags. Use ful when we want to avoid generating a new set
	    //of randomized tags.
	    CmdLineParser.Option downSampleFlag = argParser.addIntegerOption('D', "downSampled");
	    //Minimum quality score required for a read to be included
	    CmdLineParser.Option qualityScoreFlag = argParser.addDoubleOption('q', "qualityScore");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    nMismatches = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    windowSize = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(normalizeFlag);
	    normalizationFactor = tmp!=null ? ((Double)tmp).doubleValue() : READS_NORMALIZATION_FACTOR;
	    tmp = argParser.getOptionValue(downSampleFlag);
	    downSample = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    verbose = argParser.getOptionValue(verboseFlag)!=null;
	    unique = argParser.getOptionValue(uniqueFlag)!=null;
	    orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    extend = argParser.getOptionValue(extendFlag)!=null;
	    noHeader = argParser.getOptionValue(noHeaderFlag)!=null;
	    tmp = argParser.getOptionValue(qualityScoreFlag);
	    qualityScore = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    //////////////////////////////////////////////////////////////////
	    GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	    
	    if (chr==-1) { //Look at all chromosomes
		analyze();
		//for (int i=1; i<=21; i++)
		//  analyze(i);
	    }
	    else
		analyze(chr);
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    public static void analyze(int c) {
	ABIRead reads[] = PeakFinder2.parseFiles(c, fileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, AnalyzeControls.getSaveFileName(fileName, "read_stats", c, fragmentLength, "txt"));
	PeakFinder2.savePositionsReads(reads, fileName + "_" + (extend ? "E" : "") + fragmentLength + "_" + c + ".bed", c, !noHeader, normalizationFactor*PeakFinder2.CHROMOSOME_LENGTH[c-1]/(reads.length*PeakFinder2.TOTAL_CHROMOSOME_LENGTH));
	//TKParser tkParser = new TKParser(fileName, c, nMismatches, unique, fragmentLength);
    }
    
    public static void analyze() {
	//Do all chromosomes in one big file
	for (int i=1; i<=21; i++) {
	    ABIRead reads[] = PeakFinder2.parseFiles(i, fileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, AnalyzeControls.getSaveFileName(fileName, "read_stats", i, fragmentLength, "txt"));
	    //PeakFinder2.savePositionsReads(reads, fileName + "_" + (extend ? "E" : "") + fragmentLength + ".bed", i, i==1 && !noHeader, normalizationFactor*PeakFinder2.CHROMOSOME_LENGTH[i-1]/(reads.length*PeakFinder2.TOTAL_CHROMOSOME_LENGTH), i!=1);
	    //TKParser tkParser = new TKParser(fileName, c, nMismatches, unique, fragmentLength);
	}
    }

}
