import java.util.*;

public class EST implements Comparable{

    public static final int MIN_EST_INTRON_LENGTH = 50;

    protected String code = "";
    protected int start, end, chr, tissue = -1;
    protected boolean strand, spliced = false;

    protected Vector<Integer> exonStarts = new Vector<Integer>(), intronStarts = new Vector<Integer>();

    public EST(int c, int s, int e, boolean st){
	chr = c;
	start = s;
	end = e;
	strand = st;
	exonStarts.addElement(s);
    }

    public EST(int c, int s, int e, boolean st, String co){
	chr = c;
	start = s;
	end = e;
	strand = st;
	code = co;
	exonStarts.addElement(s);
    }

    public void merge(EST e){
	//Use this constructor when merging to ESTs
	if (chr==e.chr && strand==e.strand) {
	    if (Math.abs(e.start - end)>MIN_EST_INTRON_LENGTH) {
		intronStarts.addElement(end + (strand ? 1 : -1));
		exonStarts.addElement(e.start);
		//Make sure that it is sorted
		Collections.sort(exonStarts);
		Collections.sort(intronStarts);
		if (!strand) { //reverse the order
		    Vector<Integer> tmp = new Vector<Integer>(exonStarts);
		    for (int i=tmp.size()-1; i>=0; i--)
			exonStarts.setElementAt(tmp.elementAt(i), tmp.size() - i - 1);
		    tmp = new Vector<Integer>(intronStarts);
		    for (int i=tmp.size()-1; i>=0; i--)
			intronStarts.setElementAt(tmp.elementAt(i), tmp.size() - i - 1);
		}
	    }
	    if (strand) {
		start = Math.min(start, e.start);
		end = Math.max(end, e.end);
	    }
	    else {
		start = Math.max(start, e.start);
		end = Math.min(end, e.end);
	    }
	}
	else 
	    System.err.println("Error merging ESTs " + toString() + " and " + e.toString());
    }

    public int getStart(){ return start; }

    public int getEnd(){ return end; }

    public int getLength(){ return Math.abs(end - start); }

    public boolean getStrand(){ return strand; }
    
    public void setCode(String c) { code = c; }

    public String getCode(){ return code; }

    public int getChromosome() { return chr; }

    public void setSpliced(boolean s) { spliced = s; }

    public boolean isSpliced() { return spliced; }

    public void setTissueType(int t) { tissue = t; }
    
    public int getTissueType() { return tissue; }

    /////////////////////////////////

    public int numberOfIntrons() { return intronStarts.size(); }

    public int getExonStart(int ind) { return exonStarts.elementAt(ind); }

    public int getIntronStart(int ind) { return intronStarts.elementAt(ind); }

    public int getExonLength(int ind) { 
	if (ind==0 && exonStarts.size()==1)
	    return getLength();
	else if (ind==0) 
	    return Math.abs(intronStarts.elementAt(0) - exonStarts.elementAt(0));
	else if (ind==exonStarts.size()-1) 
	    return Math.abs(end - exonStarts.lastElement());
	else 
	    return Math.abs(intronStarts.elementAt(ind) - exonStarts.elementAt(ind));
    }

    public int getIntronLength(int ind) { 
	return Math.abs(intronStarts.elementAt(ind-1) - exonStarts.elementAt(ind));
    }

    //////////////////////////////////////////////////////////////////

    public String toString() { return code + " chr" + chr + ":" + (strand ? start : end) + "-" + (strand ? end : start) + (strand ? "" : "(-)"); }

    public boolean equals(Object o) {
	if (!(o instanceof EST))
	    System.err.println("Error comparing reads!");
	EST a = (EST)o;
	return this.start==a.start && this.strand==a.strand && this.end==a.end && this.chr==a.chr;
    }

    public int hashCode() { return chr*93 + 17*start + 29*end; }

    public int compareTo(Object o) {
	//Sort according to position
	if (!(o instanceof EST))
	    System.err.println("Error comparing ESTs!");
	EST a = (EST)o;
	if (this.start<a.start)
	    return -1;
	else if (this.start>a.start)
	    return 1;
	else {
	    if (this.strand && !a.strand)
		return -1;
	    else if (!this.strand && a.strand)
		return 1;
	    else { //Let the ones with few mismatches come first
		if (this.end<a.end)
		    return -1;
		else if (this.end>a.end)
		    return 1;
		else
		    return 0;
	    }
	}
    }
}
