//Second version is more streamlined and meant to be used for the
//transcript calling stuff rather than the enhancer stuff.

import edu.uci.ics.jung.graph.impl.*;
import java.util.*;

public class Enhancer2 extends UndirectedSparseVertex implements Comparable {

    protected int chr, center;

    protected TreeMap<String, ABIRead[]> factorReads = new TreeMap<String, ABIRead[]>();


    //////////////////////////////////////////////////////////////////

    public Enhancer2(int c, int ce) {
	chr = c;
	center = ce;
    }

    ///////////////////////////////////////////////////////

    public int getChromosome() { return chr; }

    public int getCenter() { return center; }

    public boolean overlaps(int st, int en) { return st<center && en>center; }

    public boolean overlaps(int st, int en, int maxDist) { return st-maxDist<center && en+maxDist>center; }

    public void setFactorReads(String tfName, ABIRead[] reads) { factorReads.put(tfName, reads); }

    public ABIRead[] getFactorReads(String tfName) { return factorReads.containsKey(tfName) ? factorReads.get(tfName) : null; }

    //public double getReadDensity(String tfName) { return (double)getFactorReads(tfName).length/getLength(); }



    //////////////////////////////////////////////////////////////////

    public String toString() {
	String str = "chr" + chr + ":" + center;
	return str;
    }

    public int compareTo(Object o) {
	if (!(o instanceof Enhancer2))
	    System.err.println("ERROR comparing enhancers!");
	Enhancer2 e = (Enhancer2)o;
	if (this.chr<e.chr)
	    return -1;
	else if (this.chr>e.chr)
	    return 1;
	else {
	    if (getCenter()<e.getCenter())
		return -1;
	    else if (getCenter()>e.getCenter())
		return 1;
	    else 
		return 0;
	}
    }

    public boolean equals(Object o) {
	if (!(o instanceof Enhancer2))
	    return false;
	Enhancer2 e = (Enhancer2)o;
	return chr==e.chr && e.center==center;
    }

    public int hashCode() { return 93*chr + 29*center; }

}
