//Use this method to enumerate all 35-mers in the genome

import java.util.*;
import java.io.*;

public class EnumerateMers {

    protected HashSet uniqueMers = new HashSet(100000);
    protected HashMap nonUniqueMers = new HashMap(100000);

    public EnumerateMers() { }

    public void findMers(int chr, int len, boolean strand, boolean unique) {
	String fileName = "/media/exthd/Genomes/Mm/v9/" + AnalyzeControls.getChromosomeString(chr);
	try {
	    //Open the genome
	    BufferedReader br = new BufferedReader(new FileReader("/media/exthd/Genomes/Mm/v9/chr18.fa"));
	    //BufferedReader br = new BufferedReader(new FileReader(fileName + ".fa"));
	    //Go through 
	    String line = br.readLine();
	    int bytesRead = line.length();
	    while (line.startsWith(">")) {
		line = br.readLine();
		bytesRead += line.length();
	    }
	    char l[] = line.toCharArray();
	    char window[] = new char[len];
	    for (int i=0; i<window.length; i++) //Initalize
		window[i] = l[i];
	    storeMer(new String(window), unique);
	    window = scanGenome(line.substring(len).toCharArray(), window, unique);
	    while (line!=null) {
		window = scanGenome(line.toCharArray(), window, unique);
		line = br.readLine();
		bytesRead += line.length();
		if (uniqueMers.size()>0 && uniqueMers.size() % 1000==0)
		    System.out.println("Found " + uniqueMers.size() + " unique mers after " + bytesRead);
	    }
	    br.close();
	    System.out.println("Found " + uniqueMers.size() + " unique " + len + "-mers and " + nonUniqueMers.size() + " non-unique");
	}
	catch (IOException e) { 
	    System.err.println("Error reading " + fileName);
	}
	try {
	    //Write the two lists to file
	    BufferedReader br = new BufferedReader(new FileReader("/media/exthd/Genomes/Mm/v9/chr18.fa"));
	    String line = br.readLine();
	    while (line.startsWith(">"))
		line = br.readLine();
	    char l[] = line.toCharArray();
	    char window[] = new char[len];
	    for (int i=0; i<window.length; i++) //Initalize
		window[i] = l[i];
	    String saveFileName = "chr" + AnalyzeControls.getChromosomeString(chr) + "_U" + (strand ? "f" : "r") + "_" + len + ".txt";
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    while (line!=null) {
		saveUniqueMers(line.toCharArray(), window, bw);
		line = br.readLine();
	    }
	    bw.close();
	    br.close();
	    if (!unique) {
		saveFileName = "chr" + AnalyzeControls.getChromosomeString(chr) + "_NU" + (strand ? "f" : "r") + "_" + len + ".txt";
		bw = new BufferedWriter(new FileWriter(saveFileName));
		Iterator it = nonUniqueMers.keySet().iterator();
		while (it.hasNext()) {
		    String mer = (String)it.next();
		    if (mer.indexOf('N')==-1 && mer.indexOf('n')==-1)
			bw.write(mer + "\t" + ((Integer)nonUniqueMers.get(mer)).toString() + "\n");
		}
		bw.close();
	    }
	}
	catch (IOException e) { }

    }

    public char[] saveUniqueMers(char[] l, char[] window, BufferedWriter bw) {
	try {
	    int len = window.length;
	    for (int i=0; i<l.length; i++) {
		char[] newWin = new char[len];
		newWin[len-1] = l[i];
		for (int j=1; j<window.length; j++)
		    newWin[j-1] = window[j];
		bw.write(new String(newWin) + "\n");
		window = newWin;
	    }
	}
	catch (IOException e) { }
	return window;
    }

    public char[] scanGenome(char[] l, char[] window, boolean unique) {
	int len = window.length;
	for (int i=0; i<l.length; i++) {
	    char[] newWin = new char[len];
	    newWin[len-1] = l[i];
	    for (int j=1; j<window.length; j++)
		newWin[j-1] = window[j];
	    storeMer(new String(newWin), unique);
	    window = newWin;
	}
	return window;
    }

    public void storeMer(String str, boolean unique) {
	if (str.indexOf('N')!=-1 || str.indexOf('n')!=-1)
	    return;
	boolean noLongerUnique = false;
	Integer hash = new Integer(str.hashCode());
	if (uniqueMers.contains(hash)) 
	    noLongerUnique = uniqueMers.remove(hash);
	if (nonUniqueMers.containsKey(hash)) 
	    nonUniqueMers.put(hash, new Integer(((Integer)nonUniqueMers.get(hash)).intValue() + 1));
	else if (noLongerUnique) {
	    nonUniqueMers.put(hash, new Integer(2));
	    //System.out.println("Found non-unique " + str);
	}
	else { //completely new unique element
	    uniqueMers.add(hash);
	    //System.out.println("Found unique " + str);
	}

    }

    public HashSet getUniqueMers() { return uniqueMers; }

    public HashMap getNonUniqueMers() { return nonUniqueMers; }

    public static String reverseComplement(String str){
	int l = str.length();
	char c[] = str.toCharArray(), r[] = new char[l];
	for (int i=0; i<c.length; i++) {
	    int j = l - i - 1;
	    if (c[i]=='a')
		r[j] = 't';
	    else if (c[i]=='c')
		r[j] = 'g';
	    else if (c[i]=='g')
		r[j] = 'c';
	    else if (c[i]=='t')
		r[j] = 'a';
	    else if (c[i]=='A')
		r[j] = 'T';
	    else if (c[i]=='C')
		r[j] = 'G';
	    else if (c[i]=='G')
		r[j] = 'C';
	    else if (c[i]=='T')
		r[j] = 'A';
	    else if (c[i]=='n')
		r[j] = 'n';
	    else if (c[i]=='N')
		r[j] = 'N';
	}
	return new String(r);
    }
}
