//Use this mehtod to try to account for all the reads in the
//experiment and see how many fall into each category. Together with
//the matlab sript ismb_poster_plots.m, this is the file mainly used
//for creating table 1 in the paper, the one accounting for the
//distribution of reads and transcripts.
//
//bsub -q shared_2h java -Xmx4096m -server EstimateGenicTranscription -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/merged_total/merged_total -o
//
//bsub -q shared_2h java -Xmx4096m -server EstimateGenicTranscription -n /groups/neuroduo/martin/ChIP-Seq/HeLa/mRNA/HeLac -o --species hs

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class EstimateGenicTranscription {

    public static final int MIN_READS_FOR_ARTEFACT = 10;
    public static final double MIN_READS_FROM_ONE_BASE_PROPORTION_FOR_ARTEFACT = .90;

    public static DecimalFormat format = new DecimalFormat("###.#######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option intronicFlag = argParser.addBooleanOption('I', "intronic");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option reverseFlag = argParser.addBooleanOption('r', "reverse");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option bedFileNameFlag = argParser.addStringOption('b', "bedFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option maxDistanceFlag = argParser.addIntegerOption('d', "maxDistance");	    
	    CmdLineParser.Option maxEnhancerDistFlag = argParser.addIntegerOption('e', "maxEnhancerDist");	    
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");	    
	    CmdLineParser.Option minReadsFlag = argParser.addIntegerOption('m', "minReads");	    
	    CmdLineParser.Option minDensFlag = argParser.addDoubleOption("minDens");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean intronic = argParser.getOptionValue(intronicFlag)!=null;
	    boolean forward = argParser.getOptionValue(reverseFlag)==null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(bedFileNameFlag);
	    String bedFileName = tmp!=null ? (String)tmp : fileName;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : fileName + "_" + bedFileName.substring(bedFileName.lastIndexOf("/") + 1) + "_reads_categories" + "_" + (forward ? "f" : "r") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    tmp = argParser.getOptionValue(maxDistanceFlag);
	    int maxDistance = tmp!=null ? ((Integer)tmp).intValue() : 10000;
	    tmp = argParser.getOptionValue(maxEnhancerDistFlag);
	    int maxEnhancerDist = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(minReadsFlag);
	    int minReads = tmp!=null ? ((Integer)tmp).intValue() : 4;
	    tmp = argParser.getOptionValue(minDensFlag);
	    double minDens = tmp!=null ? ((Double)tmp).doubleValue() : .001;
	    ///////////////////////////////////////////////////////
	    String cellType = "Helas3";
	    String exonFileName = "/groups/neuroduo/martin/Genomes/" + (species.equals("mm") ? "Mm/v9" : "Hs/v36") + "/all_EXNs.tsv";

	    String extragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_lRNA_lMe3_real.tsv";
	    TreeSet<Enhancer2> extragenicEnhancers = FindTranscripts2.parseEnhancerFile(extragenicEnhancerFileName);
	    String intragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_intragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_in_lMe3_real.tsv";
	    TreeSet<Enhancer2> intragenicEnhancers = FindTranscripts2.parseEnhancerFile(intragenicEnhancerFileName);
	    TreeSet<Enhancer2> allEnhancers = new TreeSet<Enhancer2>();
	    allEnhancers.addAll(extragenicEnhancers);
	    allEnhancers.addAll(intragenicEnhancers);
	    String lncRNAFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/landerRegions.bed";
	    String lncRNAExonFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/exonsNPC.gff";
	    TreeSet<LncRNA> lncRNAs = AnalyzeTranscriptReadProfiles.loadLncRNA(lncRNAFileName, species);
	    String macroRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/macroRNA.txt";
	    TreeSet<MacroRNA> macroRNAs = AnalyzeTranscriptReadProfiles.loadMacroRNA(macroRNAFileName, species);
	    String miRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/miRNA.txt";
	    TreeSet<MiRNA> miRNAs = AnalyzeTranscriptReadProfiles.loadMiRNA(miRNAFileName, species);
	    String peakFileName = !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/hela_tf_names.txt" : "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/MmDataPeakFiles/tf_kcl_names";
	    String tfNamesInStudy[] = species.equals("mm") ? new String[]{"CBP_Millipore_KCl_B1","CREB_SC_KCl_B1","Npas4_Npas4_KCl_B1","SRF_H300_KCl_B1"} : EncodePeaks.getEncodeTFNames(peakFileName, cellType);
	    Vector<Peak> peaks = !species.equals("mm") ? new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(peakFileName, cellType, chr)) : new Vector<Peak>(FindEnhancerSites.parsePeakFile(peakFileName, chr).values());
	    System.out.println("Found " + peaks.size() + " peaks");
	    //Get the K4me3 peaks as well
	    String k4me3PeakFileName = !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak" : "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/MmDataPeakFiles/H3K4Me3_KCl_peaks_names";
	    Vector<Peak> k4me3Peaks = !species.equals("mm") ? new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(k4me3PeakFileName, cellType, chr)) : new Vector<Peak>(FindEnhancerSites.parsePeakFile(k4me3PeakFileName, chr).values());
	    System.out.println("Found " + k4me3Peaks.size() + " H3K4me3 peaks");

	    String randomLocsFileName = "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/Transcripts/other_random_tfbs_locations";

	    int nChr = FindTranscripts.numberOfChromosomes(species), readLength = 35;
	    String rnaFileName = fileName + ".tsv";
	    RefSeqUCSCEnsemblGenes geneInfo[] = new RefSeqUCSCEnsemblGenes[nChr];
	    //Keep track of the number of conserved islands that are explained by exons that are expressed
	    TreeSet<ConservationIsland> conservedExpressedExons = new TreeSet<ConservationIsland>(), conservedExpressedNcExons = new TreeSet<ConservationIsland>(), conservedExpressedERNAs = new TreeSet<ConservationIsland>(), conservedExpressedDivergent = new TreeSet<ConservationIsland>(), conservedExpressedAS = new TreeSet<ConservationIsland>(), conservedExpressedPeaks = new TreeSet<ConservationIsland>();

	    RepeatInfo repeatInfo[] = new RepeatInfo[nChr];
	    TreeMap<Integer, Integer> remainingReadsHist = new TreeMap<Integer, Integer>();
	    //use the normalization factor to calculate the densities of the regions that we pick up without using the transcript caller
	    double normFactor = 1e6/FindReadsAcrossFeature2.countRNAReads(fileName, nChr);
	    //Use this structure to keep track of clusters with
	    //significant transcription at divergent sites, near
	    //peaks, enhancers and other extragenic sites.
	    TreeMap<String, TreeMap<Integer, Integer>> unannotatedReadsHist = new TreeMap<String, TreeMap<Integer, Integer>>();
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    System.out.println("Saving results to " + saveFileName);
	    String headers[] = new String[]{"Genic", "Genic_inactive", "ncRNA", "ncRNA_inactive", "eRNA_extra", "eRNA_intra", "Divergent", "Near_genes", "Near_ncRNA", "Other_AS", "Unannotated_distal", "Near_other_peaks", "Remaining"};
	    if (chr==-1 || chr==1) {
		bw.write("#chr");
		for (int i=0; i<headers.length; i++)
		    bw.write("\t" + headers[i] + "\t" + headers[i] + "_cons\t" + headers[i] + "_bp\t" + headers[i] + "_bp");
		bw.write("\n");
	    }
	    //Thresholds for determining the required number of reads for transcribed regions
	    int minReadsDivergent = minReads, minReadsEnhancers = minReads;
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    //Load the annotation
		    geneInfo[i-1] = new RefSeqUCSCEnsemblGenes(i, orchestra, species);
		    repeatInfo[i-1] = new RepeatInfo(i, forward, orchestra, species);
		    //Get the exons as well
		    TreeSet<Exon> exons = FindConservationIslands.parseExons(exonFileName, species, i);
		    TreeSet<Exon> macroRNAExons = FindConservationIslands.parseMacroRNAExons(macroRNAFileName, i);
		    TreeSet<Exon> lincRNAExons = FindConservationIslands.parseLincRNAExons(lncRNAExonFileName, i, lncRNAs);
		    ConservationIslands conservationIslands = new ConservationIslands(i, orchestra, species);
		    bw.write(i + "\t");

		    //Load reads
		    ABIRead[] rnaReadsForward = ABIReads.loadMRNAReadsChr(rnaFileName, true, i, readLength, false);
		    ABIRead[] rnaReadsReverse = ABIReads.loadMRNAReadsChr(rnaFileName, false, i, readLength, false);
		    System.out.println("Found " + rnaReadsForward.length + " reads forward and " + rnaReadsReverse.length + " reverse reads for chr " + i);
		    //Investigate random regions as well to use as negative controls
		    //int randLocVec[] = FindReadsAcrossFeature2.parseFeatureFile(randomLocsFileName + "_" + i + ".txt", i);
		    //boolean randStrandVec[] = FindReadsAcrossFeature2.parseFeatureFileStrandInfo(randomLocsFileName + "_" + i + ".txt", i, randLocVec);

		    //Set reads to genes and transcripts
		    geneInfo[i-1].setFactorReads(rnaFileName, rnaReadsReverse, -1);
		    geneInfo[i-1].setFactorReads(rnaFileName, rnaReadsForward, 1);
		    geneInfo[i-1].setGeneIDsKeep(true, rnaFileName, minDens);
		    geneInfo[i-1].setGeneIDsKeep(false, rnaFileName, minDens);
		    //Set reads to the ncRNAs
		    AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, rnaFileName, forward ? rnaReadsForward : rnaReadsReverse, i, forward ? 1 : -1);
		    AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, rnaFileName, forward ? rnaReadsForward : rnaReadsReverse, i);
		    AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, forward ? rnaReadsForward : rnaReadsReverse, i, forward ? 1 : -1);
		    //Load the called transcripts
		    ReducedModel2 reducedModel = new ReducedModel2(bedFileName  + FindTranscripts.BED_FILE_SUFFIX + (forward ? "_f.bed" : "_r.bed"), forward ? rnaReadsForward : rnaReadsReverse, rnaFileName, i, forward, species);
		    //Assign categories
		    reducedModel.compare2Annotation(maxDistance, geneInfo[i-1], lncRNAs, miRNAs, macroRNAs, allEnhancers);
		    reducedModel.categorizeUnannotatedTranscribedRegions(maxEnhancerDist, species.equals("mm"));
		    /////////////////////////////////
		    //Keep track of which bps are covered by transcripts
		    TreeMap<Integer, Integer> bpsCoveredAnnotation = new TreeMap<Integer, Integer>(), bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    ////////////////////////////////////////////
		    //Find out how many reads in transcripts/genes can be explained by the presence of a nearby K4me3 peak.
		    String saveReadsFileName = fileName + "_reads_initiated_by_peaks_" + (forward ? "f" : "r") + "_" + i + ".txt", readDensSaveFile = fileName + "_transcripts_dens_" + (forward ? "f" : "r") + "_" + i + ".txt";
		    System.out.println("Saving reads initiated by peaks to " + saveReadsFileName);
		    TreeMap<String, Integer> readsDrivenByPeaks = findReadsDrivenByPeaks(chr, forward, k4me3Peaks, peaks, rnaFileName, geneInfo[i-1], reducedModel, macroRNAs, lncRNAs, 5000, saveReadsFileName);
		    ////////////////////////////////////////////
		    int startReads = forward ? rnaReadsForward.length : rnaReadsReverse.length;
		    //Subtract the annotated reads from the total
		    ABIRead readsRemaining[] = subtractReads(forward ? rnaReadsForward : rnaReadsReverse, geneInfo[i-1], true, rnaFileName, forward, bpsCoveredAnnotation, minDens, readDensSaveFile.replace("dens", "coding_dens"), false, normFactor, true, exons, conservationIslands, conservedExpressedExons);
		    readsRemaining = subtractReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.CODING_STR, TranscribedRegion2.CODING_FRAGMENT_STR, TranscribedRegion2.MULTI_GENE_STR}, rnaFileName, forward, bpsCoveredTranscripts, minDens, true, /*readDensSaveFile.replace("dens", "coding_dens"),*/ normFactor, exons, conservationIslands, conservedExpressedExons);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredAnnotation, bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //Do the inactive genes
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    bpsCoveredAnnotation = new TreeMap<Integer, Integer>();
		    readsRemaining = subtractReads(readsRemaining, geneInfo[i-1], true, rnaFileName, forward, bpsCoveredAnnotation, minDens, false);
		    readsRemaining = subtractReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.CODING_STR, TranscribedRegion2.CODING_FRAGMENT_STR, TranscribedRegion2.MULTI_GENE_STR}, rnaFileName, forward, bpsCoveredTranscripts, minDens, false);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredAnnotation, bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //////////////////////
		    //Find out how many are at ncRNAs
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    bpsCoveredAnnotation = new TreeMap<Integer, Integer>();
		    readsRemaining = subtractReads(readsRemaining, geneInfo[i-1], false, rnaFileName, forward, bpsCoveredAnnotation, minDens, readDensSaveFile.replace("dens", "ncRNA_dens"), true, normFactor, true, exons, conservationIslands, conservedExpressedNcExons);
		    readsRemaining = subtractReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.NON_CODING_ANNOTATED_STR, TranscribedRegion2.MI_RNA_STR, TranscribedRegion2.LNC_RNA_STR, TranscribedRegion2.MACRO_RNA_STR}, rnaFileName, forward, bpsCoveredTranscripts, minDens, true, /*readDensSaveFile.replace("dens", "ncRNA_dens"),*/ normFactor, exons, conservationIslands, conservedExpressedNcExons);
		    readsRemaining = subtractMacroRNAReads(readsRemaining, macroRNAs, rnaFileName, i, forward, bpsCoveredAnnotation, minDens, true, exons, conservationIslands, conservedExpressedExons);
		    //readsRemaining = subtractMiRNAReads(readsRemaining, miRNAs, rnaFileName, i, forward);
		    readsRemaining = subtractLncRNAReads(readsRemaining, lncRNAs, rnaFileName, i, forward, bpsCoveredAnnotation, minDens, true, exons, conservationIslands, conservedExpressedExons);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredAnnotation, bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //Inactive ncRNAs
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    bpsCoveredAnnotation = new TreeMap<Integer, Integer>();
		    readsRemaining = subtractReads(readsRemaining, geneInfo[i-1], false, rnaFileName, forward, bpsCoveredAnnotation, minDens, false);
		    readsRemaining = subtractReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.NON_CODING_ANNOTATED_STR, TranscribedRegion2.MI_RNA_STR, TranscribedRegion2.LNC_RNA_STR, TranscribedRegion2.MACRO_RNA_STR}, rnaFileName, forward, bpsCoveredTranscripts, minDens, false);
		    readsRemaining = subtractMacroRNAReads(readsRemaining, macroRNAs, rnaFileName, i, forward, bpsCoveredAnnotation, minDens, false);
		    //readsRemaining = subtractMiRNAReads(readsRemaining, miRNAs, rnaFileName, i, forward);
		    readsRemaining = subtractLncRNAReads(readsRemaining, lncRNAs, rnaFileName, i, forward, bpsCoveredAnnotation, minDens, false);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredAnnotation, bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //////////////////////
		    //Find out how many are near enhancers
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    if (species.equals("mm")) {
			readsRemaining = subtractReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.ERNA_STR}, rnaFileName, forward, bpsCoveredTranscripts, minDens, true, readDensSaveFile.replace("dens", "eRNA_dens"), normFactor, null, conservationIslands, conservedExpressedERNAs);
			ABIRead extragenicERNAReads[] = assignEnhancerReads(readsRemaining, extragenicEnhancers, maxEnhancerDist, rnaFileName, i, forward, unannotatedReadsHist, "extragenic_enhancers", bpsCoveredTranscripts, minReadsEnhancers, readDensSaveFile.replace("dens", "eRNA_dens"), true, normFactor, conservationIslands, conservedExpressedERNAs);
			readsRemaining = subtractReads(readsRemaining, extragenicERNAReads);
		    }
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;


		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    if (species.equals("mm")) {
			ABIRead intragenicERNAReads[] = assignEnhancerReads(readsRemaining, intragenicEnhancers, maxEnhancerDist, rnaFileName, i, forward, geneInfo[i-1], unannotatedReadsHist, "intragenic_enhancers", bpsCoveredTranscripts, minReadsEnhancers, readDensSaveFile.replace("dens", "eRNA_dens"), true, normFactor, conservationIslands, conservedExpressedERNAs);
			readsRemaining = subtractReads(readsRemaining, intragenicERNAReads);
		    }
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //////////////////////
		    //Find out how many are divergent
		    //assignReads(forward ? rnaReadsForward : rnaReadsReverse, randLocVec, randStrandVec, i, forward, maxDistance, unannotatedReadsHist, "extragenic_random_" + maxDistance);
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    bpsCoveredAnnotation = new TreeMap<Integer, Integer>();
		    readsRemaining = subtractReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.DIVERGENT_STR}, rnaFileName, forward, bpsCoveredTranscripts, minDens, true, readDensSaveFile.replace("dens", "divergent_dens"), normFactor, null, conservationIslands, conservedExpressedDivergent);
		    readsRemaining = subtractNearbyReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.CODING_STR, TranscribedRegion2.CODING_FRAGMENT_STR, TranscribedRegion2.MULTI_GENE_STR}, rnaFileName, !forward, maxDistance, windowSize, bpsCoveredTranscripts, minReadsDivergent, readDensSaveFile.replace("dens", "divergent_dens"), true, normFactor, conservationIslands, conservedExpressedDivergent);
		    readsRemaining = subtractNearbyReads(readsRemaining, geneInfo[i-1], true, rnaFileName, !forward, maxDistance, 0, windowSize, bpsCoveredAnnotation, minReadsDivergent, readDensSaveFile.replace("dens", "divergent_dens"), true, normFactor, conservationIslands, conservedExpressedDivergent);
		    readsRemaining = subtractNearbyReads(readsRemaining, geneInfo[i-1], false, rnaFileName, !forward, 0, maxDistance, windowSize, bpsCoveredAnnotation, minReadsDivergent, conservationIslands, conservedExpressedDivergent);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredAnnotation, bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //////////////////////
		    //Find other reads that are within 10 kb of an annotated star/end or start/end of one of the categorized annotated on the sense strand.
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    bpsCoveredAnnotation = new TreeMap<Integer, Integer>();
		    readsRemaining = subtractNearbyReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.CODING_STR, TranscribedRegion2.CODING_FRAGMENT_STR, TranscribedRegion2.MULTI_GENE_STR}, rnaFileName, forward, maxDistance, windowSize, bpsCoveredTranscripts, minReadsDivergent);
		    readsRemaining = subtractNearbyReads(readsRemaining, geneInfo[i-1], true, rnaFileName, forward, maxDistance, 0, windowSize, bpsCoveredAnnotation, minReadsDivergent, conservationIslands, conservedExpressedDivergent);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredAnnotation, bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //////////////////////
		    //Find out how many are near the ncRNAs
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    bpsCoveredAnnotation = new TreeMap<Integer, Integer>();
		    readsRemaining = subtractNearbyReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.NON_CODING_ANNOTATED_STR, TranscribedRegion2.MACRO_RNA_STR, TranscribedRegion2.MI_RNA_STR, TranscribedRegion2.LNC_RNA_STR}, rnaFileName, forward, maxDistance, windowSize, bpsCoveredTranscripts, minReadsDivergent);
		    readsRemaining = subtractNearbyReads(readsRemaining, geneInfo[i-1], false, rnaFileName, forward, 0, maxDistance, windowSize, bpsCoveredAnnotation, minReadsDivergent);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredAnnotation, bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredAnnotation, bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //////////////////////
		    //Find out how many are AS, consider only the ones in transcribed regions
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    bpsCoveredAnnotation = new TreeMap<Integer, Integer>();
		    readsRemaining = subtractReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.NON_CODING_AS_STR}, rnaFileName, forward, bpsCoveredTranscripts, minDens, true, readDensSaveFile.replace("dens", "otherAS_dens"), normFactor, null, conservationIslands, conservedExpressedAS);
		    readsRemaining = subtractReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.CODING_STR, TranscribedRegion2.CODING_FRAGMENT_STR, TranscribedRegion2.MULTI_GENE_STR}, rnaFileName, !forward, new TreeMap<Integer, Integer>(), minDens, true, normFactor, null, conservationIslands, conservedExpressedAS);
		    readsRemaining = subtractReads(readsRemaining, geneInfo[i-1], true, rnaFileName, !forward, new TreeMap<Integer, Integer>(), minDens);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //////////////////////
		    //Find out how many are in unannotated transcripts
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    readsRemaining = subtractReads(readsRemaining, reducedModel, new String[]{TranscribedRegion2.NON_CODING_OTHER_RNA_STR}, rnaFileName, forward, bpsCoveredTranscripts, minDens, true, readDensSaveFile.replace("dens", "novel_dens"), normFactor, null, null, null);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    //////////////////////
		    //Find out how many are close to other peaks
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    //assignReads(readsRemaining, randLocVec, i, forward, maxEnhancerDist, unannotatedReadsHist, "extragenic_random_" + maxEnhancerDist);
		    readsRemaining = subtractReads(readsRemaining, peaks, tfNamesInStudy, i, forward, maxEnhancerDist, unannotatedReadsHist, "other_peaks", bpsCoveredTranscripts, minReadsEnhancers, readDensSaveFile.replace("dens", "peaks_dens"), false, normFactor, conservationIslands, conservedExpressedPeaks);
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredTranscripts) + "\t");
		    startReads = readsRemaining.length;
		    /////////////////////////////////
		    //Check out how much seems to be coming from repeat elements
		    readsRemaining = subtractReads(readsRemaining, repeatInfo[i-1], i, forward);
		    //bw.write((startReads - readsRemaining.length) + "\t");
		    startReads = readsRemaining.length;
		    //bw.write(readsRemaining.length + "\t");
		    /////////////////////////////////
		    //Check the clustering of the remaining reads
		    bpsCoveredTranscripts = new TreeMap<Integer, Integer>();
		    String saveName = "RemainingReadsHistograms/remaining_reads_transcripts_min_reads_" + minReads + "_" + (forward ? "f" : "r") + (chr==-1 ? "" : "_" + chr) + ".txt";
		    remainingReadsHist = getReadsHistogram(readsRemaining, windowSize, i, forward, geneInfo[i-1], reducedModel, macroRNAs, lncRNAs, minReads, saveName, bpsCoveredTranscripts, species, readDensSaveFile.replace("dens", "remaining_dens"), false, normFactor);
		    unannotatedReadsHist.put("remaining", remainingReadsHist);
		    //Create a null model whereby the same amount of reads are randomly assigned throughout the genome
		    ABIRead[] remainingReadsRandom = generateRandomReadsModel(i, remainingReadsHist, forward, geneInfo[i-1], reducedModel, macroRNAs, lncRNAs, species);
		    unannotatedReadsHist.put("remaining_random", getReadsHistogram(remainingReadsRandom, windowSize, i, forward, geneInfo[i-1], reducedModel, macroRNAs, lncRNAs, new TreeMap<Integer, Integer>(), species));
		    bw.write((startReads - readsRemaining.length) + "\t" + numberOfBpsTranscripts(bpsCoveredTranscripts) + "\t" + numberOfTranscriptsStartingConserved(bpsCoveredTranscripts, i, forward, conservationIslands) + "\t" + numberOfBpsCovered(bpsCoveredTranscripts) + "\n");
		}
	    }
	    bw.close();
	    //Save the number of CIs that are explained by transcription
	    String saveCIName = fileName + "_expressed_CIs_" + (forward ? "f" : "r") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    System.out.println("Saving expressed CIs to " + saveCIName);
	    bw = new BufferedWriter(new FileWriter(saveCIName));
	    if (chr==1 || chr==-1)
		bw.write("#Genic\tncRNA\teRNA\tDivergent\tAS\tPeaks\n");
	    TreeSet<ConservationIsland> conservationIslandsAlreadyExplained = new TreeSet<ConservationIsland>(conservedExpressedExons);
	    bw.write(conservedExpressedExons.size() + "\t");
	    bw.write(removeDuplicateConservedIslands(conservedExpressedNcExons, conservationIslandsAlreadyExplained) + "\t");
	    bw.write(removeDuplicateConservedIslands(conservedExpressedERNAs, conservationIslandsAlreadyExplained) + "\t");
	    bw.write(removeDuplicateConservedIslands(conservedExpressedDivergent, conservationIslandsAlreadyExplained) + "\t");
	    bw.write(removeDuplicateConservedIslands(conservedExpressedAS, conservationIslandsAlreadyExplained) + "\t");
	    bw.write(removeDuplicateConservedIslands(conservedExpressedPeaks, conservationIslandsAlreadyExplained) + "\n");
	    bw.close();

	    for (String cat : unannotatedReadsHist.keySet()) {
		String saveName = "RemainingReadsHistograms/" + fileName.substring(fileName.lastIndexOf("/") + 1) + "_" + cat + "_reads_histogram_" + (forward ? "f" : "r") + (chr==-1 ? "" : "_" + chr) + ".txt";
		System.out.println("Saving histogram to " + saveName);
		bw = new BufferedWriter(new FileWriter(saveName));
		for (Integer k : unannotatedReadsHist.get(cat).keySet())
		    bw.write(k + "\t" + unannotatedReadsHist.get(cat).get(k) + "\n");
		bw.close();
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////

    public static ABIRead[] subtractReads(ABIRead[] rnaReads, RefSeqUCSCEnsemblGenes geneInfo, boolean coding, String fileName, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens) throws IOException { return subtractReads(rnaReads, geneInfo, coding, fileName, forward, bpsCovered, minDens, null, false, 0, true, null, null, null); }

    public static ABIRead[] subtractReads(ABIRead[] rnaReads, RefSeqUCSCEnsemblGenes geneInfo, boolean coding, String fileName, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, boolean overThres) throws IOException { return subtractReads(rnaReads, geneInfo, coding, fileName, forward, bpsCovered, minDens, null, false, 0, overThres, null, null, null); }

    public static ABIRead[] subtractReads(ABIRead[] rnaReads, RefSeqUCSCEnsemblGenes geneInfo, boolean coding, String fileName, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, String saveDensFileName, boolean appendSaveDensFileName, double normFactor, boolean overThres, TreeSet<Exon> exons, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) throws IOException {
	//Subtract the reads in annotated regions from the total
	//number and return the remaining reads. Also, find out how
	//many bps are covered and how many of the exons that are
	//conserved.
	BufferedWriter bw = saveDensFileName==null ? null : new BufferedWriter(new FileWriter(saveDensFileName, appendSaveDensFileName));	TreeSet<ABIRead> reads2remove = new TreeSet<ABIRead>();
	Vector<ABIRead> remainingReads = new Vector<ABIRead>();
	//Since multiple genes are often overlapping, we want to avoid double counting and only consider the one of the genes at a given loci, this is especially problematic for non-coding RNAs
	for (RefSeqUCSCEnsemblGene g : geneInfo.getNonOverlappingGenes(forward)) {
	    if (g.isNonCoding()==!coding) {
		ABIRead[] rs = g.getFactorReads(fileName);
		//Make sure that the read density is high enough
		if (g.getReadDensity(fileName)>minDens==overThres) {
		    for (int i=0; i<rs.length; i++) 
			reads2remove.add(rs[i]);
		    bpsCovered.put(forward ? g.getStart() : g.getEnd(), forward ? g.getEnd() : g.getStart());
		    if (conservedExpressedExons!=null)
			conservedExpressedExons.addAll(conservationIslands.conservedExons(g.getOverlappingExons(exons)).keySet());
		    if (bw!=null) 
			bw.write(format.format((normFactor*(double)g.getFactorReads(fileName).length)/g.getLength()) + "\n");
		}
	    }
	}
	for (int i=0; i<rnaReads.length; i++) {
	    if (!reads2remove.contains(rnaReads[i]))
		remainingReads.addElement(rnaReads[i]);
	}
	ABIRead[] readsRemaining = new ABIRead[remainingReads.size()];
	for (int i=0; i<remainingReads.size(); i++)
	    readsRemaining[i] = remainingReads.elementAt(i);
	if (bw!=null)
	    bw.close();
	return readsRemaining;
    }

    public static ABIRead[] subtractMacroRNAReads(ABIRead[] rnaReads, TreeSet<MacroRNA> macroRNAs, String fileName, int chr, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, boolean overThres) { return subtractMacroRNAReads(rnaReads, macroRNAs, fileName, chr, forward, bpsCovered, minDens, overThres, null, null, null); }

    public static ABIRead[] subtractMacroRNAReads(ABIRead[] rnaReads, TreeSet<MacroRNA> macroRNAs, String fileName, int chr, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, boolean overThres, TreeSet<Exon> exons, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) {
	//Subtract the reads in annotated regions from the total
	//number and return the remaining reads.
	TreeSet<ABIRead> reads2remove = new TreeSet<ABIRead>();
	Vector<ABIRead> remainingReads = new Vector<ABIRead>();
	for (MacroRNA m : macroRNAs) {
	    if (m.getStrand()==forward && m.getChromosome()==chr) {
		ABIRead[] rs = m.getFactorReads(fileName);
		if (rs!=null) {
		    if (m.getReadDensity(fileName)>minDens==overThres) {
			for (int i=0; i<rs.length; i++) 
			    reads2remove.add(rs[i]);
			bpsCovered.put(forward ? m.getStart() : m.getEnd(), forward ? m.getEnd() : m.getStart());
			if (conservedExpressedExons!=null)
			    conservedExpressedExons.addAll(conservationIslands.conservedExons(m.getOverlappingExons(exons)).keySet());
		    }
		}
	    }
	}
	for (int i=0; i<rnaReads.length; i++) {
	    if (!reads2remove.contains(rnaReads[i]))
		remainingReads.addElement(rnaReads[i]);
	}
	ABIRead[] readsRemaining = new ABIRead[remainingReads.size()];
	for (int i=0; i<remainingReads.size(); i++)
	    readsRemaining[i] = remainingReads.elementAt(i);
	return readsRemaining;
    }

    public static ABIRead[] subtractMiRNAReads(ABIRead[] rnaReads, TreeSet<MiRNA> miRNAs, String fileName, int chr, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens) {
	//Subtract the reads in annotated regions from the total
	//number and return the remaining reads.
	TreeSet<ABIRead> reads2remove = new TreeSet<ABIRead>();
	Vector<ABIRead> remainingReads = new Vector<ABIRead>();
	for (MiRNA m : miRNAs) {
	    if (m.getStrand()==forward && m.getChromosome()==chr) {
		ABIRead[] rs = m.getFactorReads(fileName);
		if (rs!=null) {
		    for (int i=0; i<rs.length; i++) 
			reads2remove.add(rs[i]);
		    if (m.getReadDensity(fileName)>minDens)
			bpsCovered.put(forward ? m.getStart() : m.getEnd(), forward ? m.getEnd() : m.getStart());
		}
	    }
	}
	for (int i=0; i<rnaReads.length; i++) {
	    if (!reads2remove.contains(rnaReads[i]))
		remainingReads.addElement(rnaReads[i]);
	}
	ABIRead[] readsRemaining = new ABIRead[remainingReads.size()];
	for (int i=0; i<remainingReads.size(); i++)
	    readsRemaining[i] = remainingReads.elementAt(i);
	return readsRemaining;
    }

    public static ABIRead[] subtractLncRNAReads(ABIRead[] rnaReads, TreeSet<LncRNA> lncRNAs, String fileName, int chr, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, boolean overThres) { return subtractLncRNAReads(rnaReads, lncRNAs, fileName, chr, forward, bpsCovered, minDens, overThres, null, null, null); }

    public static ABIRead[] subtractLncRNAReads(ABIRead[] rnaReads, TreeSet<LncRNA> lncRNAs, String fileName, int chr, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, boolean overThres, TreeSet<Exon> exons, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) {
	//Subtract the reads in annotated regions from the total
	//number and return the remaining reads.
	TreeSet<ABIRead> reads2remove = new TreeSet<ABIRead>();
	Vector<ABIRead> remainingReads = new Vector<ABIRead>();
	for (LncRNA m : lncRNAs) {
	    if (m.getChromosome()==chr) {
		ABIRead[] rs = m.getFactorReads(fileName);
		if (rs!=null) {
		    if (m.getReadDensity(fileName)>minDens==overThres) {
			if (!overThres && random.nextBoolean()) {
			    for (int i=0; i<rs.length; i++) 
				reads2remove.add(rs[i]);
			    bpsCovered.put(m.getStart(), m.getEnd());
			    if (conservedExpressedExons!=null)
				conservedExpressedExons.addAll(conservationIslands.conservedExons(m.getOverlappingExons(exons)).keySet());
			}
		    }
		}
	    }
	}
	for (int i=0; i<rnaReads.length; i++) {
	    if (!reads2remove.contains(rnaReads[i]))
		remainingReads.addElement(rnaReads[i]);
	}
	ABIRead[] readsRemaining = new ABIRead[remainingReads.size()];
	for (int i=0; i<remainingReads.size(); i++)
	    readsRemaining[i] = remainingReads.elementAt(i);
	return readsRemaining;
    }

    public static ABIRead[] subtractReads(ABIRead[] reads, ReducedModel2 reducedModel, String categories[], String fileName, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, boolean overThres) throws IOException { return subtractReads(reads, reducedModel, categories, fileName, forward, bpsCovered, minDens, overThres, null, false, 0, null, null, null); }

    public static ABIRead[] subtractReads(ABIRead[] reads, ReducedModel2 reducedModel, String categories[], String fileName, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, boolean overThres, double normFactor, TreeSet<Exon> exons, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) throws IOException { return subtractReads(reads, reducedModel, categories, fileName, forward, bpsCovered, minDens, overThres, null, false, normFactor, exons, conservationIslands, conservedExpressedExons); }

    public static ABIRead[] subtractReads(ABIRead[] reads, ReducedModel2 reducedModel, String categories[], String fileName, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, boolean overThres, String saveDensFileName, double normFactor, TreeSet<Exon> exons, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) throws IOException { return subtractReads(reads, reducedModel, categories, fileName, forward, bpsCovered, minDens, overThres, saveDensFileName, false, normFactor, exons, conservationIslands, conservedExpressedExons); }

    public static ABIRead[] subtractReads(ABIRead[] reads, ReducedModel2 reducedModel, String categories[], String fileName, boolean forward, TreeMap<Integer, Integer> bpsCovered, double minDens, boolean overThres, String saveDensFileName, boolean appendSaveDensFileName, double normFactor, TreeSet<Exon> exons, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) throws IOException {
	BufferedWriter bw = saveDensFileName==null ? null : new BufferedWriter(new FileWriter(saveDensFileName, appendSaveDensFileName));
	TreeSet<ABIRead> reads2remove = new TreeSet<ABIRead>();
	Vector<ABIRead> remainingReads = new Vector<ABIRead>();
	for (int i=0; i<categories.length; i++) {
	    System.out.println("Found " + reducedModel.getSegments(categories[i]).size() + " transcripts for " + categories[i]);
	    for (TranscribedRegion2 t : reducedModel.getSegments(categories[i])) {
		if (t.getStrand()==forward) {
		    if (t.getReadDensity(fileName)>minDens==overThres) {
			ABIRead[] rs = t.getFactorReads(fileName);
			for (int j=0; j<rs.length; j++) 
			    reads2remove.add(rs[j]);
			bpsCovered.put(t.getStart(), t.getEnd());
			if (conservedExpressedExons!=null) {
			    if (exons==null) 
				//For unannotated transcripts (eg divergent) where we don't know about the splicing. A conservative assumption is to check the entire transcribed region for overlap with conservation islands
				conservedExpressedExons.addAll(conservationIslands.getOverlappingConservationIslands(t));
			    else
				conservedExpressedExons.addAll(conservationIslands.conservedExons(t.getOverlappingExons(exons)).keySet());
			}
			if (bw!=null)
			    bw.write(format.format((normFactor*(double)t.getFactorReads(fileName).length)/t.getLength()) + "\n");
		    }
		}
	    }
	    System.out.println("Found " + reads2remove.size() + " reads to remove for " + categories[i]);
	}
	for (int i=0; i<reads.length; i++) {
	    if (!reads2remove.contains(reads[i]))
		remainingReads.addElement(reads[i]);
	}
	ABIRead[] readsRemaining = new ABIRead[remainingReads.size()];
	for (int i=0; i<remainingReads.size(); i++)
	    readsRemaining[i] = remainingReads.elementAt(i);
	System.out.println("Returning " + readsRemaining.length + " reads.");
	if (bw!=null)
	    bw.close();
	return readsRemaining;
    }

    ////////////////////////////////////////////

    public static ABIRead[] subtractNearbyReads(ABIRead[] rnaReads, RefSeqUCSCEnsemblGenes geneInfo, boolean start, String fileName, boolean forward, int upstream, int downstream, int stepSize, TreeMap<Integer, Integer> bpsCovered, int minReads) throws IOException { return subtractNearbyReads(rnaReads, geneInfo, start, fileName, forward, upstream, downstream, stepSize, null, null, bpsCovered, minReads, null, false, 0, null, null); }

    public static ABIRead[] subtractNearbyReads(ABIRead[] rnaReads, RefSeqUCSCEnsemblGenes geneInfo, boolean start, String fileName, boolean forward, int upstream, int downstream, int stepSize, TreeMap<Integer, Integer> bpsCovered, int minReads, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressed) throws IOException { return subtractNearbyReads(rnaReads, geneInfo, start, fileName, forward, upstream, downstream, stepSize, null, null, bpsCovered, minReads, null, false, 0, conservationIslands, conservedExpressed); }

    public static ABIRead[] subtractNearbyReads(ABIRead[] rnaReads, RefSeqUCSCEnsemblGenes geneInfo, boolean start, String fileName, boolean forward, int upstream, int downstream, int stepSize, TreeMap<Integer, Integer> bpsCovered, int minReads, String saveDensFileName, boolean appendSaveDensFileName, double normFactor) throws IOException { return subtractNearbyReads(rnaReads, geneInfo, start, fileName, forward, upstream, downstream, stepSize, null, null, bpsCovered, minReads, saveDensFileName, appendSaveDensFileName, normFactor, null, null); }

    public static ABIRead[] subtractNearbyReads(ABIRead[] rnaReads, RefSeqUCSCEnsemblGenes geneInfo, boolean start, String fileName, boolean forward, int upstream, int downstream, int stepSize, TreeMap<Integer, Integer> bpsCovered, int minReads, String saveDensFileName, boolean appendSaveDensFileName, double normFactor, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressed) throws IOException { return subtractNearbyReads(rnaReads, geneInfo, start, fileName, forward, upstream, downstream, stepSize, null, null, bpsCovered, minReads, saveDensFileName, appendSaveDensFileName, normFactor, conservationIslands, conservedExpressed); }

    public static ABIRead[] subtractNearbyReads(ABIRead[] rnaReads, RefSeqUCSCEnsemblGenes geneInfo, boolean start, String fileName, boolean forward, int upstream, int downstream, int stepSize, TreeMap<String, TreeMap<Integer, Integer>> unannotatedReadsHist, String keyName, TreeMap<Integer, Integer> bpsCovered, int minReads) throws IOException { return subtractNearbyReads(rnaReads, geneInfo, start, fileName, forward, upstream, downstream, stepSize, unannotatedReadsHist, keyName, bpsCovered, minReads, null, false, 0, null, null); }
    
    public static ABIRead[] subtractNearbyReads(ABIRead[] rnaReads, RefSeqUCSCEnsemblGenes geneInfo, boolean start, String fileName, boolean forward, int upstream, int downstream, int stepSize, TreeMap<String, TreeMap<Integer, Integer>> unannotatedReadsHist, String keyName, TreeMap<Integer, Integer> bpsCovered, int minReads, String saveDensFileName, boolean appendSaveDensFileName, double normFactor, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressed) throws IOException {
	//Subtract the reads in annotated regions from the total
	//number and return the remaining reads. Since there tend to
	//be so many of the reads adjacent to annotated reads/genes we
	//take all the reads in the downstream and upstream
	//regions. We also go onwards with windows of stepSize as long
	//as there is anything contained there
	BufferedWriter bw = saveDensFileName==null ? null : new BufferedWriter(new FileWriter(saveDensFileName, appendSaveDensFileName));
	TreeMap<Integer, Integer> readsHist = new TreeMap<Integer, Integer>();
	TreeSet<ABIRead> reads2remove = new TreeSet<ABIRead>();
	Vector<ABIRead> remainingReads = new Vector<ABIRead>();
	for (RefSeqUCSCEnsemblGene g : geneInfo.getGenes()) {
	    if (g.getStrand()==forward) {
		ABIRead rs[];
		int s, e;
		if (start) {
		    s = g.getStart() + (forward ? -upstream : -downstream);
		    e = g.getStart() + (forward ? downstream : upstream);
		}
		else {
		    s = g.getEnd() + (forward ? -upstream : -downstream);
		    e = g.getEnd() + (forward ? downstream : upstream);
		}
		rs = getReadSubset(rnaReads, g.getChromosome(), forward, s, e);
		if (rs!=null) {
		    for (int i=0; i<rs.length; i++) 
			reads2remove.add(rs[i]);
		    if (!readsHist.containsKey(rs.length))
			readsHist.put(rs.length, 0);
		    readsHist.put(rs.length, readsHist.get(rs.length) + 1);
		    if (rs.length>minReads) {
			bpsCovered.put(s, e);
			if (conservedExpressed!=null)
			    conservedExpressed.addAll(conservationIslands.getOverlappingConservationIslands(geneInfo.getChromosome(), s, e));
		    }
		}
		/////////////////////////////////
		//contiue stepping outwards
		int nSteps = 1;
		while (rs!=null && rs.length>0) {
		    if (start) {
			s = g.getStart() + (forward ? -upstream - nSteps*stepSize: upstream + (nSteps - 1)*stepSize);
			e = g.getStart() + (forward ? -upstream - (nSteps - 1)*stepSize : upstream + nSteps*stepSize);
		    }
		    else {
			s = g.getEnd() + (forward ? downstream + (nSteps - 1)*stepSize : -downstream - nSteps*stepSize);
			e = g.getEnd() + (forward ? downstream + nSteps*stepSize : -downstream - (nSteps - 1)*stepSize);
		    }
		    rs = getReadSubset(rnaReads, g.getChromosome(), forward, s, e);
		    if (rs!=null) {
			for (int i=0; i<rs.length; i++) 
			    reads2remove.add(rs[i]);
			if (rs.length>minReads) {
			    bpsCovered.put(s, e);
			    if (conservedExpressed!=null)
				conservedExpressed.addAll(conservationIslands.getOverlappingConservationIslands(geneInfo.getChromosome(), s, e));
			}
		    }
		    nSteps++;
		}
		if (bw!=null && rs.length>minReads)
		    bw.write(format.format(((double)rs.length/(e-s))*normFactor) + "\n");
	    }
	}
	for (int i=0; i<rnaReads.length; i++) {
	    if (!reads2remove.contains(rnaReads[i]))
		remainingReads.addElement(rnaReads[i]);
	}
	ABIRead[] readsRemaining = new ABIRead[remainingReads.size()];
	for (int i=0; i<remainingReads.size(); i++)
	    readsRemaining[i] = remainingReads.elementAt(i);
	if (unannotatedReadsHist!=null)
	    unannotatedReadsHist.put(keyName, readsHist);
	if (bw!=null)
	    bw.close();
	return readsRemaining;
    }

    public static ABIRead[] subtractNearbyReads(ABIRead[] reads, ReducedModel2 reducedModel, String categories[], String fileName, boolean forward, int maxDistance, int stepSize, TreeMap<Integer, Integer> bpsCovered, int minReads) throws IOException { return subtractNearbyReads(reads, reducedModel, categories, fileName, forward, maxDistance, stepSize, bpsCovered, minReads, null, false, 0, null, null); }

    public static ABIRead[] subtractNearbyReads(ABIRead[] reads, ReducedModel2 reducedModel, String categories[], String fileName, boolean forward, int maxDistance, int stepSize, TreeMap<Integer, Integer> bpsCovered, int minReads, String saveDensFileName, boolean appendSaveDensFileName, double normFactor, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) throws IOException {
	//Remove the reads that are within maxDistance of the transcripts of type categories
	BufferedWriter bw = saveDensFileName==null ? null : new BufferedWriter(new FileWriter(saveDensFileName, appendSaveDensFileName));
	TreeSet<ABIRead> reads2remove = new TreeSet<ABIRead>();
	Vector<ABIRead> remainingReads = new Vector<ABIRead>();
	for (int i=0; i<categories.length; i++) {
	    for (TranscribedRegion2 t : reducedModel.getSegments(categories[i])) {
		if (t.getStrand()==forward) {
		    ABIRead rs[] = getReadSubset(reads, t.getChromosome(), forward, t.getStart() - maxDistance, t.getStart());
		    if (rs!=null) {
			for (int j=0; j<rs.length; j++) 
			    reads2remove.add(rs[j]);
			if (rs.length>minReads)
			    bpsCovered.put(t.getStart() - maxDistance, t.getStart());
			if (conservedExpressedExons!=null)
			    conservedExpressedExons.addAll(conservationIslands.getOverlappingConservationIslands(t.getChromosome(), t.getStart() - maxDistance, t.getStart()));
		    }
		    /////////////////////////////////
		    //contiue stepping outwards
		    int nSteps = 1;
		    while (rs!=null && rs.length>0) {
			rs = getReadSubset(reads, t.getChromosome(), forward, t.getStart() - maxDistance - nSteps*stepSize, t.getStart() - maxDistance - (nSteps - 1)*stepSize);
			if (rs!=null) {
			    for (int j=0; j<rs.length; j++) 
				reads2remove.add(rs[j]);
			    if (rs.length>minReads) {
				int s = t.getStart() - maxDistance - nSteps*stepSize, e = t.getStart() - maxDistance - (nSteps - 1)*stepSize;
				bpsCovered.put(s, e);
				if (conservedExpressedExons!=null)
				    conservedExpressedExons.addAll(conservationIslands.getOverlappingConservationIslands(t.getChromosome(), s, e));
			    }
			}
			nSteps++;
		    }
		    rs = getReadSubset(reads, t.getChromosome(), forward, t.getEnd(), t.getEnd() + maxDistance);
		    if (rs!=null) {
			for (int j=0; j<rs.length; j++) 
			    reads2remove.add(rs[j]);
			if (rs.length>minReads) {
			    bpsCovered.put(t.getEnd(), t.getEnd() + maxDistance);
			    if (conservedExpressedExons!=null)
				conservedExpressedExons.addAll(conservationIslands.getOverlappingConservationIslands(t.getChromosome(), t.getEnd(), t.getEnd() + maxDistance));
			}
		    }
		    nSteps = 1;
		    while (rs!=null && rs.length>0) {
			rs = getReadSubset(reads, t.getChromosome(), forward, t.getEnd() + maxDistance + (nSteps - 1)*stepSize, t.getEnd() + maxDistance + nSteps*stepSize);
			if (rs!=null) {
			    for (int j=0; j<rs.length; j++) 
				reads2remove.add(rs[j]);
			    if (rs.length>minReads) {
				int s = t.getEnd() + maxDistance + (nSteps - 1)*stepSize, e = t.getEnd() + maxDistance + nSteps*stepSize;
				bpsCovered.put(s, e);
				if (conservedExpressedExons!=null)
				    conservedExpressedExons.addAll(conservationIslands.getOverlappingConservationIslands(t.getChromosome(), s, e));
			    }
			    nSteps++;
			    if (bw!=null && rs.length>minReads)
				bw.write(format.format(((double)rs.length/(maxDistance + nSteps*stepSize)*normFactor)) + "\n");
			}
		    }
		}
	    }
	    System.out.println("Found " + reads2remove.size() + " reads to remove near " + categories[i]);
	}
	for (int i=0; i<reads.length; i++) {
	    if (!reads2remove.contains(reads[i]))
		remainingReads.addElement(reads[i]);
	}
	ABIRead[] readsRemaining = new ABIRead[remainingReads.size()];
	for (int i=0; i<remainingReads.size(); i++)
	    readsRemaining[i] = remainingReads.elementAt(i);
	System.out.println("Returning " + readsRemaining.length + " reads.");
	if (bw!=null)
	    bw.close();
	return readsRemaining;
    }
	
    public static ABIRead[] getReadSubset(ABIRead reads[], int chr, boolean forward, int start, int end) {
	int startInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, start));
	startInd = startInd<0 ? -startInd-1 : startInd;
	int endInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, end));
	endInd = endInd<0 ? -endInd-1 : endInd;
	ABIRead rs[] = new ABIRead[endInd - startInd];
	for (int i=startInd; i<endInd; i++) 
	    rs[i-startInd] = reads[i]; 
	return rs;
    }

    public static boolean isArtefact(ABIRead rs[]) {
	//Use this method to look for spikes with lots of reads in the same location and no others nearby.
	int nReads = rs.length, nReadsFromOneBase = 0, maxReadsFromOneBase = 0;
	if (nReads>=MIN_READS_FOR_ARTEFACT) {
	    if (readsOverlapping(rs))
		return true;
	    for (int i=1; i<rs.length; i++) {
		if (rs[i-1].equals(rs[i])) {
		    nReadsFromOneBase++;
		    maxReadsFromOneBase = nReadsFromOneBase>maxReadsFromOneBase ? nReadsFromOneBase : maxReadsFromOneBase;
		}
		else 
		    nReadsFromOneBase = 0;
	    }
	    return (double)maxReadsFromOneBase/nReads>=MIN_READS_FROM_ONE_BASE_PROPORTION_FOR_ARTEFACT || maxReadsFromOneBase>MIN_READS_FROM_ONE_BASE_PROPORTION_FOR_ARTEFACT*10;
	}
	return false;
    }

    //public static int maxReadsFromOneBase(

    public static ABIRead[] assignEnhancerReads(ABIRead[] reads, TreeSet<Enhancer2> enhancers, int maxDist, String rnaFileName, int chr, boolean forward, TreeMap<String, TreeMap<Integer, Integer>> unannotatedReadsHist, String keyName, TreeMap<Integer, Integer> bpsCovered, int minReads, String saveDensFileName, boolean appendSaveDensFileName, double normFactor, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) throws IOException {
	//Use for extragenic enhancers
	BufferedWriter bw = saveDensFileName==null ? null : new BufferedWriter(new FileWriter(saveDensFileName, appendSaveDensFileName));
	TreeMap<Integer, Integer> readsHist = new TreeMap<Integer, Integer>();
	Vector<ABIRead> readsKeep = new Vector<ABIRead>();
	for (Enhancer2 e : enhancers) {
	    if (e.getChromosome()==chr) {
		ABIRead rs[] = getReadSubset(reads, chr, forward, e.getCenter() + (forward ? 0 : -maxDist), e.getCenter() + (forward ? maxDist : 0));
		if (!isArtefact(rs)) {
		    e.setFactorReads(rnaFileName, rs);
		    if (!readsHist.containsKey(rs.length))
			readsHist.put(rs.length, 0);
		    readsHist.put(rs.length, readsHist.get(rs.length) + 1);
		    for (int i=0; i<rs.length; i++)
			readsKeep.addElement(rs[i]);
		    if (rs.length>minReads) {
			bpsCovered.put(e.getCenter() + (forward ? 0 : -maxDist), e.getCenter() + (forward ? maxDist : 0));
			if (conservedExpressedExons!=null)
			    conservedExpressedExons.addAll(conservationIslands.getOverlappingConservationIslands(e.getChromosome(), e.getCenter() + (forward ? 0 : -maxDist), e.getCenter() + (forward ? maxDist : 0), forward ? 1 : -1));
			if (bw!=null)
			    bw.write(format.format(((double)rs.length/maxDist)*normFactor) + "\n");
			System.out.println("Assigned " + rs.length + " reads to " + e.toString());
		    }
		}
	    }
	}
	System.out.println((bpsCovered.size()*maxDist) + " bps removed for extragenic and removing " + readsKeep.size());
	ABIRead eRNAs[] = new ABIRead[readsKeep.size()];
	for (int i=0; i<eRNAs.length; i++)
	    eRNAs[i] = readsKeep.elementAt(i);
	unannotatedReadsHist.put(keyName, readsHist);
	if (bw!=null)
	    bw.close();
	return eRNAs;
    }

    public static ABIRead[] assignEnhancerReads(ABIRead[] reads, TreeSet<Enhancer2> enhancers, int maxDist, String rnaFileName, int chr, boolean forward, RefSeqUCSCEnsemblGenes geneInfo, TreeMap<String, TreeMap<Integer, Integer>> unannotatedReadsHist, String keyName, TreeMap<Integer, Integer> bpsCovered, int minReads, String saveDensFileName, boolean appendSaveDensFileName, double normFactor, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) throws IOException {
	//Use for intragenic enhancers
	BufferedWriter bw = saveDensFileName==null ? null : new BufferedWriter(new FileWriter(saveDensFileName, appendSaveDensFileName));
	TreeMap<Integer, Integer> readsHist = new TreeMap<Integer, Integer>();
	Vector<ABIRead> readsKeep = new Vector<ABIRead>();
	for (Enhancer2 e : enhancers) {
	    if (e.getChromosome()==chr) {
		//Find the overlapping gene
		RefSeqUCSCEnsemblGene g = geneInfo.getOverlappingGene(e.getCenter());
		if (g!=null && g.getStrand()!=forward) {
		    ABIRead rs[] = getReadSubset(reads, chr, forward, e.getCenter() + (forward ? 0 : -maxDist), e.getCenter() + (forward ? maxDist : 0));
		    if (!isArtefact(rs)) {
			e.setFactorReads(rnaFileName, rs);
			if (!readsHist.containsKey(rs.length))
			    readsHist.put(rs.length, 0);
			readsHist.put(rs.length, readsHist.get(rs.length) + 1);
			for (int i=0; i<rs.length; i++)
			    readsKeep.addElement(rs[i]);
			if (rs.length>minReads) {
			    bpsCovered.put(e.getCenter() + (forward ? 0 : -maxDist), e.getCenter() + (forward ? maxDist : 0));
			    if (conservedExpressedExons!=null)
				conservedExpressedExons.addAll(conservationIslands.getOverlappingConservationIslands(e.getChromosome(), e.getCenter() + (forward ? 0 : -maxDist), e.getCenter() + (forward ? maxDist : 0), forward ? 1 : -1));
			    if (bw!=null)
				bw.write(format.format(((double)rs.length/maxDist)*normFactor) + "\n");
			}
		    }
		}
	    }
	}
	System.out.println((bpsCovered.size()*maxDist) + " bps removed for intragenic and removing " + readsKeep.size());
	ABIRead eRNAs[] = new ABIRead[readsKeep.size()];
	for (int i=0; i<eRNAs.length; i++)
	    eRNAs[i] = readsKeep.elementAt(i);
	unannotatedReadsHist.put(keyName, readsHist);
	if (bw!=null)
	    bw.close();
	return eRNAs;
    }

    public static ABIRead[] subtractReads(ABIRead[] reads, ABIRead[] subtractReads) {
	Vector<ABIRead> readsKeep = new Vector<ABIRead>();
	for (int i=0; i<reads.length; i++) {
	    if (Arrays.binarySearch(subtractReads, reads[i])<0)
		readsKeep.addElement(reads[i]);
	}
	ABIRead[] rs = new ABIRead[readsKeep.size()];
	for (int i=0; i<rs.length; i++)
	    rs[i] = readsKeep.elementAt(i);
	return rs;
    }

    public static ABIRead[] subtractReads(ABIRead[] reads, Vector<Peak> peaks, String[] tfNamesInStudy, int chr, boolean forward, int maxDist, TreeMap<String, TreeMap<Integer, Integer>> unannotatedReadsHist, String keyName, TreeMap<Integer, Integer> bpsCovered, int minReads, String saveDensFileName, boolean appendSaveDensFileName, double normFactor, ConservationIslands conservationIslands, TreeSet<ConservationIsland> conservedExpressedExons) throws IOException {
	//Remove reads that are within maxDist of a peak
	BufferedWriter bw = saveDensFileName==null ? null : new BufferedWriter(new FileWriter(saveDensFileName, appendSaveDensFileName));
	TreeMap<Integer, Integer> readsHist = new TreeMap<Integer, Integer>();
	TreeSet<ABIRead> reads2remove = new TreeSet<ABIRead>();
	Vector<ABIRead> remainingReads = new Vector<ABIRead>();
	for (Peak p : peaks) {
	    if (p.getChromosome()==chr && p.hasFactor(tfNamesInStudy)) {
		ABIRead[] rs = getReadSubset(reads, chr, forward, p.getCenter() + (forward ? 0 : -maxDist), p.getCenter() + (forward ? maxDist : 0));
		if (rs!=null) {
		    for (int j=0; j<rs.length; j++) 
			reads2remove.add(rs[j]);
		    if (!isArtefact(rs)) {
			if (rs.length>100)
			    System.out.println("Found " + rs.length + " reads near chr" + chr + ":" + (p.getCenter()-1e4) + "-" + (p.getCenter()+1e4));
			if (!readsHist.containsKey(rs.length))
			    readsHist.put(rs.length, 0);
			readsHist.put(rs.length, readsHist.get(rs.length) + 1);
			if (rs.length>minReads) {
			    bpsCovered.put(p.getCenter() + (forward ? 0 : -maxDist), p.getCenter() + (forward ? maxDist : 0));
			    if (conservedExpressedExons!=null)
				conservedExpressedExons.addAll(conservationIslands.getOverlappingConservationIslands(p.getChromosome(), p.getCenter() + (forward ? 0 : -maxDist), p.getCenter() + (forward ? maxDist : 0), forward ? 1 : -1));
			    if (bw!=null)
				bw.write(format.format(((double)rs.length/maxDist)*normFactor) + "\n");
			}
		    }
		}
	    }
	}
	System.out.println("Found " + reads2remove.size() + " reads to remove near peaks.");
	for (int i=0; i<reads.length; i++) {
	    if (!reads2remove.contains(reads[i]))
		remainingReads.addElement(reads[i]);
	}
	ABIRead[] readsRemaining = new ABIRead[remainingReads.size()];
	for (int i=0; i<remainingReads.size(); i++)
	    readsRemaining[i] = remainingReads.elementAt(i);
	System.out.println("Returning " + readsRemaining.length + " reads");
	unannotatedReadsHist.put(keyName, readsHist);
	if (bw!=null)
	    bw.close();
	return readsRemaining;
    }

    public static ABIRead[] subtractReads(ABIRead[] reads, RepeatInfo repeatInfo, int chr, boolean forward) {
	TreeSet<ABIRead> reads2remove = new TreeSet<ABIRead>();
	Vector<ABIRead> remainingReads = new Vector<ABIRead>();
	Arrays.sort(reads);
	Iterator<RepeatRegion> repeatIt = repeatInfo.getRepeatRegions().iterator();
	if (repeatIt.hasNext()) {
	    RepeatRegion r = repeatIt.next();
	    for (int j=0; j<reads.length; j++) {
		if (reads[j].getPosition()>r.getEnd() && repeatIt.hasNext())
		    r = repeatIt.next();
		if (r.overlaps(reads[j])) 
		    reads2remove.add(reads[j]);
	    }
	}
	System.out.println("Found " + reads2remove.size() + " reads to remove at repeat elements.");
	for (int i=0; i<reads.length; i++) {
	    if (!reads2remove.contains(reads[i]))
		remainingReads.addElement(reads[i]);
	}
	ABIRead[] readsRemaining = new ABIRead[remainingReads.size()];
	for (int i=0; i<remainingReads.size(); i++)
	    readsRemaining[i] = remainingReads.elementAt(i);
	return readsRemaining;
    }

    public static void assignReads(ABIRead[] reads, int[] randLocVec, int chr, boolean forward, int maxDistance, TreeMap<String, TreeMap<Integer, Integer>> unannotatedReadsHist, String keyName) {
	//Use this method to figure out how many reads are near random
	//regions. By default, we do not subtract them.
	TreeMap<Integer, Integer> readsHist = new TreeMap<Integer, Integer>();
	for (int i=0; i<randLocVec.length; i++) {
	    ABIRead rs[] = getReadSubset(reads, chr, forward, randLocVec[i] + (forward ? 0 : -maxDistance), randLocVec[i] + (forward ? maxDistance : 0));
	    if (!isArtefact(rs)) { 
		if (!readsHist.containsKey(rs.length))
		    readsHist.put(rs.length, 0);
		readsHist.put(rs.length, readsHist.get(rs.length) + 1);
	    }
	}
	unannotatedReadsHist.put(keyName, readsHist);
    }

    public static TreeMap<String, Integer> findReadsDrivenByPeaks(int chr, boolean forward, Vector<Peak> k4me3Peaks, Vector<Peak> tfPeaks, String rnaFileName, RefSeqUCSCEnsemblGenes geneInfo, ReducedModel2 reducedModel, TreeSet<MacroRNA> macroRNAs, TreeSet<LncRNA> lncRNAs, int maxDist, String saveReadsFileName) {
	//Find out how many reads that appear to be driven by the TF and the
	//K4me3 peaks. Separate into coding, non-coding and other categories.
	TreeMap<String, Integer> categories2readsMap = new TreeMap<String, Integer>(), categories2peaksMap = new TreeMap<String, Integer>();
	categories2readsMap.put("Coding,k4", 0);
	categories2readsMap.put("Noncoding,k4", 0);
	categories2readsMap.put("Other,k4", 0);

	categories2peaksMap.put("Coding,k4", 0);
	categories2peaksMap.put("Noncoding,k4", 0);
	categories2peaksMap.put("Other,k4", 0);
	
	//To avoid double-counting, we need to exclude the elements that have already been accounted for.
	TreeSet<RefSeqUCSCEnsemblGene> genesAccountedFor = new TreeSet<RefSeqUCSCEnsemblGene>();
	TreeSet<TranscribedRegion2> transcriptsAccountedFor = new TreeSet<TranscribedRegion2>();
	TreeSet<MacroRNA> macroRNAsAccountedFor = new TreeSet<MacroRNA>();
	TreeSet<LncRNA> lncRNAsAccountedFor = new TreeSet<LncRNA>();
	for (Peak p : k4me3Peaks) {
	    if (p.getChromosome()==chr) {
		RefSeqUCSCEnsemblGene nearbyTSS = geneInfo.getNearestTSSGene(p.getCenter(), forward);
		int nCodingReads = 0, nNonCodingReads = 0, nOtherReads = 0;
		boolean categoryFound = false;
		if (Math.abs(nearbyTSS.getStart() - p.getCenter())<maxDist) {
		    if (!genesAccountedFor.contains(nearbyTSS)) {
			if (nearbyTSS.isNonCoding()) {
			    nNonCodingReads = nearbyTSS.getFactorReads(rnaFileName).length;
			    System.out.println("Found " + nNonCodingReads + " from gene " + nearbyTSS.toString() + " near " + p.toString());
			}
			else {
			    nCodingReads = nearbyTSS.getFactorReads(rnaFileName).length;
			    //System.out.println("Found " + nCodingReads + " from gene " + nearbyTSS.toString() + " near " + p.toString());
			}
		    }
		    if (nearbyTSS.isNonCoding())
			categories2peaksMap.put("Coding,k4", categories2peaksMap.get("Coding,k4") + 1);
		    else 
			categories2peaksMap.put("Noncoding,k4", categories2peaksMap.get("Noncoding,k4") + 1);
		    categoryFound = true;
		}
		//Compare with macro or lncRNAs
		MacroRNA m = PeaksNearTranscripts.getNearestMacroRNA(macroRNAs, p, maxDist);
		if (m!=null && macroRNAsAccountedFor.contains(m)) {
		    nNonCodingReads = m.getFactorReads(rnaFileName).length>nNonCodingReads ? m.getFactorReads(rnaFileName).length : nNonCodingReads;
		    if (!categoryFound) {
			categories2peaksMap.put("Noncoding,k4", categories2peaksMap.get("Noncoding,k4") + 1);
			categoryFound = true;
		    }
		}
		LncRNA l = PeaksNearTranscripts.getNearestLncRNA(lncRNAs, p, maxDist);
		if (l!=null && lncRNAsAccountedFor.contains(l)) {
		    nNonCodingReads = l.getFactorReads(rnaFileName).length>nNonCodingReads ? l.getFactorReads(rnaFileName).length : nNonCodingReads;
		    if (!categoryFound) {
			categories2peaksMap.put("Noncoding,k4", categories2peaksMap.get("Noncoding,k4") + 1);
			categoryFound = true;
		    }
		}
		//Get the nearest annotated transcript and see if that includes more reads
		TranscribedRegion2 transcriptNearestStart = reducedModel.getNearestTranscript(p.getCenter(), true);
		if (transcriptNearestStart!=null && !transcriptsAccountedFor.contains(transcriptNearestStart) && ((forward && Math.abs(transcriptNearestStart.getStart() - p.getCenter())<maxDist) || (!forward && Math.abs(transcriptNearestStart.getEnd() - p.getCenter())<maxDist))) {
		    if (transcriptNearestStart.isCodingAnnotated()) {
			nCodingReads = nCodingReads<transcriptNearestStart.getFactorReads(rnaFileName).length ? transcriptNearestStart.getFactorReads(rnaFileName).length : nCodingReads;
			if (!categoryFound) 
			    categories2peaksMap.put("Coding,k4", categories2peaksMap.get("Coding,k4") + 1);
		    }
		    else if (transcriptNearestStart.isNonCodingAnnotated()) {
			nNonCodingReads = nNonCodingReads<transcriptNearestStart.getFactorReads(rnaFileName).length ? transcriptNearestStart.getFactorReads(rnaFileName).length : nNonCodingReads;
			System.out.println("Found " + nNonCodingReads + " from transcript " + transcriptNearestStart.toString() + " near " + p.toString());
			if (!categoryFound) 
			    categories2peaksMap.put("Noncoding,k4", categories2peaksMap.get("Noncoding,k4") + 1);
		    }
		    else {
			nOtherReads = transcriptNearestStart.getFactorReads(rnaFileName).length;
			if (!categoryFound) 
			    categories2peaksMap.put("Other,k4", categories2peaksMap.get("Other,k4") + 1);
		    }
		}
		//Find out which of the candidates to consider
		if (nCodingReads>=nOtherReads && nCodingReads>=nNonCodingReads) {
		    categories2readsMap.put("Coding,k4", categories2readsMap.get("Coding,k4") + nCodingReads);
		    genesAccountedFor.addAll(geneInfo.getOverlappingGenes(nearbyTSS, true));
		    if (transcriptNearestStart!=null) {	transcriptsAccountedFor.add(transcriptNearestStart); }
		}
		else if (nNonCodingReads>=nOtherReads && nNonCodingReads>=nCodingReads) {
		    categories2readsMap.put("Noncoding,k4", categories2readsMap.get("Noncoding,k4") + nNonCodingReads);
		    genesAccountedFor.addAll(geneInfo.getOverlappingGenes(nearbyTSS, false));
		    if (transcriptNearestStart!=null) {	transcriptsAccountedFor.add(transcriptNearestStart); }
		    if (m!=null) { macroRNAsAccountedFor.add(m); }
		    if (l!=null) { lncRNAsAccountedFor.add(l); }
		}
		else if (nOtherReads>=nCodingReads && nOtherReads>=nNonCodingReads) {
		    categories2readsMap.put("Other,k4", categories2readsMap.get("Other,k4") + nOtherReads);
		    transcriptsAccountedFor.add(transcriptNearestStart);
		}
		else
		    System.out.println("Error could not find category for " + transcriptNearestStart.toString());
		//System.out.println("Found " + nCodingReads + ", " + nNonCodingReads + " and " + nOtherReads + " near " + p.toString());
	    }
	}
	try {
	    BufferedWriter peakReadsBw = new BufferedWriter(new FileWriter(saveReadsFileName));
	    if (chr==1) {
		peakReadsBw.write("#chr");
		for (String key : categories2readsMap.keySet()) 
		    peakReadsBw.write("\t" + key + ",reads\t" + key + ",peaks");
		peakReadsBw.write("\n");
	    }
	    peakReadsBw.write((new Integer(chr)).toString());
	    for (String key : categories2readsMap.keySet()) 
		peakReadsBw.write("\t" + categories2readsMap.get(key) + "\t" + categories2peaksMap.get(key));
	    peakReadsBw.write("\n");
	    peakReadsBw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving peaks " + e.toString());
	}
	return categories2readsMap;
   }

    public static boolean overlapsAnnotatedElement(int chr, int start, int end, boolean forward, RefSeqUCSCEnsemblGenes geneInfo, ReducedModel2 reducedModel, TreeSet<MacroRNA> macroRNAs, TreeSet<LncRNA> lncRNAs) {
	//Returns true if the window is overlapping a gene, a transcribed region or a ncRNA.
	if (geneInfo.overlapsGene(start, end, forward))
	    return true;
	if (reducedModel.getOverlappingSegment(start, end)!=null) 
	    return true;
	for (MacroRNA m : macroRNAs) {
	    if (m.getChromosome()==chr && m.getStrand()==forward && m.overlaps(start, end))
		return true;
	}
	for (LncRNA m : lncRNAs) {
	    if (m.getChromosome()==chr && m.overlaps(start, end))
		return true;
	}
	return false;
    }

    public static ABIRead[] generateRandomReadsModel(int chr, TreeMap<Integer, Integer> unannotatedReadsHist, boolean forward, RefSeqUCSCEnsemblGenes geneInfo, ReducedModel2 reducedModel, TreeSet<MacroRNA> macroRNAs, TreeSet<LncRNA> lncRNAs, String species) {
	//Randomly assign as many reads as are present in unannotatedReadsHist and make sure that they are not overlapping any of the annotated regions.
	int nReads = 0;
	for (Integer k : unannotatedReadsHist.keySet())
	    nReads += unannotatedReadsHist.get(k)*k;
	ABIRead randReads[] = new ABIRead[nReads];
	System.out.println("Assigning " + nReads + " random reads.");
	for (int i=0; i<randReads.length; i++) {
	    int start;
	    do {
		start = random.nextInt(RefSeqUCSCEnsemblGenes.getChromosomeLength(chr, species) - 35);
	    }
	    while (overlapsAnnotatedElement(chr, start, start + 35, forward, geneInfo, reducedModel, macroRNAs, lncRNAs));
	    randReads[i] = new ABIRead(chr, forward, start);
	}
	return randReads;
    }

    public static TreeMap<Integer, Integer> getReadsHistogram(ABIRead reads[], int windowSize, int chr, boolean forward, RefSeqUCSCEnsemblGenes geneInfo, ReducedModel2 reducedModel, TreeSet<MacroRNA> macroRNAs, TreeSet<LncRNA> lncRNAs, TreeMap<Integer, Integer> bpsCovered, String species) { return getReadsHistogram(reads, windowSize, chr, forward, geneInfo, reducedModel, macroRNAs, lncRNAs, Integer.MAX_VALUE, null, bpsCovered, species, null, false, 0); }
	

    public static TreeMap<Integer, Integer> getReadsHistogram(ABIRead reads[], int windowSize, int chr, boolean forward, RefSeqUCSCEnsemblGenes geneInfo, ReducedModel2 reducedModel, TreeSet<MacroRNA> macroRNAs, TreeSet<LncRNA> lncRNAs, int minReads, String saveFileName, TreeMap<Integer, Integer> bpsCovered, String species, String saveDensFileName, boolean appendSaveDensFileName, double normFactor) {
	//If consecutive windows have more than minReads, we assign them to a transcript which we save in saveFileName
	TreeMap<Integer, Integer> readsHist = new TreeMap<Integer, Integer>();
	try {
	    BufferedWriter bwDens = saveDensFileName==null ? null : new BufferedWriter(new FileWriter(saveDensFileName, appendSaveDensFileName));
	    BufferedWriter bw = saveFileName==null ? null : new BufferedWriter(new FileWriter(saveFileName));
	    int windowStart = 0, transcriptStart = -1, nReadsInTranscript = 0;
	    while (windowStart + windowSize<RefSeqUCSCEnsemblGenes.getChromosomeLength(chr, species)) {
		if (!overlapsAnnotatedElement(chr, windowStart, windowStart + windowSize, forward, geneInfo, reducedModel, macroRNAs, lncRNAs)) {
		    ABIRead rs[] = getReadSubset(reads, chr, forward, windowStart, windowStart + windowSize);
		    if (!isArtefact(rs)) {
			int nReads = rs.length;
			if (nReads>minReads && transcriptStart==-1) {
			    transcriptStart = windowStart;
			    nReadsInTranscript += nReads;
			}
			else if (nReads<=minReads && transcriptStart!=-1 && bw!=null) {
			    //First check if this region is close to one of the other regions. If that is the case, we assume that it belongs to that region instead
			    if (!overlapsAnnotatedElement(chr, transcriptStart - 10000, windowStart + windowSize + 10000, forward, geneInfo, reducedModel, macroRNAs, lncRNAs)) {
				bw.write("chr" + chr + ":" + transcriptStart + "-" + windowStart + "\t" + (forward ? "" : "-") + format.format(nReadsInTranscript/*/(windowStart + windowSize - transcriptStart)*/) + "\n");
				if (bwDens!=null)
				    bwDens.write(format.format(((double)nReadsInTranscript/(windowStart - transcriptStart))*normFactor) + "\n");
			    }
			    bpsCovered.put(transcriptStart, windowStart);
			    transcriptStart = -1;
			    nReadsInTranscript = 0;
			}
			if (!readsHist.containsKey(nReads))
			    readsHist.put(nReads, 0);
			readsHist.put(nReads, readsHist.get(nReads) + 1);
		    }
		}
		windowStart += windowSize;
	    }
	    if (bw!=null) { bw.close(); }
	    if (bwDens!=null) { bwDens.close(); }
	    //for (Integer k : readsHist.keySet())
	    //System.out.println(k + "\t" + readsHist.get(k));
	}
	catch (IOException e) {
	    System.err.println("Error saving remaining transcripts " + e.toString());
	}
	return readsHist;
    }
	
    public static boolean readsOverlapping(ABIRead reads[]) {
	//Returns true if all of the reads are part of an overlapping cluster
	for (int i=1; i<reads.length; i++) {
	    if (!reads[i-1].overlaps(reads[i]))
		return false;
	}
	return true;
    }

    public static TreeMap<Integer, Integer> mergeCoverageMaps(TreeMap<Integer, Integer> bpsCoveredAnnotation, TreeMap<Integer, Integer> bpsCoveredTranscripts) {
	TreeMap<Integer, Integer> mergedCover = new TreeMap<Integer, Integer>();
	for (Integer s1 : bpsCoveredAnnotation.keySet()) {
	    int e1 = bpsCoveredAnnotation.get(s1), mergedStart = s1, mergedEnd = e1;
	    //Check if there are any other transcripts that overlap this one
	    for (Integer s2 : bpsCoveredTranscripts.keySet()) {
		int e2 = bpsCoveredTranscripts.get(s2);
		if (s1<=e2 && e1>=e2)
		    mergedStart = s2;
		if (s1<=s2 && e1>=s2)
		    mergedEnd = e2;
	    }
	    mergedCover.put(mergedStart, mergedEnd);
	}
	//Make sure that we have not created overlapping segments from the merging procedure
	Iterator<Integer> it = mergedCover.keySet().iterator();
	if (it.hasNext()) {
	    int prevStart = it.next();
	    int prevEnd = mergedCover.get(prevStart);
	    TreeSet<Integer> removeSet = new TreeSet<Integer>();
	    while (it.hasNext()) {
		int s = it.next(), e = mergedCover.get(s);
		if (prevStart<=s && prevEnd>=s) {
		    removeSet.add(prevStart);
		    mergedCover.put(prevStart, s-1);
		}
		prevStart = s;
		prevEnd = e;
	    }
	    for (Integer k : removeSet)
		mergedCover.remove(k);
	}
	return mergedCover;
    }
    
    public static int numberOfBpsCovered(TreeMap<Integer, Integer> bpsCoveredAnnotation, TreeMap<Integer, Integer> bpsCoveredTranscripts) {
	return numberOfBpsCovered(mergeCoverageMaps(bpsCoveredAnnotation, bpsCoveredTranscripts));
    }
	    
    public static int numberOfBpsCovered(TreeMap<Integer, Integer> mergedCover) {
	int nBpsCovered = 0;
	for (Integer s : mergedCover.keySet())
	    nBpsCovered += mergedCover.get(s) - s;
	return nBpsCovered;
    }

    public static int numberOfBpsTranscripts(TreeMap<Integer, Integer> bpsCoveredAnnotation, TreeMap<Integer, Integer> bpsCoveredTranscripts) {
	return mergeCoverageMaps(bpsCoveredAnnotation, bpsCoveredTranscripts).size();
    }

    public static int numberOfBpsTranscripts(TreeMap<Integer, Integer> bpsCoveredAnnotation) { return bpsCoveredAnnotation.size(); }

    public static int numberOfTranscriptsStartingConserved(TreeMap<Integer, Integer> bpsCoveredAnnotation, TreeMap<Integer, Integer> bpsCoveredTranscripts, int chr, boolean forward, ConservationIslands conservationIslands) {
	return numberOfTranscriptsStartingConserved(mergeCoverageMaps(bpsCoveredAnnotation, bpsCoveredTranscripts), chr, forward, conservationIslands);
    }

    public static int numberOfTranscriptsStartingConserved(TreeMap<Integer, Integer> bpsCovered, int chr, boolean forward, ConservationIslands conservationIslands) {
	int nCons = 0;
	for (Integer s : bpsCovered.keySet()) {
	    s = forward ? s : bpsCovered.get(s);
	    nCons += conservationIslands!=null && conservationIslands.overlaps(chr, s) ? 1 : 0;
	}
	return nCons;
    }

    ////////////////////////////////////////////

    

    ////////////////////////////////////////////

    public static int removeDuplicateConservedIslands(TreeSet<ConservationIsland> cis, TreeSet<ConservationIsland> conservationIslandsAlreadyExplained) {
	//Count the number of CIs in the set cis that are not found in
	//conservationIslandsAlreadyExplained and then add the
	//remaining elements to the
	//conservationIslandsAlreadyExplained set.
	int nNonOverlapping = 0;
	for (ConservationIsland c : cis) {
	    if (!conservationIslandsAlreadyExplained.contains(c)) {
		nNonOverlapping++;
		conservationIslandsAlreadyExplained.add(c);
	    }
	}
	return nNonOverlapping;
    }

}