//Use this class to find which windows are significant for a given
//pair of thresholds (one for #ChIP-#input and one restricting the
//maximum #input). This method utilizes the output from
//CalculateCriticalWindowDifference.java and takes the histograms of
//the distributions as well as the list of windows
//generated. Unfortunately, the analysis cannot be done in Matlab
//directly since we run out of memory because of Matlab's crappy
//method for handling it.
//
//Example:
//
//>java -Xmx512m FindCriticalWindows -n Npas4/KCl/Npas4_KCl -p .01
//



import jargs.gnu.CmdLineParser;
import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
//import umontreal.iro.lecuyer.stat.Tally;
import org.apache.commons.math.stat.descriptive.SummaryStatisticsImpl;


public class FindCriticalWindows {

    public static DecimalFormat format = new DecimalFormat("###.########");

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    //Use this flag to for the location of the .fa-files
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Decide what file to write the results to
	    CmdLineParser.Option saveFileFlag = argParser.addStringOption('f', "saveFile");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption('p', "fdr");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    String GENOME_LOC = "./";
	    int nChr = 21;
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(saveFileFlag);
	    String saveFileStart = tmp!=null && !tmp.toString().equals("") ? tmp.toString() : fileName;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 240;
	    //////////////////////////////////////////////////////////////////
	    //Parse the file with the cut-offs
	    WindowFinder winFinder = new WindowFinder(fileName, fdr);
	    Vector mergedWindows = new Vector();
	    int nWins = 0;
	    if (chr==-1) {
		for (int i=0; i<nChr; i++) {
		    int wins[][] = winFinder.findSignificantWindows(i+1, windowSize);
		    mergedWindows.addElement(wins);
		    nWins += wins.length;
		}
	    }
	    else {
		int wins[][] = winFinder.findSignificantWindows(chr, windowSize);
		nWins += wins.length;
	    }
	    //Count the total number of windows
	    System.out.println(nWins);
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }
}
