//Use this method to calculate the overlap btwn the peaks for a given TF from different cell-types
//
//for ((i=1; i<25; i++)) ; do bsub -q shared_1d -o /dev/null java -Xmx4096m -server FindEncodePeaksOverlap -o  -f dnase -p /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase1_names.txt -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase2_names.txt --nMC 0 -c $i ; done
//
//For mouse DHS data:
//
//for ((i=1; i<22; i++)) ; do bsub -q shared_12h -o /dev/null java -Xmx4096m -server FindEncodePeaksOverlap -o  -f dnase -p /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase_mm1_names.txt -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase_mm2_names.txt --species mm --nMC 0 -c $i ; done

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class FindEncodeDHSOverlap {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option nMCFlag = argParser.addIntegerOption("nMC");	    
	    CmdLineParser.Option factorNameFlag = argParser.addStringOption('f', "factorName");
	    CmdLineParser.Option overlapFactorNameFlag = argParser.addStringOption('F', "overlapFactorName");
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option overlapPeakFileNameFlag = argParser.addStringOption('P', "overlapPeakFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean replicate = argParser.getOptionValue(replicateFlag)!=null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? (String)tmp : "hs";
	    tmp = argParser.getOptionValue(factorNameFlag);
	    String factorName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(overlapFactorNameFlag);
	    String overlapFactorName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : (!orchestra ? "/home/mhemberg" : "/home/mh168") + "/Projects/ChIP-Seq/Encode/PeakFiles/dnase_names.txt";
	    tmp = argParser.getOptionValue(overlapPeakFileNameFlag);
	    String overlapPeakFileName = replicate ? null : (tmp!=null ? (String)tmp : null);
	    String replicatePeakFileName = replicate ? (String)tmp : null;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(nMCFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : 100;
	    tmp = argParser.getOptionValue(peakCategoryFlag);
	    String category = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "PeakLocations/" + (overlapPeakFileName!=null ? overlapPeakFileName.substring(overlapPeakFileName.lastIndexOf("/") + 1, overlapPeakFileName.lastIndexOf(".")) + "_" : "") + (peakFileName.contains("/") ? peakFileName.substring(peakFileName.lastIndexOf("/") + 1, peakFileName.lastIndexOf(".")) : peakFileName.substring(0, Math.max(peakFileName.length(), peakFileName.lastIndexOf(".")))) + "_unique_peaks" + (category!=null ? "_" + category : "") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    
	    ///////////////////////////////////////////////////////
	    int nChr = species.toLowerCase().equals("hs") ? 24 : 21;
	    System.out.println("Saving to " + saveFileName.replace("unique_peaks", "peaks_heatmap").replace("PeakLocations", "HeatMaps"));
	    BufferedWriter bwOverlapEnrichment = new BufferedWriter(new FileWriter(saveFileName==null ? "/dev/null" : saveFileName.replace("unique_peaks", "peaks_heatmap").replace("PeakLocations", "HeatMaps")));
	    System.out.println("Saving to " + saveFileName.replace("unique_peaks", "unique_peaks_pos"));
	    BufferedWriter bwOverlapPeakSizes = new BufferedWriter(new FileWriter(saveFileName==null ? "/dev/null" : saveFileName.replace("unique_peaks", "unique_peaks_pos")));

	    String cellTypes[] = EncodePeaks.getEncodeCellTypes(peakFileName, factorName);
	    String overlapCellTypes[] = EncodePeaks.getEncodeCellTypes(overlapPeakFileName, overlapFactorName);
	    if ((chr==-1 || chr==1)) {
		bwOverlapEnrichment.write("#chr\tPos");
		for (int i=0; i<overlapCellTypes.length; i++) {
		    bwOverlapEnrichment.write("\t" + overlapCellTypes[i]);
		}
		bwOverlapEnrichment.write("\n");
		bwOverlapPeakSizes.write("#chr\tStart\tEnd\n");
	    }
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    bw.write(Integer.toString(i));
		    RefSeqUCSCEnsemblGenes geneInfo = new RefSeqUCSCEnsemblGenes(i, orchestra, species);
		    EncodePeaks overlapPeaks = new EncodePeaks(overlapPeakFileName, overlapFactorName, i);
		    System.out.println("Found " + overlapPeaks.numberOfPeaks() + " peaks from " + overlapPeakFileName);
		    EncodePeaks encodePeaks = new EncodePeaks(peakFileName, replicatePeakFileName, factorName, i);
		    System.out.println("Found " + encodePeaks.numberOfPeaks() + " peaks from " + peakFileName);

		    int nUniquePeaks[][] = overlapPeaks.numberOfUniquePeaks(encodePeaks, nMC);
		    
		    //TreeSet<Peak> overlappingPeaks = overlapPeaks.findOverlappingEnrichment(encodePeaks);
		    TreeSet<Peak> overlappingPeaks = overlapPeaks!=null ? encodePeaks.findOverlappingEnrichment(overlapPeaks) : encodePeaks.getPeaks();

		    encodePeaks = category==null ? (replicate ? new EncodePeaks(peakFileName, replicatePeakFileName, factorName, i, conservationIslands) : new EncodePeaks(peakFileName, factorName, i, conservationIslands)) : new EncodePeaks(peakFileName, category, i, false);
		    if (nMC>0) {
			for (int j=0; j<nUniquePeaks.length; j++)
			    bw.write("\t" + nUniquePeaks[j][0] + "\t" + format.format(nUniquePeaks[j][1]/Math.sqrt(nMC)) + "\t" + nUniquePeaksCons[j][0] + "\t" + format.format(nUniquePeaksCons[j][1]/Math.sqrt(nMC)));
			bw.write("\n");
		    }
		    for (Peak p : overlappingPeaks) {
			bwOverlapEnrichment.write(p.getChromosome() + "\t" + p.getCenter());
			for (int j=0; j<overlapCellTypes.length; j++)
			    bwOverlapEnrichment.write("\t" + (p.hasFactor(overlapCellTypes[j]) ? p.getFactor(overlapCellTypes[j]) : 0));
			bwOverlapEnrichment.write("\n");
			bwOverlapPeakSizes.write(p.getChromosome() + "\t" + p.getStart() + "\t" + p.getEnd() + "\n");
		    }
		}
	    }
	    bw.close();
	    bwOverlapEnrichment.close();
	    bwOverlapEnrichmentCons.close();
	    bwOverlapPeakSizes.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////


}


