//Use this method to calculate the overlap btwn the peaks for a given TF from different cell-types
//
//bsub -q shared_2h java -Xmx4096m -server FindEncodePeaksOverlap -o -c 22 -f dnase
//
//java -Xmx4096m -server FindEncodePeaksOverlap -o -c 22 -f Ctcf --nMC 10 -p ../../Encode/PeakFiles/ctcf_names.txt 
//
//java -Xmx4096m -server FindEncodePeaksOverlap -o -c 22 -f H3k4me3 --nMC 10 -p ../../Encode/PeakFiles/h3k4me3_names.txt 
//
//For looking at the overlap btwn h3k4me3 in HeLa and DHS sites in other cell types
//
//bsub -q shared_2h java -Xmx4096m -server FindEncodePeaksOverlap -o -c 22 -f dnase -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak
//
//When looking at only a specific category of HeLa DHS and h3k4me3 in other cell types
//
//java -Xmx4096m -server FindEncodePeaksOverlap -o -c 22 -f H3k4me3 -p ../../Encode/PeakFiles/h3k4me3_names.txt --nMC 100 -P hela_dnase -C Extragenic_other
//
//If we are interested in only two factors from one cell type
//
//bsub -q shared_2h java -Xmx4096m -server FindEncodePeaksOverlap -o -c 22 -f dnase -p /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwDnaseSeqPeaksRep1Helas3.narrowPeak -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak
//
//For finding DHS sites that are present in both replicates:
//
//bsub -q shared_12h java -Xmx4096m -server FindEncodePeaksOverlap -o  -f dnase -p /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase1_names.txt -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase2_names.txt --replicate
//
//for ((i=1; i<25; i++)) ; do bsub -q shared_1d -o /dev/null java -Xmx4096m -server FindEncodePeaksOverlap -o  -f dnase -p /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase1_names.txt -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase2_names.txt --nMC 0 --replicate -c $i ; done
//
//For mouse DHS data:
//
//for ((i=1; i<22; i++)) ; do bsub -q shared_12h -o /dev/null java -Xmx4096m -server FindEncodePeaksOverlap -o  -f dnase -p /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase_mm1_names.txt -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/dnase_mm2_names.txt --species mm --nMC 0 --replicate -c $i ; done

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class FindEncodePeaksOverlap {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option replicateFlag = argParser.addBooleanOption("replicate");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option nMCFlag = argParser.addIntegerOption("nMC");	    
	    CmdLineParser.Option factorNameFlag = argParser.addStringOption('f', "factorName");
	    CmdLineParser.Option overlapFactorNameFlag = argParser.addStringOption('F', "overlapFactorName");
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option overlapPeakFileNameFlag = argParser.addStringOption('P', "overlapPeakFileName");
	    //Only consider peaks that fall into a given category
	    CmdLineParser.Option peakCategoryFlag = argParser.addStringOption('C', "category");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean replicate = argParser.getOptionValue(replicateFlag)!=null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? (String)tmp : "hs";
	    tmp = argParser.getOptionValue(factorNameFlag);
	    String factorName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(overlapFactorNameFlag);
	    String overlapFactorName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : (!orchestra ? "/home/mhemberg" : "/home/mh168") + "/Projects/ChIP-Seq/Encode/PeakFiles/dnase_names.txt";
	    tmp = argParser.getOptionValue(overlapPeakFileNameFlag);
	    String overlapPeakFileName = replicate ? null : (tmp!=null ? (String)tmp : null);
	    String replicatePeakFileName = replicate ? (String)tmp : null;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(nMCFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : 100;
	    tmp = argParser.getOptionValue(peakCategoryFlag);
	    String category = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "PeakLocations/" + (overlapPeakFileName!=null ? overlapPeakFileName.substring(overlapPeakFileName.lastIndexOf("/") + 1, overlapPeakFileName.lastIndexOf(".")) + "_" : "") + (peakFileName.contains("/") ? peakFileName.substring(peakFileName.lastIndexOf("/") + 1, peakFileName.lastIndexOf(".")) : peakFileName.substring(0, Math.max(peakFileName.length(), peakFileName.lastIndexOf(".")))) + "_unique_peaks" + (category!=null ? "_" + category : "") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    
	    ///////////////////////////////////////////////////////
	    boolean removePromoters = false, removeIntergenic = false, removeEnhancerOverlappingPeaks = false;
	    int nChr = species.toLowerCase().equals("hs") ? 24 : 21, maxTSSDist = 1000, maxEnhancerDist = 1000, maxPeakDist = 100;
	    System.out.println("Saving to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(nMC==0 || saveFileName==null ? "/dev/null" : saveFileName));
	    System.out.println("Saving to " + saveFileName.replace("unique_peaks", "peaks_heatmap").replace("PeakLocations", "HeatMaps"));
	    BufferedWriter bwOverlapEnrichment = new BufferedWriter(new FileWriter(saveFileName==null ? "/dev/null" : saveFileName.replace("unique_peaks", "peaks_heatmap").replace("PeakLocations", "HeatMaps")));
	    BufferedWriter bwOverlapEnrichmentCons = new BufferedWriter(new FileWriter(saveFileName==null ? "/dev/null" : saveFileName.replace("unique_peaks", "peaks_heatmap_cons").replace("PeakLocations", "HeatMaps")));
	    System.out.println("Saving to " + saveFileName.replace("unique_peaks", "unique_peaks_pos"));
	    BufferedWriter bwOverlapPeakSizes = new BufferedWriter(new FileWriter(saveFileName==null ? "/dev/null" : saveFileName.replace("unique_peaks", "unique_peaks_pos")));

	    String cellTypes[] = EncodePeaks.getEncodeCellTypes(peakFileName, factorName);
	    String overlapCellTypes[] = overlapPeakFileName!=null ? EncodePeaks.getEncodeCellTypes(overlapPeakFileName, overlapFactorName) : cellTypes;
	    if ((chr==-1 || chr==1)) {
		bw.write("#chr");
		for (int i=1; i<=cellTypes.length; i++) 
		    bw.write("\tMean" + i + "\tStdErr" + i + "\tConsMean" + i + "\tConsStdErr" + i);
		bw.write("\n");

		bwOverlapEnrichment.write("#chr\tPos");
		bwOverlapEnrichmentCons.write("#chr\tPos");
		for (int i=0; i<overlapCellTypes.length; i++) {
		    bwOverlapEnrichment.write("\t" + overlapCellTypes[i]);
		    bwOverlapEnrichmentCons.write("\t" + overlapCellTypes[i]);
		}
		bwOverlapEnrichment.write("\n");
		bwOverlapEnrichmentCons.write("\n");

		bwOverlapPeakSizes.write("#chr\tStart\tEnd\n");
	    }
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    bw.write(Integer.toString(i));
		    EncodePeaks overlapPeaks = null;
		    RefSeqUCSCEnsemblGenes geneInfo = new RefSeqUCSCEnsemblGenes(i, orchestra, species);
		    if (overlapPeakFileName!=null) { //Get a set of peaks and consider how the peaks in peakFileName overlaps this set
			overlapPeaks = new EncodePeaks(overlapPeakFileName, overlapFactorName, i);
			System.out.println("Found " + overlapPeaks.numberOfPeaks() + " peaks from " + overlapPeakFileName);
			if (removePromoters || removeIntergenic) {
			    if (removePromoters) {
				overlapPeaks.removePromoterProximalPeaks(geneInfo, true, maxTSSDist);
				overlapPeaks.removePromoterProximalPeaks(geneInfo, false, maxTSSDist);
			    }
			    System.out.println("Found " + overlapPeaks.numberOfPeaks() + " after removing the ones overlapping promoters.");
			    if (removeIntergenic)
				overlapPeaks.removeIntergenicPeaks(geneInfo);
			}
			if (removeEnhancerOverlappingPeaks) {
			    String extragenicEnhancerFileName = "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed";
			    TreeSet<Enhancer2> enhancers = FindTranscripts2.parseEnhancerFile(extragenicEnhancerFileName);
			    String intragenicEnhancerFileName = "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed";
			    enhancers.addAll(FindTranscripts2.parseEnhancerFile(intragenicEnhancerFileName));
			    overlapPeaks.removeEnhancerOverlappingPeaks(enhancers, maxEnhancerDist);
			}
		    }
		    EncodePeaks encodePeaks = category==null ? (replicate ? new EncodePeaks(peakFileName, replicatePeakFileName, factorName, i) : new EncodePeaks(peakFileName, factorName, i)) : new EncodePeaks(peakFileName, category, i, false);
		    System.out.println("Found " + encodePeaks.numberOfPeaks() + " peaks from " + peakFileName);
		    if (removePromoters || removeIntergenic) {
			if (removePromoters) {
			    encodePeaks.removePromoterProximalPeaks(geneInfo, true, maxTSSDist);
			    encodePeaks.removePromoterProximalPeaks(geneInfo, false, maxTSSDist);
			}
			System.out.println("Found " + encodePeaks.numberOfPeaks() + " after removing the ones overlapping promoters.");
			if (removeIntergenic)
			    encodePeaks.removeIntergenicPeaks(geneInfo);
			/*String extragenicEnhancerFileName = "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed";
			TreeSet<Enhancer2> enhancers = FindTranscripts2.parseEnhancerFile(extragenicEnhancerFileName);
			String intragenicEnhancerFileName = "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed";
			enhancers.addAll(FindTranscripts2.parseEnhancerFile(intragenicEnhancerFileName));
			overlapPeaks.removeEnhancerOverlappingPeaks(enhancers, maxEnhancerDist);*/
			System.out.println("Found " + encodePeaks.numberOfPeaks() + " after removing the ones overlapping enhancers.");
			//Just look at the overlap for the encodePeaks and the overlapPeaks
			System.out.println("There are " + encodePeaks.numberOfPeaks(null, overlapPeaks) + " overlapping peaks.");
		    }
		    //int nUniquePeaks[][] = encodePeaks.numberOfUniquePeaks(overlapPeaks, nMC);
		    int nUniquePeaks[][] = overlapPeaks!=null ? overlapPeaks.numberOfUniquePeaks(encodePeaks, nMC) : encodePeaks.numberOfUniquePeaks(nMC);
		    
		    //TreeSet<Peak> overlappingPeaks = overlapPeaks.findOverlappingEnrichment(encodePeaks);
		    TreeSet<Peak> overlappingPeaks = overlapPeaks!=null ? encodePeaks.findOverlappingEnrichment(overlapPeaks) : encodePeaks.getPeaks();

		    ConservationIslands conservationIslands = category==null ? new ConservationIslands(i, orchestra, species) : null;
		    encodePeaks = category==null ? (replicate ? new EncodePeaks(peakFileName, replicatePeakFileName, factorName, i, conservationIslands) : new EncodePeaks(peakFileName, factorName, i, conservationIslands)) : new EncodePeaks(peakFileName, category, i, false);
		    //int nUniquePeaksCons[][] = encodePeaks.numberOfUniquePeaks(overlapPeaks, nMC);
		    int nUniquePeaksCons[][] = overlapPeaks!=null ? overlapPeaks.numberOfUniquePeaks(encodePeaks, nMC) : encodePeaks.numberOfUniquePeaks(nMC);
		    TreeSet<Peak> overlappingPeaksCons = overlapPeaks!=null ? overlapPeaks.findOverlappingEnrichment(encodePeaks) : encodePeaks.getPeaks();
		    if (nMC>0) {
			for (int j=0; j<nUniquePeaks.length; j++)
			    bw.write("\t" + nUniquePeaks[j][0] + "\t" + format.format(nUniquePeaks[j][1]/Math.sqrt(nMC)) + "\t" + nUniquePeaksCons[j][0] + "\t" + format.format(nUniquePeaksCons[j][1]/Math.sqrt(nMC)));
			bw.write("\n");
		    }
		    for (Peak p : overlappingPeaks) {
			//if (overlapPeaks!=null || conservationIslands.overlaps(p)) {
			bwOverlapEnrichment.write(p.getChromosome() + "\t" + p.getCenter());
			for (int j=0; j<overlapCellTypes.length; j++)
			    bwOverlapEnrichment.write("\t" + (p.hasFactor(overlapCellTypes[j]) ? p.getFactor(overlapCellTypes[j]) : 0));
			bwOverlapEnrichment.write("\n");
			//}
			bwOverlapPeakSizes.write(p.getChromosome() + "\t" + p.getStart() + "\t" + p.getEnd() + "\n");
		    }
		    for (Peak p : overlappingPeaksCons) {
			bwOverlapEnrichmentCons.write(p.getChromosome() + "\t" + p.getCenter());
			for (int j=0; j<overlapCellTypes.length; j++)
			    bwOverlapEnrichmentCons.write("\t" + (p.hasFactor(overlapCellTypes[j]) ? p.getFactor(overlapCellTypes[j]) : 0));
			bwOverlapEnrichmentCons.write("\n");
		    }
		}
	    }
	    bw.close();
	    bwOverlapEnrichment.close();
	    bwOverlapEnrichmentCons.close();
	    bwOverlapPeakSizes.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////


}


