//
//
//java -Xmx512m FindEnrichedWords -f Sequences/ap1_fos_npas4_cbp_peaks_extragenic_distal_ids_w_100  -c Sequences/ap1_fos_peaks_extragenic_distal_ids_w_100 -m 6 -M 7 -p .0001

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;


public class FindEnrichedWords {



    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option motifFlag = argParser.addBooleanOption("motifs");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option seqFileNameFlag = argParser.addStringOption('f', "seqFileName");
	    CmdLineParser.Option ctrlSeqFileNameFlag = argParser.addStringOption('c', "ctrlSeqFileName");
	    CmdLineParser.Option pValFlag = argParser.addDoubleOption('p', "pVal");
	    CmdLineParser.Option minLenFlag = argParser.addIntegerOption('m', "minLen");
	    CmdLineParser.Option maxLenFlag = argParser.addIntegerOption('M', "maxLen");
	    CmdLineParser.Option nClustersFlag = argParser.addIntegerOption('k', "nClusters");
	    CmdLineParser.Option peakWidthFlag = argParser.addIntegerOption('w', "width");
	    CmdLineParser.Option subsampleFlag = argParser.addIntegerOption('S', "subsample");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    boolean motifs = argParser.getOptionValue(motifFlag)!=null;
	    tmp = argParser.getOptionValue(pValFlag);
	    double pVal = tmp!=null ? ((Double)tmp).doubleValue() : .00001;
	    tmp = argParser.getOptionValue(minLenFlag);
	    int minLen = tmp!=null ? ((Integer)tmp).intValue() : 6;
	    tmp = argParser.getOptionValue(maxLenFlag);
	    int maxLen = tmp!=null ? ((Integer)tmp).intValue() : 6;
	    tmp = argParser.getOptionValue(nClustersFlag);
	    int nClusters = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(peakWidthFlag);
	    int peakWidth = tmp!=null ? ((Integer)tmp).intValue() : 100;
	    tmp = argParser.getOptionValue(subsampleFlag);
	    int subsample = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(seqFileNameFlag);
	    String seqFileName = tmp!=null ? (String)tmp : "Sequences/ap1_fos_npas4_cbp_peaks_distal_w_" + peakWidth;
	    tmp = argParser.getOptionValue(ctrlSeqFileNameFlag);
	    String ctrlSeqFileName = tmp!=null ? (String)tmp : "Sequences/enhancers_control_w_" + peakWidth;
	    //////////////////////////////////////////////////////////////////
	    EnrichedWordFinder wf = new EnrichedWordFinder(seqFileName, ctrlSeqFileName, subsample, orchestra);
	    if (motifs) {
		Vector<Vector<String>> clusters = wf.findEnrichedClusters(minLen, maxLen, pVal, nClusters);
		int k = 0;
		for (Vector<String> c : clusters) 
		    System.out.println(format.format((double)wf.numberOfClusterOccurences(c)/wf.numberOfSequences()) + " ( " + format.format((double)wf.numberOfClusterOccurences(c, true)/wf.numberOfSequences(true)) + " ) " + c.toString());
		Vector<int[][]> ms = wf.findEnrichedMotifs(clusters);
		for (int[][] m : ms) {
		    System.out.println(">motif" + (k++));
		    for (int i=0; i<m.length; i++) {
			for (int j=0; j<m[i].length; j++) 
			    System.out.print(m[i][j] + " ");
			System.out.print("\n");
		    }
		}
	    }
	    else {
		Vector<Vector<String>> clusters = wf.findEnrichedClusters(minLen, maxLen, pVal, nClusters);
		int i = 0;
		for (Vector<String> c : clusters) {
		    System.out.println(format.format((double)wf.numberOfClusterOccurences(c)/wf.numberOfSequences()) + " ( " + format.format((double)wf.numberOfClusterOccurences(c, true)/wf.numberOfSequences(true)) + " ) ");
		    for (String w : c)
			System.out.println(">word_" + (i++) + "\n" + w);
		}
	    }
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////


}





