//
//
//bsub -q all_2h java -Xmx1024m FindMARs --species mm
//
//for ((i=1; i<=21; i++)) ; do bsub -q shared_2h java -Xmx4096m FindMARs --species mm -o -r -c $i; done
//

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;


public class FindMARs {

    public static DecimalFormat format = new DecimalFormat("###.####");

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option repeatMaskFlag = argParser.addBooleanOption('r', "repeatMask");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveDirNameFlag = argParser.addStringOption('d', "saveDirName");
	    
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean repeatMask = argParser.getOptionValue(repeatMaskFlag)!=null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    int nChr = species.equals("mm") ? 21 : 24, marLen = 20, nMARs = 0, nCons = 0, nUCIMARs = 0;
	    //////////////////////////////////////////////////////////////////
	    String seq, hRule = "(A){5}", hRuleRC = "([A|G|T]{5,})"; //"HHHHHHHHHHHHHHHHHHHH";
	    String saveFileName = "MARs/" + species + (repeatMask ? "_repeatMask" : "") + "_MARs_Hrule" + (chr==-1 ? "" : "_" + chr) + ".txt";
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    //Open the sequence
	    for (int k=1; k<=nChr; k++) {
		if (k==chr || chr==-1) {
		    //Load the conserved islands
		    //ConservationIslands cis = new ConservationIslands(k, orchestra, species);
		    //Load the entire genome
		    FastaSequence2 fastaReader = new FastaSequence2(species, false, k, orchestra);
		    long len = fastaReader.getChromosomeLength();
		    fastaReader = new FastaSequence2(species, false, k, orchestra);
		    //Find the locations of the H-matches. Merge overlapping regions
		    Vector<Integer> starts = new Vector<Integer>(), ends = new Vector<Integer>(), types = new Vector<Integer>();
		    Vector<Boolean> conserved = new Vector<Boolean>();
		    seq = repeatMask ? fastaReader.getSequence(0, marLen) : fastaReader.getSequence(0, marLen).toUpperCase();
		    char s[] = seq.toCharArray();
		    int nHF = 0, nHR = 0;
		    for (int j=0; j<s.length; j++) {
			nHF += s[j]=='A' || s[j]=='C' || s[j]=='T' ? 1 : 0;
			nHR += s[j]=='A' || s[j]=='G' || s[j]=='T' ? 1 : 0;
		    }
		    for (int i=marLen + 1; i<len; i++) {
			//subtract for the base that gets removed
			nHF -= s[0]=='A' || s[0]=='C' || s[0]=='T' ? 1 : 0;
			nHR -= s[0]=='A' || s[0]=='G' || s[0]=='T' ? 1 : 0;
			// Shift all elements left by one
			System.arraycopy(s, 1, s, 0, s.length-1);
			//Read the new element
			s[s.length-1] = repeatMask ? fastaReader.getNextBase().toCharArray()[0] : fastaReader.getNextBase().toUpperCase().toCharArray()[0];
			nHF += s[s.length-1]=='A' || s[s.length-1]=='C' || s[s.length-1]=='T' ? 1 : 0;
			nHR += s[s.length-1]=='A' || s[s.length-1]=='C' || s[s.length-1]=='T' ? 1 : 0;
			if (nHF==marLen || nHR==marLen) {
			    nMARs++;
			    if (!starts.isEmpty() && i>starts.lastElement() && i<ends.lastElement()) //overlap
				ends.setElementAt(i + marLen, ends.size() - 1);
			    else {
				//System.out.println("Found MAR " + new String(s) + " at " + i);
				starts.addElement(i);
				ends.addElement(i + marLen);
				//Check if there is overlap with a conservation island
				/*TreeSet<ConservationIsland> over = cis.getOverlappingConservationIslands(k, starts.lastElement(), ends.lastElement());
				if (over!=null && over.size()>0) {
				    conserved.addElement(new Boolean(true));
				    nCons++;
				    if (over.size()>1) { //Take the one that has the largest overlap
					//System.out.println(over.size() + " islands for " + starts.lastElement() + "-" + ends.lastElement());
					int largestOverlap = -1;
					ConservationIsland c = null;
					for (ConservationIsland ci : over) {
					    int overlap = Math.min(ends.lastElement(), ci.getEnd()) - Math.max(starts.lastElement(), ci.getStart());
					    if (overlap>largestOverlap) {
						c = ci;
						largestOverlap = overlap;
						if (c.getType()==8 || c.getType()==9) {
						    nUCIMARs++;
						    break; //Found a uncharacterized CI
						}
					    }
					}
					types.addElement(c.getType());
				    }
				    else {
					types.addElement(over.first().getType());
					if (types.lastElement()==8 || types.lastElement()==9) 
					    nUCIMARs++;
				    }
				}
				else {*/
				    conserved.addElement(new Boolean(false));
				    types.addElement(-1);
				    //}
			    }
			}
			//if (i % 1000000==0)
			//  System.out.println(i);
		    }
		    for (int j=0; j<starts.size(); j++)
			bw.write(k + "\t" + starts.elementAt(j) + "\t" + ends.elementAt(j) + "\t" + types.elementAt(j) + "\n");
		}	   
	    }
	    bw.close();
	    System.out.println("Found " + nMARs + " MARs with conserved " + nCons + " and " + nUCIMARs + " overlapping uncharacterized CIs.");
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
    }


}