//Use this method to calculate the distance distributions for various
//combinations of motifs.
//
//java -Xmx1024m FindMotifDistances -f Sequences/ap1_fos_peaks_extragenic_distal_ids_w_100 -M 1

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.Tally;

public class FindMotifDistances {

    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option nearestOnlyFlag = argParser.addBooleanOption('n', "nearestOnly");
	    CmdLineParser.Option absoluteFlag = argParser.addBooleanOption('a', "absolute");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option seqFileNameFlag = argParser.addStringOption('f', "seqFileName");
	    CmdLineParser.Option motifFileNameFlag = argParser.addStringOption('m', "motifFileName");
	    CmdLineParser.Option pValFlag = argParser.addDoubleOption('p', "pVal");
	    CmdLineParser.Option subsampleFlag = argParser.addIntegerOption('S', "subsample");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('M', "mismatches");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    boolean nearestOnly = argParser.getOptionValue(nearestOnlyFlag)!=null;
	    boolean absolute = argParser.getOptionValue(absoluteFlag)!=null;
	    tmp = argParser.getOptionValue(pValFlag);
	    double pVal = tmp!=null ? ((Double)tmp).doubleValue() : .00001;
	    tmp = argParser.getOptionValue(subsampleFlag);
	    int subsample = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    int mismatches = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(seqFileNameFlag);
	    String seqFileName = tmp!=null ? (String)tmp : "Sequences/ap1_fos_npas4_cbp_peaks_extragenic_distal_ids_w_100";
	    tmp = argParser.getOptionValue(motifFileNameFlag);
	    String motifFileName = tmp!=null ? (String)tmp : (orchestra ? "/home/mh168/" : "/home/mhemberg/") + "Projects/ChIP-Seq/TKData/Athar/ebox_motifs.txt";
	    //////////////////////////////////////////////////////////////////
	    Vector<String> seqs, motifs = new Vector<String>();
	    //Load the sequences
	    if (seqFileName.endsWith(".fa"))
		seqs = loadSequence(seqFileName);
	    else
		seqs = loadSequences(seqFileName);
	    //Load the motifs/words
	    if (motifFileName.endsWith(".txt"))
		motifs = loadMotifs(motifFileName);
	    String saveFileName = (seqFileName.endsWith(".fa") ? seqFileName.replace(".fa", "") : seqFileName) + (nearestOnly ? "" : "_all") + (mismatches>0 ? "_M_" + mismatches : "");
	    //For saving the distance to the nearest motif
	    BufferedWriter bwMotifDist = new BufferedWriter(new FileWriter((saveFileName + "_" + motifFileName.substring(motifFileName.lastIndexOf("/") + 1)).replace(".txt", "_dists.txt")));
	    //Write the header
	    //bwMotifDist.write("#pos");
	    for (String m1 : motifs) {
		if (!m1.matches("^NN*N$"))
		    bwMotifDist.write(m1 + "\t");
	    }
	    bwMotifDist.write("\n");
	    //Find the distribution of distances, Motif->Motif->Dist->#occurrences
	    TreeMap<String, TreeMap<String, TreeMap<Integer, Integer>>> dists = new TreeMap<String, TreeMap<String, TreeMap<Integer, Integer>>>();
	    TreeSet<Integer> allKeys = new TreeSet<Integer>();
	    for (String seq : seqs) {
		for (String m1 : motifs) {
		    if (!dists.containsKey(m1))
			dists.put(m1, new TreeMap<String, TreeMap<Integer, Integer>>());
		    Vector<Integer> p1 = findMotifs(seq, m1, mismatches);
		    for (String m2 : motifs) {
			if (!m1.equals(m2)) {
			    boolean foundDist = false;
			    if (!dists.get(m1).containsKey(m2))
				dists.get(m1).put(m2, new TreeMap<Integer, Integer>());
			    if (p1!=null && p1.size()>0) {
				
				Vector<Integer> p2 = findMotifs(seq, m2, mismatches);
				if (p2!=null && p2.size()>0) {
				    int minDist = Integer.MAX_VALUE, minD = Integer.MAX_VALUE;
				    for (Integer q1 : p1) {
					for (Integer q2 : p2) {
					    int d = absolute ? Math.abs(q1 - q2) : q1 - q2;
					    if (!dists.get(m1).get(m2).containsKey(d))
						dists.get(m1).get(m2).put(d, 0);
					    if (!nearestOnly) {
						dists.get(m1).get(m2).put(d, dists.get(m1).get(m2).get(d) + 1);
						allKeys.add(d);
					    }
					    minDist = Math.min(d, minDist);
					    minD = Math.min(Math.abs(d), minD);
					}
				    }
				    if (nearestOnly) {
					dists.get(m1).get(m2).put(minDist, dists.get(m1).get(m2).get(minDist) + 1);
					allKeys.add(minDist);
				    }
				    if (m1.matches("^NN*N$")) {
					bwMotifDist.write(minD + "\t");
					foundDist = true;
				    }
				}
			    }
			    if (m1.matches("^NN*N$") && !foundDist)
				bwMotifDist.write("-1\t");
			}
		    }
		    if (m1.matches("^NN*N$"))
			bwMotifDist.write("\n");
		}
	    }
	    bwMotifDist.close();
	    String fName = saveFileName + "_" + motifFileName.substring(motifFileName.lastIndexOf("/") + 1);
	    System.out.println("Saving to " + fName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fName));
	    bw.write("#dist");
	    for (String m1 : motifs) {
		for (String m2 : motifs) {
		    if (!m1.equals(m2)) 
			bw.write("\t" + m1 + "-" + m2);
		}
	    }
	    bw.write("\n");
	    TreeMap<String, TreeMap<String, Integer>> nFound = new TreeMap<String, TreeMap<String, Integer>>();
	    for (Integer d : allKeys) {
		bw.write(d.toString());
		for (String m1 : motifs) {
		    if (!nFound.containsKey(m1))
			nFound.put(m1, new TreeMap<String, Integer>());
		    for (String m2 : motifs) {
			if (!m1.equals(m2)) {
			    if (!nFound.get(m1).containsKey(m2))
				nFound.get(m1).put(m2, 0);
			    nFound.get(m1).put(m2, nFound.get(m1).get(m2) + (dists.get(m1).get(m2).containsKey(d) ? dists.get(m1).get(m2).get(d) : 0));
			    bw.write("\t" + (dists.get(m1).get(m2).containsKey(d) ? dists.get(m1).get(m2).get(d) : "0"));
			}
		    }
		}
		bw.write("\n");
	    }
	    //The number of sequences where no motif was found
	    bw.write("Inf");
	    for (String m1 : motifs) {
		for (String m2 : motifs) {
		    if (!m1.equals(m2)) 
			bw.write("\t" + (seqs.size() - nFound.get(m1).get(m2)));
		}
	    }
	    bw.write("\n");
	    bw.close();
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static Vector<String> loadMotifs(String fileName) {
	Vector<String> ms = new Vector<String>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    while (line!=null) {
		if (!line.startsWith("#")) {
		    String ws[] = line.split("\\s+");
		    for (int i=0; i<ws.length; i++)
			ms.add(ws[i]);
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.out.println("Error parsing motifs " + e.toString());
	}
	System.out.println("Found " + ms.size() + " motifs.");
	return ms;
    }
    
    public static Vector<String> getIUPACWords(String m, int mismatches) {
	Vector<String> ms = new Vector<String>();
	if (m.matches("^N+$")) //Used to mask out AP1
	    ms.addElement(m);
	else 
	    getIUPACWords(m.toCharArray(), 0, mismatches, ms);
	return ms;
    }

    protected static void getIUPACWords(char[] w, int i, Vector<String> ws) { getIUPACWords(w, i, 0, ws); }

    protected static void getIUPACWords(char[] w, int i, int mismatches, Vector<String> ws) {
	//Recursively generate all words
	if (i>=w.length) 
	    ws.addElement(new String(w));
	else if (w[i]=='R') {
	    char w1[] = new char[w.length];
	    System.arraycopy(w, 0, w1, 0, w.length);
	    w1[i] = 'A';
	    getIUPACWords(w1, i+1, ws); 
	    System.arraycopy(w, 0, w1, 0, w.length);
	    w1[i] = 'G';
	    getIUPACWords(w1, i+1, ws);
	}
	else if (w[i]=='Y') {
	    char w1[] = new char[w.length];
	    System.arraycopy(w, 0, w1, 0, w.length);
	    w1[i] = 'C';
	    getIUPACWords(w1, i+1, ws); 
	    System.arraycopy(w, 0, w1, 0, w.length);
	    w1[i] = 'T';
	    getIUPACWords(w1, i+1, ws);
	}
	else if (w[i]=='S') {
	    char w1[] = new char[w.length];
	    System.arraycopy(w, 0, w1, 0, w.length);
	    w1[i] = 'C';
	    getIUPACWords(w1, i+1, ws); 
	    System.arraycopy(w, 0, w1, 0, w.length);
	    w1[i] = 'G';
	    getIUPACWords(w1, i+1, ws);
	}
	else if (w[i]=='W') {
	    char w1[] = new char[w.length];
	    System.arraycopy(w, 0, w1, 0, w.length);
	    w1[i] = 'A';
	    getIUPACWords(w1, i+1, ws); 
	    System.arraycopy(w, 0, w1, 0, w.length);
	    w1[i] = 'T';
	    getIUPACWords(w1, i+1, ws);
	}
	else if (w[i]=='N') {
	    w[i] = '.';
	    getIUPACWords(w, i+1, ws);
	}
	else if (mismatches>0 && w.length>5) {
	    char w1[] = new char[w.length];
	    System.arraycopy(w, 0, w1, 0, w.length);
	    getIUPACWords(w1, i+1, mismatches, ws); 
	    System.arraycopy(w, 0, w1, 0, w.length);
	    w1[i] = '.';
	    getIUPACWords(w1, i+1, mismatches-1, ws);
	}
	else
	    getIUPACWords(w, i+1, ws);
    }

    public static Vector<Integer> findMotifs(String seq, String m) { return findMotifs(seq, m, 0, true); }

    public static Vector<Integer> findMotifs(String seq, String m, int mismatches) { return findMotifs(seq, m, mismatches, true); }

    public static Vector<Integer> findMotifs(String seq, String motif, int mismatches, boolean revcomp) {
	Vector<Integer> ps = new Vector<Integer>();
	Vector<String> ms = getIUPACWords(motif, mismatches);
	//System.out.println("Found motifs " + ms.toString());
	for (String w : ms) {
	    Pattern p = Pattern.compile(w);
	    Matcher m = p.matcher(seq);
	    int pos = 0;
	    while (m.find(pos)) {
		ps.addElement(m.start() + motif.length()/2);
		pos = m.start() + 1;
		//System.out.println("Found motif " + pos);
	    }
	    if (revcomp && !w.equals(MerTree.reverseComplement(w))) {
		p = Pattern.compile(MerTree.reverseComplement(w));
		m = p.matcher(seq);
		pos = 0;
		while (m.find(pos)) {
		    ps.addElement(m.start() + motif.length()/2);
		    pos = m.start() + 1;
		}
	    }
	}
	return ps;
    }

    public static Vector<String> loadSequences(String seqFileName) {
	Vector<String> seqs = new Vector<String>();
	int nChr = 21;
	for (int i=1; i<=nChr; i++)
	    seqs.addAll(loadSequence(seqFileName + "_" + i + ".fa"));
	System.out.println("Found " + seqs.size() + " sequences.");
	return seqs;
    }

    public static Vector<String> loadSequence(String seqFileName) {
	Vector<String> seqs = new Vector<String>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(seqFileName));
	    String line = br.readLine();
	    while (line!=null) {
		if (!line.startsWith(">"))
		    seqs.addElement(line);
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.out.println("Error parsing seqs " + e.toString());
	}
	return seqs;
    }

}