//Use this method to assemble all the reads across a given feature, eg
//TSSs, for all chromosomes. v2 is a more streamlined version of the
//bloated monster that FindReadsAcrossFeature has turned
//into. However, much of the code found there is still useful, so I
//try to re-use the functions that are good.
//
//Example:
//
//For TSS:
//
//bsub -q shared_2h java -Xmx1024m FindReadsAcrossFeature2 -f TSS -u 5000 -d 5000 -n MmDataPeakFiles/tf_kcl_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/KCl/B1/Input_Input_KCl_B1 -o -w 300 --noHists
//
//To look at a subset of promoters identified by Ivo as potentially tissue specific:
//
//bsub -q shared_2h java -Xmx1024m FindReadsAcrossFeature2 -f TSS -u 5000 -d 5000 -n MmDataPeakFiles/histones_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/MNase/B1/Input_Input_MNase_B1 -o -w 300 --select Ivo/inhibitory_specific_enhancers_d_10000_s_3.5_genes.txt
//
//To look at a subset of loci identified in Athar's fos experiments
//
//java -Xmx1024m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/Athar/fos_kcl_peaks.txt -u 1500 -d 1500 -n MmDataPeakFiles/fos_kcl_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Malik/KCl/B2/Input_Malik_KCl_B2 -o -w 300  --noHists -c 19 --motif /home/mh168/Projects/ChIP-Seq/TKData/Athar/fos_jaspar -G
//
//java -Xmx1024m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/Athar/fos_kcl_peaks.txt -u 1500 -d 1500 -n MmDataPeakFiles/fos_kcl_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Malik/KCl/B2/Input_Malik_KCl_B2 -o -w 100  --noHists -c 19 --consensus TGANTCA --consensus2 TGACGTCA -G
//
//To look at the transcription surrounding these loci
//
//java -Xmx1024m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/Athar/fos_names_other_peaks.txt -u 5000 -d 5000 -n MmDataPeakFiles/rna_names.txt -o -w 2000  --noHists -c 19
//
//To look at the subset of TSSs that have fos peaks with AP-1 motifs
//
//bsub -q shared_2h java -Xmx1024m FindReadsAcrossFeature2 -f TSS -u 5000 -d 5000 -n MmDataPeakFiles/fos_kcl_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Malik/KCl/B1/Input_Malik_KCl_B1 -o -w 200 --select Athar/ap1_fos_peaks_proximal_only_ids.txt
//
//To look at the reads surrounding the CREB peaks
//
//bsub -q shared_2h java -Xmx2048m -server FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/Timothy/creb_distal_peaks.txt -u 1500 -d 1500 -n MmDataPeakFiles/tf_kcl_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/KCl/B1/Input_Input_KCl_B1 -o -w 100  --noHists --consensus2 TGACGTCA -G
//
//For enhancers:
//
//bsub -q shared_2h java -Xmx1024m FindReadsAcrossFeature2 -f /groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/enhancers_b2_null_all.txt -u 5000 -d 5000 -n MmDataPeakFiles/tf_kcl_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/KCl/B1/Input_Input_KCl_B1 -o -w 300 --subset intergenic
//
//For looking at all CBP-peaks
//
//bsub -q shared_2h java -Xmx1024m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/MmDataPeakFiles/cbp_names_combos.txt -p CBP_Millipore_KCl_B1 -u 5000 -d 5000 -n MmDataPeakFiles/tf_kcl_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/KCl/B1/Input_Input_KCl_B1 -o -w 300 --subset intergenic
//
//bsub -q shared_2h java -Xmx1024m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/MmDataPeakFiles/cbp_names -p CBP_Millipore_KCl_B1 -u 0 -d 5000 -n MmDataPeakFiles/rna_names.txt -o -w 1500 --noHists --subset intergenic
//
//To get a subset of CBP-peaks
//
//bsub -q shared_12h java -server -Xmx4096m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/MmDataPeakFiles/cbp_names -p CBP_Millipore_KCl_B1 -u 5000 -d 5000 -n MmDataPeakFiles/tf_kcl_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/KCl/B1/Input_Input_KCl_B1 -o -w 100 --noHists --subset intergenic -S /groups/neuroduo/jesse/enhancers_for_martin/CBP_Millipore_KCl_B1_subset_enh_ex_hMe3_real.tsv
//
//To get the ESTs near CBP peaks
//
//bsub -q shared_2h java -Xmx1024m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/MmDataPeakFiles/cbp_names -p CBP_Millipore_KCl_B1 -u 5000 -d 5000 -n MmDataPeakFiles/rna_names.txt -o -w 300 --noHists --subset intergenic --EST
//
//For other features:
//
//bsub -q shared_2h java -Xmx1024m FindReadsAcrossFeature2 -f /groups/neuroduo/martin/ChIP-Seq/MmData/Brenda/promoter_peaks_Npas4_Npas4_KCl_B1.txt -u 5000 -d 5000 -n MmDataPeakFiles/rna_names.txt -o -w 1500 --noHists
//
//For Sissrs peaks:
//
//bsub -q shared_2h java -Xmx4096m FindReadsAcrossFeature2 -f CBP_Millipore_KCl_B1.sissrs -u 5000 -d 5000 -o -w 300 --noHists  -n /groups/neuroduo/martin/ChIP-Seq/MmData/CBP/Millipore/KCl/B1/CBP_Millipore_KCl_B1  -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/KCl/B1/Input_Input_KCl_B1
//
//For generating heatmaps
//
//bsub -q shared_12h java -server -Xmx4096m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/MmDataPeakFiles/cbp_names -p CBP_Millipore_KCl_B1 -u 5000 -d 5000 -n MmDataPeakFiles/histones_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/MNase/B1/Input_Input_MNase_B1 -o -w 1500 --subset intergenic -S /groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_heatmaps.tsv
//
//For internal promoters
//
//bsub -q shared_12h java -server -Xmx4096m FindReadsAcrossFeature2 -f /groups/neuroduo/martin/Genomes/Mm/v9/internal_promoters.tsv -u 0 -d 5000 -n MmDataPeakFiles/rna_names.txt -o -w 1500
//
//When we want to downsample to compare the directionality at enhancers and promoters
//
//bsub -q shared_2h java -Xmx4096m FindReadsAcrossFeature2 --noHists -f TSS -u 0 -d 5000 -n MmDataPeakFiles/rna_names.txt -o -w 1500 --downSampleDist /groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/ReadDists/CBP_Millipore_KCl_B1_subset_enh_ex_lMe3_real_divergent
//
//To look at the remaining transcripts from the transcript caller:
//
//bsub -q shared_2h java -Xmx4096m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/Transcripts/RemainingReadsHistograms/remaining_reads_transcripts_min_reads_4.txt -u 5000 -d 5000 -n ../MmDataPeakFiles/hist_names.txt -o -w 1500 --noHists -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/MNase/B1/Input_Input_MNase_B1
//
//For investigating the vicinity of conserved starts of transcribed regions
//
//bsub -q shared_12h java -server -Xmx4096m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/Transcripts/merged_total_TSS_cons_transcript_starts.txt -u 5000 -d 5000 -n MmDataPeakFiles/histone_kcl_names.txt -o -w 1500 --noHists -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/MNase/B1/Input_Input_MNase_B1 -G -C
//
//For looking at MeCP2 densities around the TSS
//
//bsub -q shared_2h java -Xmx1024m FindReadsAcrossFeature2 -f TSS -u 5000 -d 5000 -n /groups/neuroduo/martin/ChIP-Seq/MmData/MeCP2/MeCP2/2hrKClIPSonicated/B2/MeCP2_MeCP2_2hrKClIPSonicated_B2 -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/MeCP2/2hrKClSonicated/B2/Input_MeCP2_2hrKClSonicated_B2 -o -w 300 --noHists -c 19
//
//Find reads near fos-peaks, allowing for duplicates
//
//bsub -q shared_12h java -Xmx2048m FindReadsAcrossFeature2 -f /home/mh168/Projects/ChIP-Seq/TKData/Athar/ap1_fos_cfos_sc7202_KCl_B3_peaks_ids.txt -u 1500 -d 1500 -n FosPeakFiles/fos_kcl3_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Malik/KCl/B3/Input_Malik_KCl_B3.bed -l 250 -o -w 400  --noHists --keepDuplicates
//
//

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import org.apache.commons.math.distribution.BinomialDistributionImpl;
import org.apache.commons.math.MathException;

public class FindReadsAcrossFeature2{

    public static final String SCATTER_DIR = "/groups/neuroduo/martin/ChIP-Seq/MmData/Scatter/";
    public static final String READ_COUNT_DIR = "/groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/ReadDists/";
    public static final String HISTOGRAM_DIR = "/groups/neuroduo/martin/ChIP-Seq/MmData/Histograms/";
    public static final String EST_DIR = "/groups/neuroduo/martin/ChIP-Seq/MmData/ESTs/";
    public static final String DOWN_SAMPLE_STR = "downsampled";
    public static final String OUTLIER_FILE_LOCATION = "/home/mh168/Projects/ChIP-Seq/TKData/Transcripts/outliers.txt";

    public static final int ENHANCER_DOMAIN_FLANKING_START = 2500;
    public static final int ENHANCER_DOMAIN_FLANKING_END = 4000;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option strandFlag = argParser.addBooleanOption('r', "reverse");
	    CmdLineParser.Option featureFlag = argParser.addStringOption('f', "feature");
	    //Specify which type of peaks that we are interested in
	    CmdLineParser.Option peakNameFlag = argParser.addStringOption('p', "peakName");
	    //Use this flag if we only want to consider the intronic
	    //or intergenic peaks or enhancers, default is to consider all
	    CmdLineParser.Option subsetFlag = argParser.addStringOption("subset");
	    CmdLineParser.Option upstreamFlag = argParser.addIntegerOption('u', "upstream");
	    CmdLineParser.Option downstreamFlag = argParser.addIntegerOption('d', "downstream");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    //Use this flag if we want to coarse-grain the reads with
	    //a given bin-size inside the window
	    CmdLineParser.Option binSizeFlag = argParser.addIntegerOption('b', "binSize");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //use this flag when we only want to focus on features
	    //which are guaranteed to be isolated from one another.
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('I', "isolated");
	    //Use this flag to include a file with genes/peaks to filter against
	    CmdLineParser.Option selectFlag = argParser.addStringOption('s', "select");
	    //Use this flag when wish to include a subset of the peaks
	    CmdLineParser.Option selectPeakFlag = argParser.addStringOption('S', "selectPeak");
	    //Use this flag to search for the GC content in the region as well
	    CmdLineParser.Option gcFlag = argParser.addBooleanOption('G', "gc");
	    //Find the conservation scores as well
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption('C', "conservation");
	    //Avoid producing the files containing individual patterns for each loci
	    CmdLineParser.Option noHistsFlag = argParser.addBooleanOption("noHists");
	    //Only produce histograms
	    CmdLineParser.Option histsOnlyFlag = argParser.addBooleanOption("histsOnly");
	    //Extract ESTs around the feature in question
	    CmdLineParser.Option estFileNameFlag = argParser.addBooleanOption("EST");
	    //Create a set of random loci, matching the relative distance to TSS of the real one
	    CmdLineParser.Option randFlag = argParser.addBooleanOption("rand");
	    //Calculate the mappability of each locus
	    CmdLineParser.Option mappabilityFlag = argParser.addBooleanOption("mappability");
	    //Distribution to use when downsampling
	    CmdLineParser.Option downSampleDistFlag = argParser.addStringOption("downSampleDist");
	    //When we want to orient according to S and AS
	    CmdLineParser.Option senseOrientFlag = argParser.addIntegerOption("sense");
	    //For defining the size of the flanking domains, the lenght is by default the same as the windowSize
	    CmdLineParser.Option flankingDomainStartFlag = argParser.addIntegerOption("flankStart");
	    //When searching for motifs in the sequences
	    CmdLineParser.Option motifNameFlag = argParser.addStringOption("motif");
	    CmdLineParser.Option consensusSequenceFlag = argParser.addStringOption("consensus");
	    CmdLineParser.Option consensus2SequenceFlag = argParser.addStringOption("consensus2");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    //For keeping the PCR duplicates
	    CmdLineParser.Option duplicatesFlag = argParser.addBooleanOption("keepDuplicates");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : "";
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(peakNameFlag);
	    String peakName = tmp!=null ? (String)tmp : null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean forward = argParser.getOptionValue(strandFlag)==null;
	    boolean gc = argParser.getOptionValue(gcFlag)!=null;
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    boolean unique = argParser.getOptionValue(duplicatesFlag)==null;
	    tmp = argParser.getOptionValue(featureFlag);
	    String feature = tmp==null ? "TSS" : tmp.toString();
	    tmp = argParser.getOptionValue(subsetFlag);
	    String subset = tmp==null ? "" : tmp.toString();
	    tmp = argParser.getOptionValue(senseOrientFlag);
	    int senseOrient = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(flankingDomainStartFlag);
	    int flankingDomainStart = tmp!=null ? ((Integer)tmp).intValue() : ENHANCER_DOMAIN_FLANKING_START;
	    tmp = argParser.getOptionValue(upstreamFlag);
	    int upstream = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(downstreamFlag);
	    int downstream = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 300;
	    //This is problematic and for now we have to give downstreamSeq and upstreamSeq the same value as upstream and downstream to get the profile and scatter stuff to work
	    int downstreamSeq = downstream, upstreamSeq = upstream;
	    //int downstreamSeq = windowSize, upstreamSeq = windowSize;
	    tmp = argParser.getOptionValue(binSizeFlag);
	    int binSize = tmp!=null ? ((Integer)tmp).intValue() : windowSize;
	    tmp = argParser.getOptionValue(selectFlag);
	    String selectFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(downSampleDistFlag);
	    String downSampleDist = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(selectPeakFlag);
	    String selectPeaksFileName = tmp!=null ? (String)tmp : null;
	    boolean noHists = argParser.getOptionValue(noHistsFlag)!=null;
	    boolean histsOnly = argParser.getOptionValue(histsOnlyFlag)!=null;
	    String estFileName = argParser.getOptionValue(estFileNameFlag)!=null ? (orchestra ? "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/ESTs/" : null) : null;
	    boolean rand = argParser.getOptionValue(randFlag)!=null;
	    boolean mappability = argParser.getOptionValue(mappabilityFlag)!=null;
	    tmp = argParser.getOptionValue(motifNameFlag);
	    String motifName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(consensusSequenceFlag);
	    String consensusSequence = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(consensus2SequenceFlag);
	    String consensus2Sequence = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : fileName.contains("rna") ? 35 : (fileName.contains("fos") || fileName.contains("Malik") || fileName.contains("MeCP2") || fileName.contains("mecp2") ? 150 : (fileName.contains("MikkelsenData") ? 300 : 120));
	    //////////////////////////////////////////////////////////////////
	    int nChr = species.equalsIgnoreCase("mm") ? 21 : 24;
	    double pValMotifFinding = .001;
	    String saveFileName = null, scatterFileStart = null;
	    RefSeqGenes[] geneInfo = new RefSeqGenes[nChr];
	    double qualityScore = FindReadsAcrossFeature.QUALITY_SCORE_CUTOFF;
	    int isolatedIDs[] = isolated ? FindReadsAcrossFeature.parseIsolatedFile(orchestra) : null;
	    int geneIDsKeep[] = FindReadsAcrossFeature.parseSelectFile(selectFileName);
	    String repeatFileName = species.toLowerCase().equals("mm") ? EstimateGenicTranscription3.REPEAT_RNA_FILE_MM : EstimateGenicTranscription3.REPEAT_RNA_FILE_HS_36;

	    Vector<int[]> peaksKeep = parseSelectPeakFile(selectPeaksFileName);
	    Vector<String> fileNames = new Vector<String>(), shortNames = new Vector<String>();
	    if (fileName.contains("names"))
		fileNames = FindOverlappingPeaks.getFileNames(fileName, shortNames);
	    else if (fileName.length()>0) {
		fileNames.addElement(fileName);
		shortNames.addElement(fileName.substring(fileName.lastIndexOf("/") + 1));
	    }
	    System.out.println("Feature is " + feature + " and there are " + fileNames.size() + " factors.");
	    //Note! This has only been implemented for TSSs
	    if (downSampleDist!=null && !feature.equals("TSS")) 
		throw new Exception("Warning: this feature is only implemented for TSS! Aborting...");
	    TreeMap<String, Integer> nTotalReads = calculateTotalReads(fileNames, inputFileName, nChr, fragmentLength, qualityScore);
	    double histsMean[][][] = new double[fileNames.size()][nChr][FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1)];
	    double histsConsMean[][] = new double[nChr][FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1)], histsGCMean[][] = new double[nChr][FindReadsAcrossFeature.numberOfBins(upstreamSeq, downstreamSeq, 1)], histsCpGMean[][] = new double[nChr][FindReadsAcrossFeature.numberOfBins(upstreamSeq, downstreamSeq, 1)];
	    double sumHistMean[][] = new double[fileNames.size()][FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1)], sumHistVar[][] = new double[fileNames.size()][FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1)], sumHistCons[] = new double[FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1)], sumHistGC[] = new double[FindReadsAcrossFeature.numberOfBins(upstreamSeq, downstreamSeq, 1)], sumHistCpG[] = new double[FindReadsAcrossFeature.numberOfBins(upstreamSeq, downstreamSeq, 1)];
	    int sumLocs = 0, nLocs[] = new int[nChr];
	    Vector<double[][][]> nReadsInWindows = new Vector<double[][][]>();
	    Vector<double[][]> downstreamEnhancerDomains = new Vector<double[][]>(), upstreamEnhancerDomains = new Vector<double[][]>(), consInWindows = new Vector<double[][]>(), gcInWindows = new Vector<double[][]>(), cpgInWindows = new Vector<double[][]>();
	    Vector<int[]> locVecs = new Vector<int[]>(), mapVecs = new Vector<int[]>(), ids = new Vector<int[]>(), nMismatches = new Vector<int[]>(), nMismatches2 = new Vector<int[]>();
	    Vector<boolean[]> strandVecs = new Vector<boolean[]>();
	    Vector<double[]> motifScores = new Vector<double[]>();
	    //////////////////////////////////////////////////////////////////
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra);
		    int locVec[] = null, mapVec[] = null, id[] = null;
		    boolean strandVec[] = null;
		    String histSaveNameStart = null, saveReadCountDistributionStart = null;
		    double downstreamEnhancerDomain[][] = null, upstreamEnhancerDomain[][] = null;
		    TreeMap<String, double[]> downSampleDistribution = null;
		    if (feature.equals("TSS") || feature.equals("TSSi")) {
			/////////////////////////////////
			//Get the profiles surrounding the TSSs
			boolean internal = feature.equals("TSSi");
			locVec = FindReadsAcrossFeature.getTSSs(geneInfo[i-1], geneIDsKeep, isolatedIDs, internal);
			strandVec = FindReadsAcrossFeature.getStrands(geneInfo[i-1], geneIDsKeep, isolatedIDs, internal);
			id = geneInfo[i-1].getIDs(/*internal*/);
			System.out.println("Found " + locVec.length + " locations and " + id.length + " IDs.");
			if (senseOrient!=0) { //Remove the loci which are not suitable for this combination
			    locVec = senseFilterLoci(senseOrient, forward, locVec, strandVec, geneInfo[i-1]);
			    strandVec = new boolean[locVec.length];
			    Arrays.fill(strandVec, forward);
			}
			scatterFileStart = SCATTER_DIR + "tss" + (internal ? "i" : "") + "_" + (downSampleDist!=null ? DOWN_SAMPLE_STR + "_" : "") + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")) + "_" + (senseOrient==0 ? "" : senseOrient>0 ? "S_" : "AS_") + (inputFileName==null ? (fileName.contains("rna") ? (forward ? "f_" : "r_") : "") : "input_") + "w_" + windowSize + (upstream==0 ? "d" : (downstream==0 ? "u" : ""));
			if (!noHists)
			    histSaveNameStart = FindReadsAcrossFeature.TSS_HISTOGRAM_PREFIX + "_";
			saveFileName = fileName.substring(0, fileName.lastIndexOf(".")) + "_combined_TSS" + (internal ? "i" : "") + (downSampleDist!=null ? "_" + DOWN_SAMPLE_STR : "") + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (senseOrient==0 ? "" : senseOrient>0 ? "_S" : "_AS") + (inputFileName!=null ? "_input" : (fileName.contains("rna") ? (forward ? "_f" : "_r") : "")) + (upstream==0 ? "_d" : (downstream==0 ? "_u" : "")) + (chr!=-1 ? "_" + chr : "") + ".txt";
			//For saving the read counts
			if (fileName.contains("rna"))
			    saveReadCountDistributionStart = READ_COUNT_DIR + "tss" + (internal ? "i" : "") + (selectPeaksFileName==null ? "" : "_" + selectPeaksFileName.substring(selectPeaksFileName.lastIndexOf("subset_"), selectPeaksFileName.lastIndexOf("."))) + (senseOrient==0 ? "" : senseOrient>0 ? "S_" : "AS_") + (forward ? "_f" : "_r") + (upstream==0 ? "_d" : (downstream==0 ? "_u" : ""));
			downSampleDistribution = downSampleDist!=null && !internal ? loadDownSampleDistribution(downSampleDist, shortNames, i) : null;
			
		    }
		    else if (/*feature.contains("enhancer") || */feature.contains("Enhancer")) {
			Vector<Enhancer> enhancers = FindReadsAcrossFeature.parseEnhancerFile(feature, i);
			locVec = new int[enhancers.size()];
			for (int j=0; j<locVec.length; j++)
			    locVec[j] = enhancers.elementAt(j).getCenter();
			strandVec = new boolean[locVec.length];
			Arrays.fill(strandVec, true);
			scatterFileStart = SCATTER_DIR + feature.substring(feature.lastIndexOf("/") + 1, feature.lastIndexOf(".")) + "_" + fileName + "_" + FindReadsAcrossFeature.AGGREGATE_SUFFIX + (subset.length()>0 ? subset + "_" : "") + (inputFileName==null ? (fileName.contains("rna") ? (forward ? "f_" : "r_") : "") : "input_") + "w_" + windowSize;
			if (!noHists)
			    histSaveNameStart = HISTOGRAM_DIR + feature.substring(feature.lastIndexOf("/") + 1, feature.lastIndexOf(".")); // + "_" + fileName + "_" + FindReadsAcrossFeature.AGGREGATE_SUFFIX + (subset.length()>0 ? subset + "_" : "") + (inputFileName==null ? (fileName.contains("rna") ? (forward ? "f_" : "r_") : "") : "input_") + "w_" + windowSize;
			saveFileName = fileName.substring(0, fileName.lastIndexOf(".")) + "_combined_" + feature.substring(feature.lastIndexOf("/"), feature.lastIndexOf(".")) + (inputFileName!=null ? "_input" : (fileName.contains("rna") ? (forward ? "f_" : "r_") : ""));
		    }
		    else if (feature.contains("PeakFiles") || feature.contains("peakfiles")) {
			//Get the peak info
			TreeMap<Integer, Peak> peaks = FindEnhancerSites.parsePeakFile(feature, i, false);
			TreeSet<String> peakNames = new TreeSet<String>(FindEnhancerSites.parsePeakFileHeader(feature, i).values());
			TreeSet<Peak> removePeaks = new TreeSet<Peak>();
			for (Peak p : peaks.values()) {
			    if (p.hasFactor(peakName)) {
				RefSeqGene overlappingGene = geneInfo[i-1].getOverlappingGene(p.getCenter());
				if (!(subset==null || subset.equals("") || (subset.equals(FindReadsAcrossFeature.INTERGENIC_STR) && overlappingGene==null) || (subset.equals(FindReadsAcrossFeature.INTRONIC_STR) && overlappingGene!=null)))
				    removePeaks.add(p);
				else {
				    if (!(peaksKeep==null || Arrays.binarySearch(peaksKeep.elementAt(i-1), p.getCenter())>=0))
					removePeaks.add(p);
				    else if (senseOrient!=0 && overlappingGene!=null) {
					//If we want to orient according to S/AS, we need to remove the peaks that are on the "wrong" strand. This is a last-minute ugly hack!
					if (!strandOK(senseOrient, forward, overlappingGene.getStrand()))
					    removePeaks.add(p);
				    }
				}
			    }
			    else
				removePeaks.add(p);
			}
			for (Peak p : removePeaks) 
			    peaks.remove(p.getCenter());
			System.out.println("Found " + peaks.size() + " for chromosome " + i + " and removed " + removePeaks.size());
			if (rand) 
			    peaks = createRandomPeaks(peaks, subset, geneInfo[i-1], feature + "_" + peakName + "_" + (subset.length()>0 ? subset + "_" : "") + "rand_peak_loci_" + i + ".txt");
			scatterFileStart = SCATTER_DIR + peakName + "_" + (fileNames.size()>1 ? fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")) : fileName.substring(fileName.lastIndexOf("/") + 1)) + "_" + (subset.length()>0 ? subset + "_" : "") + (rand ? "rand_" : "") + (selectPeaksFileName==null ? "" : selectPeaksFileName.substring(selectPeaksFileName.lastIndexOf("subset_"), selectPeaksFileName.lastIndexOf(".")) + "_") + (senseOrient==0 ? "" : senseOrient>0 ? "S_" : "AS_") + (inputFileName==null ? (fileName.contains("rna") ? (forward ? "f_" : "r_") : "") : "input_") + "w_" + windowSize + (upstream==0 ? "d" : (downstream==0 ? "u" : "")) + (windowSize>binSize ? "_b_" + binSize : "");
			if (!noHists) 
			    histSaveNameStart = HISTOGRAM_DIR + peakName + "_" + (subset.length()>0 ? subset + "_" : "") + (rand ? "rand_" : "") + (selectPeaksFileName==null ? "" : selectPeaksFileName.substring(selectPeaksFileName.lastIndexOf("subset_"), selectPeaksFileName.lastIndexOf(".")) + "_") + (senseOrient==0 ? "" : senseOrient>0 ? "S_" : "AS_") + (inputFileName==null ? (fileName.contains("rna") ? (forward ? "f_" : "r_") : "") : "input_"); 
			saveFileName = fileNames.size()>1 ? fileName.substring(0, fileName.lastIndexOf(".")) + "_combined_" + peakName + (subset.length()>0 ?  "_" + subset : "") + (rand ? "_rand" : "") + (selectPeaksFileName==null ? "" : "_" + selectPeaksFileName.substring(selectPeaksFileName.lastIndexOf("subset_"), selectPeaksFileName.lastIndexOf("."))) + (senseOrient==0 ? "" : senseOrient>0 ? "_S" : "_AS") + (inputFileName!=null ? "_input" : (fileName.contains("rna") ? (forward ? "_f" : "_r") : "")) + (upstream==0 ? "_d" : (downstream==0 ? "_u" : "")) + (chr!=-1 ? "_" + chr : "") + ".txt" : null;
			//For saving the read counts
			if (fileName.contains("rna") && !histsOnly) {
			    saveReadCountDistributionStart = READ_COUNT_DIR + peakName + (subset.length()>0 ?  "_" + subset : "") + (rand ? "_rand" : "") + (selectPeaksFileName==null ? "" : "_" + selectPeaksFileName.substring(selectPeaksFileName.lastIndexOf("subset_"), selectPeaksFileName.lastIndexOf("."))) + (senseOrient==0 ? "" : senseOrient>0 ? "S_" : "AS_") + (forward ? "_f" : "_r") + (upstream==0 ? "_d" : (downstream==0 ? "_u" : ""));
			    System.out.println("saving counts to " + saveReadCountDistributionStart);
			}
			/////////////////////////////////
			//EST stuff
			LinkedHashMap<String, EST> estTagsMapForward = estFileName==null ? null : CompareCageEstRNASeqStarts.parseEstTags(estFileName, i, true, species);
			LinkedHashMap<String, EST> estTagsMapReverse = estFileName==null ? null : CompareCageEstRNASeqStarts.parseEstTags(estFileName, i, false, species);
			String saveESTsName = EST_DIR + peakName + "_" + FindEnhancerSites.ENHANCER_EST_FILE_SUFFIX + (rand ? "_rand" : "") + (selectPeaksFileName==null ? "" : "_" + selectPeaksFileName.substring(selectPeaksFileName.lastIndexOf("subset_"), selectPeaksFileName.lastIndexOf("."))) + "_w_" + FindEnhancerSites.EST_ENHANCER_DISTANCE + "_" + i + ".txt";
			BufferedWriter estWriter = estFileName==null ? null : new BufferedWriter(new FileWriter(saveESTsName));
			if (estFileName!=null) { 
			    System.out.println("Saving est sites to " + saveESTsName);
			    if (estWriter!=null && chr==1)
				estWriter.write("#chr\tCBP\tGeneID\tEST_5p\tEST_3p\tStrand\tSpliced\t\tOverlapsTSS\tName\n");
			}
			//Mappability stuff
			Mappability map = mappability ? new Mappability(orchestra, i, 2*windowSize + 1, 35, species) : null;
			locVec = new int[peaks.size()];
			mapVec = new int[peaks.size()];
			int j = 0, prevLoc = 0;
			for (Peak p : peaks.values()) {
			    RefSeqGene overlappingGene = geneInfo[i - 1].getOverlappingGene(p.getCenter());
			    locVec[j] = p.getCenter();
			    if (estFileName!=null) { //discard if there are too many ESTs starting near the enhancer
				FindEnhancerSites.findNearbyESTs(estTagsMapForward, i, true, p.getCenter(), FindEnhancerSites.EST_ENHANCER_DISTANCE, overlappingGene==null ? -1 : overlappingGene.getID(), geneInfo[i - 1], estWriter);
				FindEnhancerSites.findNearbyESTs(estTagsMapReverse, i, false, p.getCenter(), FindEnhancerSites.EST_ENHANCER_DISTANCE, overlappingGene==null ? -1 : overlappingGene.getID(), geneInfo[i - 1], estWriter);
			    }
			    if (mappability && p.getChromosome()==i) {
				int moveWindowDist = p.getCenter() - prevLoc - windowSize;
				mapVec[j] = map.numberOfPotentialReadsNextWindow(moveWindowDist);
				if (mapVec[j]<0)
				    System.err.println("Error, found negative mappability " + mapVec[j] + " for chr" + i + ":" + p.getCenter());
				prevLoc = p.getCenter() - windowSize;
			    }
			    j++;
			}
			strandVec = new boolean[locVec.length]; //getStrandVec(subset, geneInfo[i-1], locVec);
			if (estWriter!=null)
			    estWriter.close();
			//Investigate the flanking domains
			if (fileName.contains("rna")) {
			    if (downstream>0)
				downstreamEnhancerDomain = new double[fileNames.size()][locVec.length];
			    if (upstream>0)
				upstreamEnhancerDomain = new double[fileNames.size()][locVec.length];
			}

		    }
		    else if (feature.contains("transcripts")) {
			//A list of regions in .bed format. use the center of the region by default
			locVec = parseBedFile(feature, i, 0);
			strandVec = parseBedFileStrandInfo(feature, i);
			scatterFileStart = SCATTER_DIR + feature.substring(feature.lastIndexOf("/") + 1, feature.lastIndexOf(".")) + "_" + (fileNames.contains(".") ? fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")) : fileName.substring(fileName.lastIndexOf("/") + 1)) + "_" + (inputFileName==null ? (fileName.contains("rna") ? (forward ? "f_" : "r_") : "") : "input_") + "w_" + windowSize + (upstream==0 ? "d" : (downstream==0 ? "u" : "")) + (windowSize>binSize ? "_b_" + binSize : "");
			if (noHists)
			    histSaveNameStart = feature.substring(0, feature.lastIndexOf("."));
			saveFileName = (fileName.contains(".") ? fileName.substring(0, fileName.lastIndexOf(".")) : fileName) + "_combined_" + feature.substring(feature.lastIndexOf("/")+1, feature.lastIndexOf(".")) + (inputFileName!=null ? "_input" : (fileName.contains("rna") ? (forward ? "_f" : "_r") : "")) + (upstream==0 ? "_d" : (downstream==0 ? "_u" : "")) + (chr!=-1 ? "_" + chr : "") + ".txt";
		    }
		    else {
			//Just some list of features, assume that the
			//file contains two columns representing chr
			//and coordinates. Optionally, we have a third
			//column with strand info
			locVec = parseFeatureFile(feature, i, true/*feature.contains("sissrs")*/);
			strandVec = parseFeatureFileStrandInfo(feature, i, locVec);
			if (senseOrient!=0) {
			    locVec = senseFilterLoci(senseOrient, forward, locVec, strandVec, geneInfo[i-1]);
			    strandVec = new boolean[locVec.length];
			    Arrays.fill(strandVec, forward);
			}
			scatterFileStart = SCATTER_DIR + feature.substring(feature.lastIndexOf("/") + 1, feature.lastIndexOf(".")) + "_" + (fileNames.contains(".") ? fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")) : fileName.substring(fileName.lastIndexOf("/") + 1)) + (senseOrient==0 ? "" : senseOrient>0 ? "_S" : "_AS") + "_" + (!unique ? "duplicates_" : "") + (inputFileName==null ? (fileName.contains("rna") ? (forward ? "f_" : "r_") : "") : "input_") + "w_" + windowSize + (upstream==0 ? "d" : (downstream==0 ? "u" : "")) + (windowSize>binSize ? "_b_" + binSize : "");
			if (!noHists) {
			    histSaveNameStart = HISTOGRAM_DIR + "Peaks/" + feature.substring(feature.lastIndexOf("/") + 1, feature.lastIndexOf(".")) + "/"; //feature.substring(0, feature.lastIndexOf("."));
			    File dir = new File(histSaveNameStart);
			    dir.mkdirs();
			    System.out.println("Created " + histSaveNameStart);
			}
			if (fileName.length()>0) 
			    saveFileName = (fileName.contains(".") ? fileName.substring(0, fileName.lastIndexOf(".")) : fileName) + "_combined_" + feature.substring(feature.lastIndexOf("/")+1, feature.lastIndexOf(".")) + (senseOrient==0 ? "" : senseOrient>0 ? "_S" : "_AS") + (inputFileName!=null ? "_input" : (fileName.contains("rna") ? (forward ? "_f" : "_r") : "")) + (upstream==0 ? "_d" : (downstream==0 ? "_u" : "")) + (chr!=-1 ? "_" + chr : "") + ".txt";
			else
			    saveFileName = feature.substring(0, feature.lastIndexOf(".")) + "_combined_" + (senseOrient==0 ? "" : senseOrient>0 ? "_S" : "_AS") + (inputFileName!=null ? "_input" : (fileName.contains("rna") ? (forward ? "_f" : "_r") : "")) + (upstream==0 ? "_d" : (downstream==0 ? "_u" : "")) + (chr!=-1 ? "_" + chr : "") + ".txt";
		    }
		    nLocs[i-1] = locVec.length;
		    sumLocs += locVec.length;
		    int nBins = fileName.contains("rna") ? (/*upstream==0 || downstream==0 ? windowSize/binSize : 2*/windowSize/binSize) : 1;
		    double nReadsInWindow[][][] = new double[fileNames.size()][locVec.length][nBins], consInWindow[][] = new double[locVec.length][nBins], gcInWindow[][] = new double[locVec.length][nBins], cpgInWindow[][] = new double[locVec.length][nBins];
		    double motifScore[] = new double[locVec.length];
		    int nMismatch[] = new int[locVec.length], nMismatch2[] = new int[locVec.length];
		    /////////////////////////////////
		    for (int k=0; k<fileNames.size(); k++) {
			//Load the reads
			boolean rna = (fileNames.elementAt(k).contains("RNA") || fileNames.elementAt(k).contains("rna")) && !(fileNames.elementAt(k).contains("RNAi") || fileNames.elementAt(k).contains("rnai"));
			ABIRead reads[] = null, readsInput[] = null;
			if (!rna) {
			    if (fileNames.elementAt(k).contains("MikkelsenData"))
				reads = PeakFinder2.parseBroadFiles(i, fileNames.elementAt(k), fragmentLength, true);
			    else if (fileNames.elementAt(k).endsWith("bed"))
				reads = CreateWigFiles.parseBedFiles(i, fileNames.elementAt(k), unique, fragmentLength);
			    else
				reads = PeakFinder2.parseFiles(i, fileNames.elementAt(k), fragmentLength, true, 0, qualityScore); //unique option not implemented here
			}
			if (inputFileName!=null) {
			    if (inputFileName.contains("MikkelsenData"))
				readsInput = PeakFinder2.parseBroadFiles(i, inputFileName, fragmentLength, true);
			    else if (inputFileName.endsWith("bed"))
				readsInput = CreateWigFiles.parseBedFiles(i, inputFileName, true, fragmentLength);
			    else
				readsInput = PeakFinder2.parseFiles(i, inputFileName, fragmentLength, true, 0, qualityScore);
			}
			//Get the histograms surrounding the feature
			String histSaveName = noHists ? null : histSaveNameStart + shortNames.elementAt(k) + "_" + i + ".txt";
			String saveReadCountDistributionFile = rna && saveReadCountDistributionStart!=null ? saveReadCountDistributionStart + "_" + fileNames.elementAt(k).substring(fileNames.elementAt(k).lastIndexOf("/") + 1, fileNames.elementAt(k).lastIndexOf(".")) + (chr!=-1 ? "_" + chr : "") + ".txt" : null;
			double hists[][], histsCons[][], histsGC[][], histsCpG[][];
			if (rna) {
			    ABIRead readsForward[] = ABIReads.loadMRNAReadsChr(fileNames.elementAt(k), true, i, 35, false, repeatFileName);
			    ABIRead readsReverse[] = ABIReads.loadMRNAReadsChr(fileNames.elementAt(k), false, i, 35, false, repeatFileName);
			    hists = getRNAHistograms(i, locVec, strandVec, forward, readsForward, readsReverse, downstream, upstream, 1, 35, (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(fileNames.elementAt(k)));
			}
			else {
			    double normFactor = (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(fileNames.elementAt(k));
			    double inputNormFactor = inputFileName==null ? 0 : (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(inputFileName);
			    hists = getHistograms(i, locVec, strandVec, reads, readsInput, downstream, upstream, fragmentLength, histSaveName, normFactor, inputNormFactor, rna, downSampleDistribution==null ? null : downSampleDistribution.get(shortNames.elementAt(k)), saveReadCountDistributionFile, chr==-1 && i>1);
			}
			//Find the number of reads for the scatter plots
			for (int l=0; l<hists.length; l++) {
			    if (windowSize>binSize) {
				int m = -1;
				for (int j=Math.max(0, upstream - windowSize); j<Math.min(hists[l].length, upstream + windowSize) - 1; j++) {
				    m += j % binSize == 0 ? 1 : 0;
				    if (j!=windowSize) //avoid counting 0
					nReadsInWindow[k][l][m] += hists[l][j]/fragmentLength;
				}
			    }
			    else {
				for (int j=Math.max(0, upstream - windowSize); j<Math.min(hists[l].length, upstream + windowSize + 1); j++) 
				    nReadsInWindow[k][l][0] += hists[l][j]/fragmentLength;
				for (int j=0; j<hists[l].length; j++) {
				    if (upstreamEnhancerDomain!=null && flankingEnhancerDomain(j, upstream, windowSize)<0)
					//Count the number of reads in the regions flanking the enhancer domain
					upstreamEnhancerDomain[k][l] += hists[l][j]/fragmentLength;
				    else if (downstreamEnhancerDomain!=null && flankingEnhancerDomain(j, upstream, windowSize)>0)
					downstreamEnhancerDomain[k][l] += hists[l][j]/fragmentLength;
				}
			    }
			}
			//Get the average histogram
			histsMean[k][i-1] = findMeanReadHistograms(hists);
		    }
		    if (conservation) {
			double histsCons[][] = findConservationHistogram(i, orchestra, species, locVec, strandVec, downstream, upstream);
			histsConsMean[i-1] = getConservationHistogram(histsCons);
			for (int l=0; l<histsCons.length; l++) {
			    if (windowSize>binSize) {
				int m = -1;
				for (int j=Math.max(0, upstream - windowSize); j<Math.min(histsCons[l].length, upstream + windowSize) - 1; j++) {
				    m += j % binSize == 0 ? 1 : 0;
				    if (j!=windowSize) //avoid counting 0
					consInWindow[l][m] += histsCons[l][j];
				}
			    }
			    else {
				for (int j=Math.max(0, upstream - windowSize); j<Math.min(histsCons[l].length, upstream + windowSize + 1); j++) 
				    consInWindow[l][0] += histsCons[l][j];
			    }
			}
		    }
		    if (gc) {
			String seqs[] = getSequences(i, orchestra, species, locVec, strandVec, upstreamSeq, downstreamSeq);
			double histsGC[][] = getGCHistogram(seqs);
			double histsCpG[][] = getCpGHistogram(seqs);
			histsGCMean[i-1] = findMeanReadHistograms(histsGC);
			histsCpGMean[i-1] = findMeanReadHistograms(histsCpG);
			for (int l=0; l<histsGC.length; l++) {
			    if (windowSize>binSize) {
				int m = -1;
				for (int j=Math.max(0, upstream - windowSize); j<Math.min(histsGC[l].length, upstream + windowSize) - 1; j++) {
				    m += j % binSize == 0 ? 1 : 0;
				    if (j!=windowSize) { //avoid counting 0
					gcInWindow[l][m] += histsGC[l][j];
					cpgInWindow[l][m] += histsCpG[l][j];
				    }
				}
			    }
			    else {
				for (int j=Math.max(0, upstream - windowSize); j<Math.min(histsGC[l].length, upstream + windowSize + 1); j++) {
				    gcInWindow[l][0] += histsGC[l][j];
				    cpgInWindow[l][0] += histsCpG[l][j];
				}
			    }
			    if (consensusSequence!=null) {
				nMismatch[l] = bestNumberOfMismatches(seqs[l], consensusSequence);
			    }
			    if (consensus2Sequence!=null) {
				nMismatch2[l] = bestNumberOfMismatches(seqs[l], consensus2Sequence);
			    }
			}
			if (motifName!=null) { //Run TFMScan
			    TFMScanner tfmScanner = new TFMScanner(seqs, orchestra);
			    int lastSlashPos = motifName.lastIndexOf("/");
			    String dirName = motifName.substring(0, lastSlashPos + 1);
			    tfmScanner.addMotifMatrix(motifName.substring(lastSlashPos + 1));
			    tfmScanner.findMotifs(pValMotifFinding, dirName);
			    motifScore = tfmScanner.getMotifScores()[0];
			}
		    }
		    nReadsInWindows.addElement(nReadsInWindow);
		    locVecs.addElement(locVec);
		    strandVecs.addElement(strandVec);
		    mapVecs.addElement(mapVec);
		    upstreamEnhancerDomains.addElement(upstreamEnhancerDomain);
		    downstreamEnhancerDomains.addElement(downstreamEnhancerDomain);
		    ids.addElement(id);
		    motifScores.addElement(motifScore);
		    nMismatches.addElement(nMismatch);
		    nMismatches2.addElement(nMismatch2);
		    if (conservation)
			consInWindows.addElement(consInWindow);
		    if (gc) {
			gcInWindows.addElement(gcInWindow);
			cpgInWindows.addElement(cpgInWindow);
		    }
		}
		else {
		    nReadsInWindows.addElement(null);
		    locVecs.addElement(null);
		    strandVecs.addElement(null);
		    mapVecs.addElement(null);
		    upstreamEnhancerDomains.addElement(null);
		    downstreamEnhancerDomains.addElement(null);
		    ids.addElement(null);
		    motifScores.addElement(null);
		    nMismatches.addElement(null);
		    nMismatches2.addElement(null);
		    if (conservation)
			consInWindows.addElement(null);
		    if (gc) {
			gcInWindows.addElement(null);
			cpgInWindows.addElement(null);
		    }
		}
	    }
	    //Save the scatter-stuff
	    if (!histsOnly) {
		String scatterFileName = scatterFileStart + (chr!=-1 ? "_" + chr : "") + ".txt";
		System.out.println("Saving scatter info to " + scatterFileName);
		BufferedWriter scatterBw = new BufferedWriter(new FileWriter(scatterFileName));
		if (chr==-1 || 1==chr) {
		    scatterBw.write("#chr\tpos\tstrand");
		    if (mappability)
			scatterBw.write("\tmappability");
		    if (feature.equals("TSS"))
			scatterBw.write("\tGeneID");
		    if (motifName!=null)
			scatterBw.write("\t" + motifName.substring(motifName.lastIndexOf("/") + 1));
		    if (consensusSequence!=null)
			scatterBw.write("\t" + consensusSequence);
		    if (consensus2Sequence!=null)
			scatterBw.write("\t" + consensus2Sequence);
		    for (int k=0; k<fileNames.size(); k++) {
			for (int m=0; m<windowSize/binSize; m++)
			    scatterBw.write("\t" + shortNames.elementAt(k) + (windowSize>binSize ? "_" + m : ""));
			if (upstreamEnhancerDomains.elementAt(0)!=null)
			    scatterBw.write("\t" + shortNames.elementAt(k) + "_upstream_flanking_enhancer_domain" );
			if (downstreamEnhancerDomains.elementAt(0)!=null)
			    scatterBw.write("\t" + shortNames.elementAt(k) + "_downstream_flanking_enhancer_domain" );
		    }
		    if (conservation)
			scatterBw.write("\tCons");
		    if (gc)
			scatterBw.write("\tGC\tCpG"); 
		    scatterBw.write("\n");
		}
		for (int i=1; i<=nChr; i++) {
		    if (chr==-1 || i==chr) { 
			for (int j=0; j<locVecs.elementAt(i-1).length; j++) {
			    scatterBw.write(i + "\t" + locVecs.elementAt(i-1)[j] + "\t" + (strandVecs.elementAt(i-1)[j] ? "1" : "0"));
			    if (mappability)
				scatterBw.write("\t" + format.format((double)mapVecs.elementAt(i-1)[j]/(2*(2*windowSize+1-35))));
			    if (feature.equals("TSS"))
				scatterBw.write("\t" + ids.elementAt(i-1)[j]);
			    if (motifName!=null)
				scatterBw.write("\t" + format.format(motifScores.elementAt(i-1)[j]));
			    if (consensusSequence!=null)
				scatterBw.write("\t" + nMismatches.elementAt(i-1)[j]);
			    if (consensus2Sequence!=null)
				scatterBw.write("\t" + nMismatches2.elementAt(i-1)[j]);
			    for (int k=0; k<fileNames.size(); k++) {
				for (int m=0; m<nReadsInWindows.elementAt(i-1)[k][j].length; m++)
				    scatterBw.write("\t" + format.format(nReadsInWindows.elementAt(i-1)[k][j][m]));
				if (upstreamEnhancerDomains.elementAt(i-1)!=null)
				    scatterBw.write("\t" + format.format(upstreamEnhancerDomains.elementAt(i-1)[k][j]));
				if (downstreamEnhancerDomains.elementAt(i-1)!=null)
				    scatterBw.write("\t" + format.format(downstreamEnhancerDomains.elementAt(i-1)[k][j]));
			    }
			    if (conservation) {
				for (int m=0; m<consInWindows.elementAt(i-1)[j].length; m++)
				    scatterBw.write("\t" + format.format(consInWindows.elementAt(i-1)[j][m]/(2*binSize+1)));
			    }
			    if (gc) {
				for (int m=0; m<gcInWindows.elementAt(i-1)[j].length; m++) {
				    double gcWin = gcInWindows.elementAt(i-1)[j][m]/(2*binSize+1);
				    double cpgWin = cpgInWindows.elementAt(i-1)[j][m]/(((double)binSize)*gcWin*gcWin/4);
				    scatterBw.write("\t" + format.format(gcWin) + "\t" + (Double.isNaN(cpgWin) ? "NaN" : (Double.isInfinite(cpgWin) ? "Inf" : format.format(cpgWin))));
				}
			    }
			    scatterBw.write("\n");
			}
		    }
		}
		scatterBw.close();
		//Add them all up and save to file
		for (int i=1; i<=nChr; i++) {
		    if (chr==-1 || i==chr) { 
			for (int k=0; k<fileNames.size(); k++) {
			    for (int j=0; j<histsMean[k][i - 1].length; j++) 
				sumHistMean[k][j] += histsMean[k][i - 1][j]*((double)nLocs[i-1])/sumLocs;
			}
		    }
		}
		if (gc) {
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) { 
			    for (int j=0; j<histsGCMean[i - 1].length; j++) {
				sumHistGC[j] += histsGCMean[i - 1][j]*((double)nLocs[i-1])/sumLocs;
				sumHistCpG[j] += histsCpGMean[i - 1][j]*((double)nLocs[i-1])/sumLocs;
			    }
			}
		    }
		}
		if (conservation) {
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) { 
			    //System.out.println("histsConsMean has " + histsConsMean[i - 1].length + " for " + i);
			    for (int j=0; j<histsConsMean[i - 1].length; j++) 
				sumHistCons[j] += histsConsMean[i - 1][j]*((double)nLocs[i-1])/sumLocs;
			}
		    }
		}
		if (saveFileName!=null) {
		    System.out.println("Saving profile to " + saveFileName);
		    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		    bw.write("#pos");
		    for (String fName : shortNames) 
			bw.write("\t" + fName);
		    if (conservation)
			bw.write("\tMean_cons");
		    if (gc)
			bw.write("\tMean_GC\tMeanCpG");
		    bw.write("\n");
		    int n = conservation ? sumHistCons.length : (gc ? sumHistGC.length : sumHistMean[0].length);
		    for (int i=0; i<n; i++) {
			bw.write("" + (-upstream + i));
			for (int j=0; j<fileNames.size(); j++)
			    bw.write("\t" + format.format(sumHistMean[j][i]));
			if (conservation)
			    bw.write("\t" + format.format(sumHistCons[i]));
			if (gc) 
			    bw.write("\t" + format.format(sumHistGC[i]) + "\t" + format.format((i==0 ? 0 : sumHistCpG[i]/(((sumHistGC[i-1]+sumHistGC[i])/4)*((sumHistGC[i-1]+sumHistGC[i])/4)))));
			bw.write("\n");
		    }
		    bw.close();
		}
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static int flankingEnhancerDomain(int j, int upstream, int windowSize) {
	if (upstream>0 && j-upstream<-ENHANCER_DOMAIN_FLANKING_START && j-upstream>=-(ENHANCER_DOMAIN_FLANKING_START + windowSize))
	    return -1;
	else if (upstream==0 && j>ENHANCER_DOMAIN_FLANKING_START && j<=(ENHANCER_DOMAIN_FLANKING_START + windowSize))
	    return 1;
	return 0;
    }

    public static TreeMap<Integer, Peak> createRandomPeaks(TreeMap<Integer, Peak> peaks, String subset, RefSeqGenes geneInfo) { return createRandomPeaks(peaks, subset, geneInfo, null); }

    public static TreeMap<Integer, Peak> createRandomPeaks(TreeMap<Integer, Peak> peaks, String subset, RefSeqGenes geneInfo, String saveFileName) {
	//Use this method to generate a set of random loci. If subset
	//is intergenic, then we constrain the distances to the TSSs,
	//while making sure that they do not overlap annotated genes.
	TreeMap<Integer, Peak> newPeaks = new TreeMap<Integer, Peak>();
	try {
	    System.out.println("Getting peaks from " + saveFileName);
	    BufferedReader br = new BufferedReader(new FileReader(saveFileName));
	    String line = br.readLine();
	    Pattern peakPattern = Pattern.compile("chr(\\d+):(\\d+)-?(\\d*)");
	    while (line!=null) {
		Matcher m = peakPattern.matcher(line);
		if (m.find()) {
		    newPeaks.put(Integer.parseInt(m.group(2)), new Peak(Integer.parseInt(m.group(1)), Integer.parseInt(m.group(2)), Integer.parseInt(m.group(2))));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException ex) {
	    System.err.println("Error reading random peaks " + ex.toString());
	    for (Peak p : peaks.values()) {
		
		Peak newPeak = null;
		int tssDist = 1, chr = p.getChromosome();
		RefSeqGene overlappingGene = geneInfo.getOverlappingGene(p.getCenter());
		if (subset.equals(FindReadsAcrossFeature.INTRONIC_STR)) {
		    if (overlappingGene==null)
			System.out.println("No overlapping gene for " + p.toString());
		    else
			tssDist = Math.abs(overlappingGene.getStart() - p.getCenter());		
		}
		else if (subset.equals(FindReadsAcrossFeature.INTERGENIC_STR)) {
		    if (overlappingGene!=null)
			System.out.println("Overlapping gene " + overlappingGene.toString() + " for " + p.toString());
		    else
			tssDist = Math.abs(geneInfo.getNearestTSS(p.getCenter()) - p.getCenter());
		}
		int pos = 0;
		while (newPeak==null) {
		    RefSeqGene randGene = geneInfo.getRandomGene(random);
		    if (subset.equals(FindReadsAcrossFeature.INTRONIC_STR)) {
			pos = randGene.getStart() + (randGene.getStrand() ? tssDist : - tssDist);
			if (tssDist<randGene.getLength() && !newPeaks.containsKey(pos) && pos>0 && pos<RefSeqGenes.CHROMOSOME_LENGTH[chr-1]) 
			    newPeak = new Peak(p.getChromosome(), pos, pos);
		    }
		    else if (subset.equals(FindReadsAcrossFeature.INTERGENIC_STR)) {
			pos = randGene.getStart() + (randGene.getStrand() ? -tssDist : tssDist);
			if (!geneInfo.overlapsGene(pos) && !newPeaks.containsKey(pos) && pos>0 && pos<RefSeqGenes.CHROMOSOME_LENGTH[chr-1])
			    newPeak = new Peak(p.getChromosome(), pos, pos);
		    }
		}
		//System.out.println("Created new peak " + p.toString());
		newPeaks.put(pos, newPeak);
	    }
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		for (Peak p : newPeaks.values())
		    bw.write(p.toString() + "\n");
		bw.close();
	    }
	    catch (IOException e) {
		System.err.println("Error saving random peaks " + e.toString());
	    }
	}
	return newPeaks;
    }

    public static double[][] getHistograms(int chr, int locVec[], boolean strandVec[], ABIRead reads[], ABIRead readsInput[], int downstream, int upstream, int fragmentLength, String saveHistName, double normFactor, double normFactorInput, boolean rna, double[] downSampleProbabilities, String saveReadCountDistributionFile, boolean append) throws IOException {
	double hd[][] = new double[locVec.length][FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1)];
	BufferedWriter bw = saveHistName!=null ? new BufferedWriter(new FileWriter(saveHistName)) : null, bwCountDist = saveReadCountDistributionFile!=null ? new BufferedWriter(new FileWriter(saveReadCountDistributionFile, append)) : null;
	if (saveHistName!=null)
	    System.out.println("Saving histogram to " + saveHistName);
	//if (saveReadCountDistributionFile!=null)
	    //System.out.println("Saving read counts to " + saveReadCountDistributionFile);
	if (rna) {
	    //System.out.println(Arrays.toString(downSampleProbabilities));
	    for (int i=0; i<locVec.length; i++) {
		//This is a super-ugly hack for coping with the
		//situation where we need to orient by S/AS direction. 
		int h[] = getReadsHistogram(chr, locVec[i], strandVec[i], downstream, upstream, reads, downSampleProbabilities==null ? 1 : downSampleProbabilities[i]), n = 0;
		for (int j=0; j<h.length; j++) {
		    n += h[j];
		    hd[i][j] = ((double)h[j])*normFactor;
		}
		if (saveReadCountDistributionFile!=null)
		    bwCountDist.write(chr + "\t" + locVec[i] + "\t" + (n/fragmentLength) + "\n");
	    }
	    if (saveReadCountDistributionFile!=null)
		bwCountDist.close();
	}
	else {
	    int h[][] = new int[locVec.length][FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1)], center = upstream + 1;
	    int hInput[][] = new int[locVec.length][FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1)];
	    for (int i=0; i<locVec.length; i++) {
		h[i] = FindReadsAcrossFeature.getHistogram(reads, null, chr, locVec[i], strandVec[i], downstream, upstream, fragmentLength);
		if (readsInput!=null)
		    hInput[i] = FindReadsAcrossFeature.getHistogram(readsInput, null, chr, locVec[i], strandVec[i], downstream, upstream, fragmentLength);
		for (int j=0; j<h[i].length; j++) 
		    hd[i][j] = ((double)h[i][j])*normFactor - ((double)hInput[i][j])*normFactorInput;
		//if (i==7) 
		//  System.out.println(Arrays.toString(h[i]) + "\n" + Arrays.toString(hInput[i]) + "\nnormFactor = " + format.format(normFactor) + "\tnormFactorInput = " + format.format(normFactorInput));
	    }
	}
	if (bw!=null) {
	    for (int i=0; i<hd.length; i++) {
		bw.write(chr + "\t" + locVec[i]);
		for (int j=0; j<hd[i].length; j++) 
		    bw.write("\t" + format.format(hd[i][j]));
		if (bw!=null) 
		    bw.write("\n");
	    }
	    if (bw!=null) 
		bw.close();
	}
	return hd;
    }

    public static TreeMap<String, double[]> loadDownSampleDistribution(String fileName, Vector<String> fileNames, int chr) {
	//Use this method to load the downsample distribution
	//estimated from the enhancers. This allows us to keep only a
	//handful of enhancer reads and hence determine the difference
	//in directionality, if it is significant.
	if (fileName==null)
	    return null;
	TreeMap<String, double[]> fs = new TreeMap<String, double[]>();
	for (String f : fileNames) {
	    double d[] = null;
	    int nZerosEnh = 0, nZerosDown = 0, nZerosTSS = 0;
	    try {
		//Assume that we have three numbers per line: chr\tpos\tnReads
		Vector<Integer> es = new Vector<Integer>(), ts = new Vector<Integer>();;
		String fNameEnhancer = fileName + "_divergent_" + f + ".tsv";
		String fNameTSS = "/groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/ReadDists/tss_divergent_" + f + ".tsv";
		BufferedReader brEnhancer = new BufferedReader(new FileReader(fNameEnhancer));
		BufferedReader brTSS = new BufferedReader(new FileReader(fNameTSS));
		String line = brEnhancer.readLine();
		Pattern pat = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)");
		while (line!=null) {
		    Matcher m = pat.matcher(line);
		    if (m.find()) {
			es.addElement(Integer.parseInt(m.group(3)));
			nZerosEnh += Integer.parseInt(m.group(3))==0 ? 1 : 0;
		    }
		    line = brEnhancer.readLine();
		}
		brEnhancer.close();

		line = brTSS.readLine();
		while (line!=null) {
		    Matcher m = pat.matcher(line);
		    if (m.find() && Integer.parseInt(m.group(1))==chr) {
			ts.addElement(Integer.parseInt(m.group(3)));
			nZerosTSS += Integer.parseInt(m.group(3))==0 ? 1 : 0;
		    }
		    line = brTSS.readLine();
		}
		brTSS.close();

		//Now we can calculate a retention probability for each promoter 
		//System.out.println("Caclulating downsample probabilities...");
		d = new double[ts.size()];
		int i = 0;
		random = new Random(1000); //make sure that we always have the same seed and hence probabilities
		for (Integer t : ts) {
		    int rand = random.nextInt(es.size()), r = Integer.MAX_VALUE;
		    while (r>t)
			r = Math.min(t, es.elementAt(rand));
		    //Proportion of reads to keep
		    d[i++] = t==0 ? 0 : (double)r/t;
		    if (Double.isNaN(d[i-1])) {
			System.out.println("Got NaN for randInd = " + rand + ", t = " + t + " and r = " + r);
			d[i-1] = Double.isNaN(d[i-1]) ? 0 : d[i-1];
		    }
		    else if (r==0)
			nZerosDown++;
		}
		fs.put(f, d);
		//System.out.println(format.format((double)nZerosEnh/es.size()) + "\t" + format.format((double)nZerosDown/ts.size()) + "\t" + format.format((double)nZerosTSS/ts.size()));
	    }
	    catch (IOException e) {
		System.err.println("Error opening distribution file: " + e.toString());
	    }
	}
	return fs;
    }

    public static int[] getReadsHistogram(int chr, int tss, boolean forward, int downstream, int upstream, ABIRead reads[], double downSampleProbability) {
	//Use this method for finding RNA-reads in a given region.
	int fragmentLength = FindNongenicTranscripts.MRNA_READ_LENGTH;
	int start = tss - upstream - fragmentLength;
	int end = tss + downstream + fragmentLength;
	int indStart = Arrays.binarySearch(reads, new ABIRead(chr, true, start), null);
	int indEnd = Arrays.binarySearch(reads, new ABIRead(chr, true, end), null);
	indStart = indStart<0 ? -indStart-1 : indStart;
	indEnd = indEnd<0 ? -indEnd-1 : indEnd;
	int nReads = indEnd - indStart, ind = 0, inds[] = new int[nReads];
	if (downSampleProbability<1) { //Extract the subset of reads
	    int r = (int)Math.round(downSampleProbability*(double)nReads);
	    inds = new int[r];
	    if (r>0) {
		inds[0] = random.nextInt(nReads);
		for (int i=1; i<r; i++) {
		    int randInd = random.nextInt(nReads);
		    while (Arrays.binarySearch(inds, randInd)>=0) 
			randInd = random.nextInt(nReads);
		    inds[i] = randInd;
		}
		Arrays.sort(inds);
	    }
	}
	else {
	    for (int i=0; i<nReads; i++)
		inds[i] = i;
	}
	ABIRead[] rs = new ABIRead[inds.length];
	for (int i=0; i<inds.length; i++)
	    rs[i] = reads[indStart + inds[i]];
	int hist[] = new int[FindReadsAcrossFeature.numberOfBins(downstream, upstream)];
	while (ind<rs.length && rs[ind].getPosition()<end) {
	    nReads++;
	    for (int i=0; i<fragmentLength; i++) {
		if (rs[ind].getPosition() + i>=start + fragmentLength && rs[ind].getPosition() + i<=end - fragmentLength) {
		    hist[-tss + upstream + rs[ind].getPosition() + i]++;
		}
	    }
	    ind++;
	}
	if (!forward) 
	    hist = FindReadsAcrossFeature.flipArray(hist);
	return hist;
    }

    public static double[][] getRNAHistograms(int chr, int locVec[], boolean strandVec[], boolean sense, ABIRead readsForward[], ABIRead readsReverse[], int downstream, int upstream, int binSize, int fragmentLength, double normFactor)  {
	double hd[][] = new double[locVec.length][FindReadsAcrossFeature.numberOfBins(upstream, downstream, binSize)];
	for (int i=0; i<locVec.length; i++) {
	    ABIRead reads[] = strandVec[i] ? (sense ? readsForward : readsReverse) : (sense ? readsReverse : readsForward);
	    int h[] = getReadsHistogram(chr, locVec[i], strandVec[i], downstream, upstream, reads, 1.0);
	    //System.out.println("Found hist at " + locVec[i] + " with " + strandVec[i]);
	    for (int j=0; j<h.length; j++) {
		if (h[j/binSize]!=Integer.MAX_VALUE) {
		    hd[i][j/binSize] += ((double)h[j])*normFactor/binSize;
		}
		else
		    hd[i][j/binSize] = Integer.MAX_VALUE;
	    }
	}
	return hd;
    }

    public static double[] findMeanReadHistograms(double hists[][]) { 
	if (hists.length>0) {
	    int norm = hists.length;
	    double hist[] = new double[hists[0].length];
	    for (int j=0; j<hists.length; j++) {
		double h[] = hists[j];
		for (int i=0; i<h.length; i++)
		    hist[i] += (double)h[i]/norm;
	    }
	    return hist;
	}
	return new double[0];
    }

    public static double[] getConservationHistogram(double pcs[][]) {
	double consHist[] = null;
	if (pcs!=null && pcs.length>0) {
	    int n[] = new int[pcs[0].length];
	    /*for (int j=0; j<pcs.length; j++) {
		for (int i=0; i<pcs[j].length; i++) 
		    n[i] += pcs[j][i]>=0 ? 1 : 0;
		    }*/
	    consHist = new double[pcs[0].length];
	    for (int j=0; j<pcs.length; j++) {
		for (int i=0; i<pcs[j].length; i++) {
		    //if (pcs[j][i]>HIGHLY_CONSERVED_THRESHOLD)
		    //consHist[i] += n[i]>0 ? 1.0/n[i] : 0;
		    consHist[i] += pcs[j][i]/pcs.length;
		}
	    }
	}
	else
	    consHist = new double[0];
	return consHist;
    }

    public static Vector<int[]> parseSelectPeakFile(String selectPeaksFileName) { return parseSelectPeakFile(selectPeaksFileName, 21); }

    public static Vector<int[]> parseSelectPeakFile(String selectPeaksFileName, int nChr) {
	if (selectPeaksFileName==null)
	    return null;
	Vector<int[]> selPeaks = new Vector<int[]>();
	Vector<Vector<Integer>> selectPeaks = new Vector<Vector<Integer>>(nChr);
	for (int i=0; i<nChr; i++)
	    selectPeaks.addElement(new Vector<Integer>());
	try {
	    BufferedReader br = new BufferedReader(new FileReader(selectPeaksFileName));
	    String line = br.readLine();
	    Pattern selectPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\w+)");
	    int nFound = 0;
	    while (line!=null) {
		Matcher m = selectPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    int pos = Integer.parseInt(m.group(2));
		    if (m.group(3).equals("TRUE")) {
			selectPeaks.elementAt(chr - 1).addElement(pos);
			nFound++;
		    }
		    else if (m.group(3).equals("FALSE") && selectPeaks.elementAt(chr - 1).contains(pos)) { //remove peak if it is already in the list
			selectPeaks.elementAt(chr - 1).remove(pos);
			nFound--;
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    System.out.println("Found " + nFound + " peaks in " + selectPeaksFileName);
	    for (int i=0; i<nChr; i++) {
		int rs[] = new int[selectPeaks.elementAt(i).size()];
		for (int j=0; j<rs.length; j++)
		    rs[j] = selectPeaks.elementAt(i).elementAt(j);
		Arrays.sort(rs);
		selPeaks.addElement(rs);
	    }
	}
	catch (IOException e) {
	    System.err.println("Error parsing peaks file: " + e.toString());
	}
	return selPeaks;
    }

    public static int[] parseFeatureFile(String featureName, int c) { return parseFeatureFile(featureName, c, false); }

    public static int[] parseFeatureFile(String featureName, int c, boolean startAndEnd) {
	int locs[] = null;
	//int outliers[] = getOutliers(c);
	try {
	    Vector<Integer> features = new Vector<Integer>();
	    BufferedReader br = new BufferedReader(new FileReader(featureName));
	    String line = br.readLine();
	    Pattern selectPattern = startAndEnd ? Pattern.compile("^(\\d+)\\s+(\\d+)\\s+(\\d+)") : Pattern.compile("^(\\d+)\\s+(\\d+)");
	    int nFound = 0;
	    while (line!=null) {
		if (!line.startsWith("#")) {
		    Matcher m = selectPattern.matcher(line);
		    if (m.find()) {
			try {
			    int chr = Integer.parseInt(m.group(1));
			    int pos = startAndEnd ? (Integer.parseInt(m.group(2)) + Integer.parseInt(m.group(3)))/2 : Integer.parseInt(m.group(2));
			    if (chr==c/* && Arrays.binarySearch(outliers, pos)<0*/)
				features.addElement(pos);
			}
			catch (java.lang.IndexOutOfBoundsException ex) {
			    selectPattern = Pattern.compile("^(\\d+)\\s+(\\d+)");
			    m = selectPattern.matcher(line);
			    if (m.find()) {
				int chr = Integer.parseInt(m.group(1));
				int pos = Integer.parseInt(m.group(2));
				if (chr==c/* && Arrays.binarySearch(outliers, pos)<0*/)
				    features.addElement(pos);
			    }
			    else
				System.out.println("Could not find match for pattern on line: " + line);
			}
		    }
		    else if (startAndEnd) {
			selectPattern = Pattern.compile("^(\\d+)\\s+(\\d+)");
			m = selectPattern.matcher(line);
			if (m.find()) {
			    int chr = Integer.parseInt(m.group(1));
			    int pos = Integer.parseInt(m.group(2));
			    if (chr==c/* && Arrays.binarySearch(outliers, pos)<0*/)
				features.addElement(pos);
			}
			else
			    System.out.println("Could not find match for pattern on line: " + line);
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    locs = new int[features.size()];
	    for (int j=0; j<locs.length; j++)
		locs[j] = features.elementAt(j);
	    Arrays.sort(locs);
	    System.out.println("Found " + locs.length + " features for chr " + c);
	    //System.out.println(Arrays.toString(locs));
	}
	catch (IOException e) {
	    System.err.println("Error parsing peaks file: " + e.toString());
	}
	return locs;
    }

    public static boolean[] parseFeatureFileStrandInfo(String featureName, int c, int[] locs) {
	//use this method to extract the strand info for the feature files
	boolean strands[] = new boolean[locs.length];
	Arrays.fill(strands, true);
	try {
	    Vector<Integer> features = new Vector<Integer>();
	    BufferedReader br = new BufferedReader(new FileReader(featureName));
	    String line = br.readLine();
	    Pattern selectPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([TRUE|FALSE]+)");
	    int nFound = 0;
	    while (line!=null) {
		Matcher m = selectPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    int pos = Integer.parseInt(m.group(2));
		    boolean strand = m.group(3).equals("TRUE");
		    if (chr==c) {
			int ind = Arrays.binarySearch(locs, pos);
			if (ind>-1)
			    strands[ind] = strand;
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing peaks file: " + e.toString());
	}
	return strands;
    }

    ////////////////////////////////////////////

    public static int[] parseBedFile(String fileName, int chr, int relLoc) {
	//relLoc determines where we will be centering on: -1 for the 5', 0 for the middle and 1 for the 3'-end.
	int locs[]  = null;
	try {
	    Vector<Integer> ls = new Vector<Integer>();
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern bedPattern = Pattern.compile("chr(\\w+):(\\d+)-(\\d+)\\s+(-*?)");//([\\d|\\.]+)");
	    while (line!=null) {
		Matcher m = bedPattern.matcher(line);
		if (m.find()) {
		    if (Integer.parseInt(m.group(1))==chr) {
			boolean strand = !m.group(4).equals("-");
			if ((relLoc<0 && strand) || (relLoc>0 && !strand))
			    ls.addElement(Integer.parseInt(m.group(strand ? 2 : 3)));
			else if ((relLoc>0 && strand) || (relLoc<0 && !strand))
			    ls.addElement(Integer.parseInt(m.group(strand ? 3 : 2)));
			else
			    ls.addElement((Integer.parseInt(m.group(2)) + Integer.parseInt(m.group(3)))/2);
		    }
		}
		else {
		    System.out.println("Failed parsing " + line);
		    break;
		}
		line = br.readLine();
	    }
	    br.close();
	    locs = new int[ls.size()];
	    for (int i=0; i<ls.size(); i++)
		locs[i] = ls.elementAt(i);
	    System.out.println("Found " + locs.length + " bed-locations for chr " + chr);
	}
	catch (IOException e) {
	    System.err.println("Error parsing bed file: " + e.toString());
	}
	return locs;
    }

    public static boolean[] parseBedFileStrandInfo(String fileName, int chr) {
	boolean strands[]  = null;
	try {
	    Vector<Boolean> ss = new Vector<Boolean>();
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern bedPattern = Pattern.compile("chr(\\w+):(\\d+)-(\\d+)\\s+(-*?)");//([\\d|-|\\.]+)");
	    while (line!=null) {
		Matcher m = bedPattern.matcher(line);
		if (m.find()) {
		    if (Integer.parseInt(m.group(1))==chr) 
			ss.addElement(!m.group(4).equals("-"));
		}
		line = br.readLine();
	    }
	    br.close();
	    strands = new boolean[ss.size()];
	    for (int i=0; i<ss.size(); i++)
		strands[i] = ss.elementAt(i);
	    System.out.println("Found " + strands.length + " strands for chr " + chr);
	}
	catch (IOException e) {
	    System.err.println("Error parsing bed file: " + e.toString());
	}
	return strands;
    }

    ////////////////////////////////////////////

    public static int[] getOutliers(int c){
	//Use this method to get a list of outliers that should be
	//excluded from further analysis.
	int outliers[] = null, i = 0;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(OUTLIER_FILE_LOCATION));
	    Vector<Integer> arr = new Vector<Integer>();
	    String line = br.readLine();
	    Pattern pattern = Pattern.compile("^(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = pattern.matcher(line);
		if (m.find()) {
		    if (Integer.parseInt(m.group(1))==c)
			arr.addElement(Integer.parseInt(m.group(2)));
		}
		line = br.readLine();
	    }
	    br.close();
	    outliers = new int[arr.size()];
	    for (Integer a : arr)
		outliers[i++] = a;
	    Arrays.sort(outliers);
	}
	catch (IOException e) {
	    System.err.println("Error reading outliers " + e.toString());
	}
	return outliers;
    }

    public static int calculateTotalReads(String fileName, String inputFileName, int nChr, int fragmentLength, double qualityScore) { 
	Vector<String> tmp = new Vector<String>();
	tmp.addElement(fileName);
	return calculateTotalReads(tmp, inputFileName, nChr, fragmentLength, qualityScore).values().iterator().next();
    }

    public static TreeMap<String, Integer> calculateTotalReads(Vector<String> fileNames, String inputFileName, int nChr, int fragmentLength, double qualityScore) {
	//Try opening the file where the stats have been
	//pre-computed. It is located in /home/mh168/Projects/ChIP-Seq/TKData/MmData_read_stats.txt with columns:
	//#Experiment     fragment_length Uniquely mappable       QS filtered     PCR filtered    Mismatches0     Mismatches1     Mismatches2     Mismatches3
	String statsFileName = "/home/mh168/Projects/ChIP-Seq/TKData/MmData_read_stats.txt";
	TreeMap<String, Integer> nTotalReads = new TreeMap<String, Integer>(), nReads = new TreeMap<String, Integer>();;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(statsFileName));
	    String line = br.readLine();
	    Pattern statsPattern = Pattern.compile("^([\\w|/|\\-|_]+)\\s+(\\d+)\\s+\\d+\\s+\\d+\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = statsPattern.matcher(line);
		if (m.find() && Integer.parseInt(m.group(2))==fragmentLength) {
		    nTotalReads.put(m.group(1), Integer.parseInt(m.group(3)));
		    //System.out.println("Found " + m.group(3) + " reads for " + m.group(1));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error opening stats file " + statsFileName);
	}
	if (inputFileName!=null) {
	    if (nTotalReads.containsKey(inputFileName.substring(0, inputFileName.lastIndexOf("/"))))
		nReads.put(inputFileName, nTotalReads.get(inputFileName.substring(0, inputFileName.lastIndexOf("/"))));
	    else {
		if (inputFileName.endsWith(".bed")) {
		    nTotalReads.put(inputFileName, CreateWigFiles.getTotalNumberOfReads(inputFileName, nChr, fragmentLength));
		    nReads.put(inputFileName, nTotalReads.get(inputFileName)); //Ugly...
		}
		else {
		    for (int i=1; i<=nChr; i++) { //For normalization purposes, we want to consider all chromosomes
			ABIRead reads[] = PeakFinder2.parseFiles(i, inputFileName, fragmentLength, true, 0, qualityScore);
			nTotalReads.put(inputFileName, nTotalReads.get(inputFileName) + reads.length);
		    }
		}
	    }
	}
	for (String fName : fileNames) {
	    if (fName.endsWith(".bed")) {
		nReads.put(fName, CreateWigFiles.getTotalNumberOfReads(fName, nChr, fragmentLength));
		nTotalReads.put(fName, nReads.get(fName)); //Ugly.....
	    }
	    else if (nTotalReads.containsKey(fName.substring(0, fName.lastIndexOf("/")))) {
		nReads.put(fName, nTotalReads.get(fName.substring(0, fName.lastIndexOf("/"))));
		System.out.println("From file, found" + nReads.get(fName) + " reads for " + fName);
	    }
	    else 
		nReads.put(fName, countReads(fName, nChr, fragmentLength, qualityScore));
	}
	return nReads;
    }

    public static int countRNAReads(String fName, int nChr) { return countReads(fName, nChr, 35, 0); }

    public static int countReads(String fName, int nChr, int fragmentLength, double qualityScore) {
	int nReads = 0;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fName + "_total_reads_f.txt"));
	    String line = br.readLine();
	    while (line!=null) {
		nReads += Integer.parseInt(line); //just one number per line
		line = br.readLine();
	    }
	    br.close();
	    br = new BufferedReader(new FileReader(fName + "_total_reads_r.txt"));
	    line = br.readLine();
	    while (line!=null) {
		nReads += Integer.parseInt(line); //just one number per line
		line = br.readLine();
	    }
	    br.close();
	    System.out.println("From file, found " + nReads + " reads for " + fName);
	}
	catch (IOException e) {
	    System.err.println("Error reading stats " + e.toString());
	    nReads = 0;
	    if (fName.contains("MikkelsenData")) {
		for (int i=1; i<=nChr; i++) //For normalization purposes, we want to consider all chromosomes
		    nReads += PeakFinder2.parseBroadFiles(i, fName, fragmentLength, true).length;
	    }
	    else if (fName.endsWith("bed")) {
		for (int i=1; i<=nChr; i++) //For normalization purposes, we want to consider all chromosomes
		    nReads += CreateWigFiles.parseBedFiles(i, fName, true, fragmentLength).length;
	    }
	    else {
		for (int i=1; i<=nChr; i++) //For normalization purposes, we want to consider all chromosomes
		    nReads += !fName.contains("RNA") ? PeakFinder2.parseFiles(i, fName, fragmentLength, true, 0, qualityScore).length : ABIReads.loadMRNAReadsChr(fName, true, i, 35, false, null).length + ABIReads.loadMRNAReadsChr(fName, false, i, 35, false, null).length;
	    }
	    System.out.println("Counting, found " + nReads + " reads for " + fName);
	}
	return nReads;
    }

    public static double[][] findConservationHistogram(int i, boolean orchestra, String species, int locVec[], boolean[] strandVec, int downstream, int upstream) {
	PhastConsParser verterbateCons = new PhastConsParser(i, true, orchestra, species);
	double[][] pcs = new double[locVec.length][FindReadsAcrossFeature.numberOfBins(downstream, upstream, 1)];
	for (int j=0; j<locVec.length; j++) {
	    pcs[j] = strandVec[j] ? verterbateCons.findScores(locVec[j] - upstream, locVec[j] + downstream + 1) : FindReadsAcrossFeature.flipArray(verterbateCons.findScores(locVec[j] - downstream, locVec[j] + upstream + 1));
	}
	return pcs;
	//return FindReadsAcrossFeature.getConservationHistogram(pcs);
    }

    public static String[] getSequences(int chr, boolean orchestra, String species, int tssVec[], boolean strandVec[], int upstream, int downstream) {
	FastaSequence2 fastaSeq = new FastaSequence2(species, false, chr, orchestra);
	int tsss[] = new int[tssVec.length];
	for (int i=0; i<tssVec.length; i++) {
	    if (strandVec[i])
		tsss[i] = tssVec[i] - upstream;
	    else
		tsss[i] = tssVec[i] - downstream;
	}
	String seqs[] = fastaSeq.getSequences(tsss, upstream + downstream + 1);
	for (int j=0; j<seqs.length; j++) {
	    if (seqs[j]!=null && !strandVec[j]) 
		seqs[j] = new String(FindReadsAcrossFeature.flipArray(seqs[j].toCharArray()));
	    //System.out.println(seqs[j].substring(0, 2));
	}
	return seqs;
    }
    
    public static double[][] getGCHistogram(String seqs[]) {
	double gcHist[][] = new double[seqs.length][seqs.length>0 ? seqs[0].length() : 0];
	for (int j=0; j<seqs.length; j++) {
	    if (seqs[j]!=null) {
		char seq[] = seqs[j].toCharArray();
		for (int i=0; i<seq.length; i++) {
		    if (seq[i]=='c' || seq[i]=='C' || seq[i]=='g' || seq[i]=='G') 
			gcHist[j][i] = 1.0;
		}
	    }
	}
	return gcHist;
    }
    
    public static double[][] getCpGHistogram(String seqs[]) {
	double cpgHist[][] = new double[seqs.length][seqs.length>0 ? seqs[0].length() : 0];
	for (int j=0; j<seqs.length; j++) {
	    if (seqs[j]!=null) {
		char seq[] = seqs[j].toCharArray();
		for (int i=1; i<seq.length; i++) {
		    if (((seq[i-1]=='c' || seq[i-1]=='C') && (seq[i]=='g' || seq[i]=='G'))) 
			cpgHist[j][i] = 1.0;
		}
	    }
	}
	return cpgHist;
    }

    public static double[] findGCHistogram(int i, boolean orchestra, int locVec[], boolean[] strandVec, int downstream, int upstream) {
	return FindReadsAcrossFeature.getGCHistogram(i, orchestra, locVec, strandVec, upstream, downstream);
    }

    public static double[] findCpGHistogram(int i, boolean orchestra, int locVec[], boolean[] strandVec, int downstream, int upstream) {
	return FindReadsAcrossFeature.getCpGHistogram(i, orchestra, locVec, strandVec, upstream, downstream);
    }

    public static boolean[] getStrandVec(String subset, RefSeqGenes geneInfo, int[] locVec) {
	boolean[] strandVec = new boolean[locVec.length];
	if (subset.equals(FindReadsAcrossFeature.INTERGENIC_STR))
	    Arrays.fill(strandVec, true);
	else {
	    for (int i=0; i<locVec.length; i++) {
		RefSeqGene g = geneInfo.getOverlappingGene(locVec[i]);
		if (g!=null)
		    strandVec[i] = g.getStrand();
		else
		    System.err.println("Error finding gene overlapping " + locVec);
	    }
	}
	return strandVec;
    }

    public static boolean strandOK(int senseOrient, boolean forward, boolean strand) {
	//use this method to determine if a loci is to be included for
	//a given combination of stranded-ness and direction. Not the
	//most good-looking solution
	if (senseOrient>0) 
	    return forward==strand;
	else 
	    return forward!=strand;
    }
    
    public static int[] senseFilterLoci(int senseOrient, boolean forward, int[] locVec, boolean[] strandVec, RefSeqGenes geneInfo) {
	//Use this method to remove loci that overlap with genes and have the wrong strand orientation
	int nKeep = 0, k = 0;
	for (int j=0; j<strandVec.length; j++)
	    nKeep += strandOK(senseOrient, forward, strandVec[j]) ? 1 : 0;
	int newLocVec[] = new int[nKeep];
	for (int j=0; j<locVec.length; j++) {
	    if (strandOK(senseOrient, forward, strandVec[j]))
		newLocVec[k++] = locVec[j];
	}
	return newLocVec;
    }

    public static int bestNumberOfMismatches(String seq, String consensusSequence) {
	//Find out how many mismatches we are away from a perfect
	//match between a substring of seq and the motif
	//consensusSequence.
	int bestMismatches = consensusSequence.length(), nMismatches[] = numberOfMismatches(seq, consensusSequence), bestInd = -1;
	for (int i=0; i<nMismatches.length; i++) {
	    if (nMismatches[i]<bestMismatches) {
		bestMismatches = nMismatches[i];
		bestInd = i;
	    }
	}
	//System.out.println("Best mismatch for " + seq + " is " + bestMismatches + " at " + bestInd);
	//System.out.println(bestMismatches + ":" + Arrays.toString(nMismatches));
	return bestMismatches;
    }

    public static int[] numberOfMismatches(String seq, String consensusSequence) {
	//Find out how many mismatches we are away from a perfect
	//match between a substring of seq and the motif
	//consensusSequence.
	if (seq!=null) {
	    int len = consensusSequence.length(), nMismatches[] = new int[seq.length() - len];
	    String consensusSequenceReverseComplement = EnumerateMers.reverseComplement(consensusSequence);
	    for (int i=0; i<seq.length()-len; i++) {
		String m = seq.substring(i, i + len);
		nMismatches[i] = Math.min(motifDistance(m, consensusSequence), motifDistance(m, consensusSequenceReverseComplement));
		//if (i % 20 == 0)
		//	System.out.println("Distance btwn " + m + " and " + consensusSequence + " is " + motifDistance(m, consensusSequence) + " and btwn " + m + " and " + consensusSequenceReverseComplement + " is " + motifDistance(m, consensusSequenceReverseComplement));
	    }
	    return nMismatches;
	}
	return new int[0];
    }

    public static int motifDistance(String m1, String m2) {
	int d = 0;
	if (m1.length()==m2.length()) {
	    for (int i=0; i<m1.length(); i++) {
		if (!m2.substring(i, i+1).equalsIgnoreCase("N") && !m1.substring(i, i+1).equalsIgnoreCase("N"))
		    d += m1.substring(i, i+1).equalsIgnoreCase(m2.substring(i, i+1)) ? 0 : 1;
	    }
	    return d;
	}
	return -1;
    }

}