//use to detect transcripts using the wavelet based method. The
//current version only contains three parameters: two length scales
//and a cut-off for the number of break points and what densities to
//include. The last parameter is similar to an FDR type cut-off and to
//sweep for different values, we may use:
//
//Version 3 is a slimmed down version of FindTranscripts2.java and some
//stuff that is no longer used has been removed, including the
//Bayesian method and everything that was used for ChIP experiments.

//java -Xmx4096m FindTranscripts3  -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr6c_b1/hr6c_b1 -b 8 -l 8 -L 20 -R 35 --merge --minDens .01 --fdr .005 -o --breakPointsFileName ../MmDataPeakFiles/Pol2_8WG16_KCl_B1_names --cost peaks --peakFileName ../MmDataPeakFiles/H3K4Me3_KCl_peaks_names --gmm 3 --log 1 --merge -c 19

//bsub -q shared_15m java -server -Xmx5096m FindTranscripts3 -o --breakPointsFileName ../MmDataPeakFiles/Pol2_8WG16_KCl_B1_names --peakFileName ../MmDataPeakFiles/H3K4Me3_KCl_peaks_names -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/merged_total/merged_total -b 8 --fdr .005 --merge --cost peaks --log 1 --gmm 2 -R 35 --minDens .001 -l 10 -L 20 -c 19

//java -Xmx4096m FindTranscripts3  -n /groups/neuroduo/martin/ChIP-Seq/HeLa/mRNA/HeLac -b 8 -l 8 -L 20 -R 35 --merge --minDens .001 --fdr .0005 -o --breakPointsFileName ../../Encode/PeakFiles/wgEncodeYaleChIPseqPeaksHelas3Pol2V2.narrowPeak --cost peaks --peakFileName ../../Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak --gmm 3 --log 1 --merge -c 19 --species hs

//java -Xmx4096m -server FindTranscripts3  -n /groups/neuroduo/martin/ChIP-Seq/GrimmondData/Hs/mRNA/lung/lung -b 8 -l 9 -L 20 -R 35 --merge --minDens .001 --fdr .01 -o --cost genes --gmm 3 --log 1 --merge -c 22 --species hs --hg19

//java -Xmx4096m FindTranscripts3  -n /groups/neuroduo/martin/ChIP-Seq/KleinsteinData/mRNA/ETX-FTX.txt -b 8 -l 9 -L 20 -R 35 --merge --minDens .001 --fdr .001 -o --cost genes --gmm 3 --log 1 --merge -c 19 --masterTable /groups/neuroduo/martin/ChIP-Seq/KleinsteinData/mRNA/mastertable_fasta_20100610_v0k1m1.sam

//For the down-sampling analysis in response to the reviewers:
//
//java -server -Xmx5096m FindTranscripts3 -o --breakPointsFileName ../MmDataPeakFiles/Pol2_8WG16_KCl_B1_names --peakFileName ../MmDataPeakFiles/H3K4Me3_KCl_peaks_names -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/merged_total/merged_total -b 8 --fdr .005 --merge --cost peaks --log 1 --gmm 2 -R 35 --minDens .001 -l 10 -L 20 -c 19 --downSample 10 -s test

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.TallyStore;

public class FindTranscripts3 {

    public static final String BED_FILE_SUFFIX = "_transcripts";
    public static final String RANDOM_READS_FILE_SUFFIX = "_random_reads_transcripts";
    public static final String DOWNSAMPLING_FILE_SUFFIX = "_downsampling_transcripts";
    public static final String LATEX_TABLE_FILE_NAME_SUFFIX = "_summary_stats.tex";
    public static final String TRANSCRIPT_LENGTHS_SUFFIX = "_lengths";
    public static final String TRANSCRIPT_DENSITIES_SUFFIX = "_dens";
    public static final String TRANSCRIPT_WITHIN_DENISTY_STD_DEV_SUFFIX = "_densStdDev";
    public static final String TRANSCRIPT_PEAKS_SUFFIX = "_peaks";
    public static final String TRANSCRIPT_CONSERVATION_SUFFIX = "_cons";
    public static final String TRANSCRIPT_WITHIN_CONSERVATION_STD_DEV_SUFFIX = "_consStdDev";
    public static final String NON_CODING_AS_INFO_SUFFIX = "Noncoding_AS_info";
    
    public static final String WWW_DIR_NAME = "neuroduo.hms.harvard.edu/docroot/bigWig/mh/Transcripts/";
    public static final String URL_NAME = "http://neuroduo.hms.harvard.edu/bigWig/mh/Transcripts/";
    public static final String TRACK_INFO_FILE_PATH = "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/Transcripts/";
    public static final String TRACK_INFO_FILE_NAME = "track_info.txt";
    public static final String BREAK_POINTS_FILE_NAME = "breakPoints.txt";
    public static final String RANDOM_DIR_NAME = "Random/";
    public static final String SEQUENCE_DIR_NAME = "Sequences/";
    public static final String AA_DIR_NAME = "AA/";

    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option readLengthFlag = argParser.addIntegerOption('R', "readLength");	    
	    CmdLineParser.Option strandFlag = argParser.addBooleanOption('r', "reverseStrand");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option logFlag = argParser.addIntegerOption("log");
	    CmdLineParser.Option gmmFlag = argParser.addIntegerOption("gmm");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption("conservation");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option sequenceFlag = argParser.addBooleanOption("sequence");
	    CmdLineParser.Option cpcFlag = argParser.addBooleanOption("CPC");
	    CmdLineParser.Option aaFlag = argParser.addIntegerOption("AA");
	    //Parameters for the transcript finding
	    CmdLineParser.Option lMinFlag = argParser.addIntegerOption('l', "lMin");
	    CmdLineParser.Option lMaxFlag = argParser.addIntegerOption('L', "lMax");
	    //For coarse-graining
	    CmdLineParser.Option binSizeFlag = argParser.addIntegerOption('b', "binSize");
	    CmdLineParser.Option deltaLFlag = argParser.addIntegerOption("deltaL");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption("fdr");
	    CmdLineParser.Option mergeFlag = argParser.addBooleanOption("merge");
	    CmdLineParser.Option minDensFlag = argParser.addDoubleOption("minDens");
	    CmdLineParser.Option costFunctionFlag = argParser.addStringOption("cost");
	    //Use these flags when we want to down-sample the total number of reads to investigate how the optimal parameters change
	    CmdLineParser.Option downSampleFlag = argParser.addIntegerOption("downSample");
	    //Use this to indicate a peak file, eg Pol2 or H3K4Me3
	    //that we wish to compare with for the optimization
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption("peakFileName");
	    //Use this flag to provide a set of peaks that are to be
	    //used for spiking the set of breakpoints, eg Pol2 peak
	    //locations
	    CmdLineParser.Option breakPointsFileNameFlag = argParser.addStringOption("breakPointsFileName");
	    //Use this flag to save information about intermediate steps of the algorithm
	    CmdLineParser.Option saveInfoFileNameFlag = argParser.addBooleanOption("saveInfo");
	    //Use this flag to create a set of randomized regions with
	    //the same lengths as the ones detected by the algorithm
	    CmdLineParser.Option randomFlag = argParser.addIntegerOption("random");
	    //Use this flag when we want to randomize the location of
	    //the reads and the breakpoints instead to determine the
	    //background detection level.
	    CmdLineParser.Option randomReadsFlag = argParser.addIntegerOption("randomReads");
	    //For comparing against known elements
	    CmdLineParser.Option maxDivergentTSSDistanceFlag = argParser.addIntegerOption("maxDivergent");
	    //Indicate the use of hg19 (v37) instead of the default hg18
	    CmdLineParser.Option hg19Flag = argParser.addBooleanOption("hg19");
	    //Flags to use when we parse the Yale data from the Kleinstein lab
	    CmdLineParser.Option masterTableFileNameFlag = argParser.addStringOption("masterTable");
	    CmdLineParser.Option fxtFlag = argParser.addBooleanOption("FXT");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    boolean sequence = argParser.getOptionValue(sequenceFlag)!=null; 
	    boolean cpc = argParser.getOptionValue(cpcFlag)!=null; 
	    boolean hg19 = argParser.getOptionValue(hg19Flag)!=null; 
	    tmp = argParser.getOptionValue(aaFlag);
	    int minAAs = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(randomFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(randomReadsFlag);
	    int nReadsMC = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    boolean rand = nMC>0; 
	    tmp = argParser.getOptionValue(gmmFlag);
	    int nComponents = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    boolean gmm = nComponents>1;
	    boolean reverseStrand = argParser.getOptionValue(strandFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp + (((String)tmp).contains(".") ? "" : ".tsv") : "";
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(breakPointsFileNameFlag);
	    String breakPointsFileName = tmp!=null ? (String)tmp : null;
	    boolean saveInfo = argParser.getOptionValue(saveInfoFileNameFlag)!=null;
	    String saveInfoFileName = saveInfo ? fileName.substring(0, fileName.lastIndexOf(".")) + (reverseStrand ? "_r" : "_f") : null;
	    String latexTableFileName = saveInfoFileName + LATEX_TABLE_FILE_NAME_SUFFIX;
	    tmp = argParser.getOptionValue(costFunctionFlag);
	    String costFunction = tmp!=null ? ((String)tmp).toLowerCase() : "sens";
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : fileName.substring(0, fileName.lastIndexOf("."));
	    String saveFileNameRandom = saveFileName.substring(0, saveFileName.lastIndexOf("/") + 1) + RANDOM_DIR_NAME + saveFileName.substring(saveFileName.lastIndexOf("/") + 1);
	    tmp = argParser.getOptionValue(logFlag);
	    int nLogs = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    boolean log = nLogs>0;
	    tmp = argParser.getOptionValue(readLengthFlag);
	    int readLength = tmp!=null ? ((Integer)tmp).intValue() : 35;
	    tmp = argParser.getOptionValue(lMinFlag);
	    int lMin = tmp!=null ? ((Integer)tmp).intValue() : 5;
	    tmp = argParser.getOptionValue(lMaxFlag);
	    int lMax = tmp!=null ? ((Integer)tmp).intValue() : 15;
	    tmp = argParser.getOptionValue(binSizeFlag);
	    int binSize = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(deltaLFlag);
	    int deltaL = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .01;
	    tmp = argParser.getOptionValue(downSampleFlag);
	    int nDownSample = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    boolean promoterLocationCategories = true;
	    tmp = argParser.getOptionValue(maxDivergentTSSDistanceFlag);
	    int maxDivergentTSSDistance = tmp!=null ? ((Integer)tmp).intValue() : promoterLocationCategories ? 10000 : 2000;
	    boolean mergeSegments = argParser.getOptionValue(mergeFlag)!=null;
	    tmp = argParser.getOptionValue(minDensFlag);
	    double minDens = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    tmp = argParser.getOptionValue(masterTableFileNameFlag);
	    String masterTableFileName = tmp!=null ? (String)tmp : null;
	    boolean kleinsteinFiles = masterTableFileName!=null;
	    boolean ext = argParser.getOptionValue(fxtFlag)!=null; 
	    //////////////////////////////////////////////////////////////////
	    int nChr = numberOfChromosomes(species) - 1;
	    int lVecCoeffs[] = getLengthVec(lMin, lMax, deltaL);
	    Pattern rnaFileNamePattern = Pattern.compile("hr(\\d)\\w");
	    Matcher matcher = rnaFileNamePattern.matcher(fileName);
	    int time = matcher.find() ? Integer.parseInt(matcher.group(1)) : 6;
	    RefSeqUCSCEnsemblGenes geneInfo[] = new RefSeqUCSCEnsemblGenes[nChr];
	    TranscriptFinder2 tf[] = new TranscriptFinder2[nChr];
	    ReducedModel2 reducedModel[] = new ReducedModel2[nChr];
	    ReducedModel2 reducedModelRandom[] = new ReducedModel2[nChr];
	    String repeatFileName = species.toLowerCase().equals("mm") ? EstimateGenicTranscription3.REPEAT_RNA_FILE_MM : (hg19 ? EstimateGenicTranscription3.REPEAT_RNA_FILE_HS_37 : EstimateGenicTranscription3.REPEAT_RNA_FILE_HS_36);
	    //Load the peaks
	    Vector<Peak> peaks = peakFileName==null ? null : new Vector<Peak>(species.equals("mm") ? FindEnhancerSites.parsePeakFile(peakFileName, chr).values() : EncodePeaks.parseEncodePeakFileCellType(peakFileName, "Helas3", chr));
	    Vector<Peak> breakPointPeaks = breakPointsFileName==null ? peaks : new Vector<Peak>(species.equals("mm") ? FindEnhancerSites.parsePeakFile(breakPointsFileName, chr).values() : EncodePeaks.parseEncodePeakFileCellType(breakPointsFileName, "Helas3", chr));

	    for (int i=1; i<=nChr; i++) {
		int nTranscriptsFoundRandomization[] = new int[nReadsMC];
		if (i==chr || chr==-1) {
		    BufferedWriter bwDownSampling = nDownSample>0 ? new BufferedWriter(new FileWriter(saveFileName + DOWNSAMPLING_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".txt")) : null;
		    if (bwDownSampling!=null) {
			if (i==1) {
			    bwDownSampling.write("#Chr");
			    for (int m=1; m<=nDownSample; m++)
				bwDownSampling.write("\t" + format.format((double)m/nDownSample) + "\tcons" + format.format((double)m/nDownSample));
			    bwDownSampling.write("\n");
			}
			bwDownSampling.write((new Integer(i)).toString());
		    }
		    for (int m=1; m<=nDownSample; m++) {
			ABIRead[] rnaReadsForward = null, rnaReadsReverse = null;
			if (kleinsteinFiles) {
			    rnaReadsForward = downSampleReads(ABIReads.loadKleinsteinReads(fileName, masterTableFileName, ext, true, i),(double)m/nDownSample);
			    rnaReadsReverse = downSampleReads(ABIReads.loadKleinsteinReads(fileName, masterTableFileName, ext, false, i),(double)m/nDownSample);
			}
			else {
			    rnaReadsForward = downSampleReads(ABIReads.loadMRNAReadsChr(fileName, true, i, readLength, false, repeatFileName),(double)m/nDownSample);
			    rnaReadsReverse = downSampleReads(ABIReads.loadMRNAReadsChr(fileName, false, i, readLength, false, repeatFileName),(double)m/nDownSample);
			}
			//Keep track of how many transcripts were found when the reads were randomized
			int nNovelCons = 0, ciSlackDist = 100;
			for (int l=0; l<Math.max(1, nReadsMC); l++) {
			    if (nReadsMC>0) {
				rnaReadsForward = randomizeReads(rnaReadsForward);
				rnaReadsReverse = randomizeReads(rnaReadsReverse);
				System.out.println("Shuffled reads");
			    }
			    geneInfo[i-1] = new RefSeqUCSCEnsemblGenes(i, orchestra, species, hg19);
			    geneInfo[i-1].setFactorReads(fileName, rnaReadsForward, 1);
			    geneInfo[i-1].setFactorReads(fileName, rnaReadsReverse, -1);
			    //Do some extra stuff for the downsampling
			    TreeSet<LncRNA> lncRNAs = null;
			    TreeSet<MiRNA> miRNAs = null;
			    TreeSet<MacroRNA> macroRNAs = null;
			    TreeSet<Enhancer2> allEnhancers = new TreeSet<Enhancer2>();
			    int maxDistance = 10000, maxEnhancerDist = 2000;
			    if (nDownSample>1) {
				String extragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_lRNA_lMe3_real.tsv";
				TreeSet<Enhancer2> extragenicEnhancers = FindTranscripts2.parseEnhancerFile(extragenicEnhancerFileName);
				String intragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_intragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_in_lMe3_real.tsv";
				TreeSet<Enhancer2> intragenicEnhancers = FindTranscripts2.parseEnhancerFile(intragenicEnhancerFileName);
				allEnhancers.addAll(extragenicEnhancers);
				allEnhancers.addAll(intragenicEnhancers);
				String lncRNAFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/landerRegions.bed";
				lncRNAs = AnalyzeTranscriptReadProfiles.loadLncRNA(lncRNAFileName, species);
				String macroRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/macroRNA.txt";
				macroRNAs = AnalyzeTranscriptReadProfiles.loadMacroRNA(macroRNAFileName, species);
				String miRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/miRNA.txt";
				miRNAs = AnalyzeTranscriptReadProfiles.loadMiRNA(miRNAFileName, species);
				//Set reads to the ncRNAs
				AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, fileName, !reverseStrand ? rnaReadsForward : rnaReadsReverse, i, !reverseStrand ? 1 : -1);
				AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, fileName, !reverseStrand ? rnaReadsForward : rnaReadsReverse, i);
				AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, !reverseStrand ? rnaReadsForward : rnaReadsReverse, i, !reverseStrand ? 1 : -1);
			    }
			    tf[i-1] = new TranscriptFinder2(reverseStrand ? rnaReadsReverse : rnaReadsForward, i, !reverseStrand, readLength, species, log);
			    tf[i-1].calculateHaarCoefficients(lVecCoeffs, binSize, saveInfoFileName);
			    geneInfo[i-1].setGeneIDsKeep(true, fileName, minDens);
			    geneInfo[i-1].setGeneIDsKeep(false, fileName, minDens);
			    int bestIterations = tf[i-1].findRefinedReducedModel(lMin, lMax, binSize, fdr, mergeSegments, gmm, saveInfoFileName, geneInfo[i-1], breakPointPeaks, peaks, maxDivergentTSSDistance, costFunction, -1);
			    reducedModel[i-1] = tf[i-1].getReducedModel();
			    if (nDownSample>1) {
				reducedModel[i-1].compare2Annotation(maxDistance, geneInfo[i-1], lncRNAs, miRNAs, macroRNAs, allEnhancers);
				reducedModel[i-1].categorizeUnannotatedTranscribedRegions(maxEnhancerDist, species.equals("mm"));
				//Find the ones that are highly conserved as well
				TreeSet<TranscribedRegion2> overlappingNovel = reducedModel[i-1].getSegments(TranscribedRegion2.NON_CODING_OTHER_RNA_STR);
				overlappingNovel.addAll(reducedModel[i-1].getSegments(TranscribedRegion2.PROMOTER_DISTAL_NCRNA_STR));
				ConservationIslands consIslands = new ConservationIslands(i, orchestra, species);
				for (TranscribedRegion2 t : overlappingNovel) {
				    //TranscribedRegionMerge t = new TranscribedRegionMerge(t2);
				    //t.assignReads(rnaReads);
				    //if (t.getReadDensity()>minDens && t.getReads().length>minReads) {
				    System.out.println("Found novel transcript " + t.toString());// + " with " + t.getReads().length);
					TreeSet<ConservationIsland> cis = consIslands.getOverlappingConservationIslands(chr, (!reverseStrand ? t.getStart() : t.getEnd()) - ciSlackDist, (!reverseStrand ? t.getStart() : t.getEnd()) + ciSlackDist);
					if (cis!=null && !cis.isEmpty()) {
					    nNovelCons++;
					    System.out.println("Found conserved novel transcript " + t.toString());// + " with " + t.getReads().length);
						    //}
					}
					//}
				}
			    }
			    else {
				reducedModel[i-1].compare2Annotation(maxDivergentTSSDistance, geneInfo[i-1]);
				reducedModel[i-1].categorizeTranscribedRegions();
			    }
			    System.out.println("Found " + reducedModel[i-1].numberOfTranscribedRegions(TranscribedRegion2.NON_CODING_OTHER_RNA_STR) + " novel transcripts with " + nNovelCons + " and " + reducedModel[i-1].numberOfTranscribedRegions(TranscribedRegion2.CODING_STR) + " coding transcripts from down-sampled model.");
			    //System.out.println(reducedModel[i-1].toString());
			    //Find out how many of the transcripts have a peak near their start
			    TreeMap<TranscribedRegion2, Peak> transcriptsWithPeaks = peaks==null ? new TreeMap<TranscribedRegion2, Peak>() : reducedModel[i-1].compare2Peaks(peaks, maxDivergentTSSDistance);
			    reducedModel[i-1].removeGenes();
			    for (int k=0; k<nMC; k++) {
				reducedModelRandom[i-1] = reducedModel[i-1].getRandomModel(readLength, reverseStrand ? rnaReadsReverse : rnaReadsForward, log);
				reducedModelRandom[i-1].compare2Annotation(maxDivergentTSSDistance, geneInfo[i-1]);
				reducedModelRandom[i-1].categorizeTranscribedRegions();
				reducedModelRandom[i-1].removeGenes();
			    }
			    if (saveFileName!=null && nReadsMC==0 && nDownSample==1) {
				String saveFile = saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".bed";
				reducedModel[i-1].saveReducedModel(saveFile, chr==1 || chr==-1 ? saveFile.substring(saveFile.lastIndexOf("/") + 1, saveFile.lastIndexOf(".")) : null, chr==-1 && i>1, hg19);
				if (orchestra && chr!=-1 && !fileName.contains("Plasmodium")) //Put it in a directory on the www-site for easy upload to the genome browser
				    createWigFile(saveFileName, false, reverseStrand, chr, false, species, hg19);
				//Save the lengths of the different transcripts
				saveTranscriptProperties(reducedModel[i-1], binSize, peaks!=null ? peaks.firstElement().getFactorNames(false) : null, orchestra, saveFileName, promoterLocationCategories, reverseStrand, chr, conservation);	
				//Save a list of novel transcribed regions
				saveNovelRegions(reducedModel[i-1], saveFileName, reverseStrand, chr);
				if (rand) {
				    saveFile = saveFileNameRandom + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".bed";
				    reducedModelRandom[i-1].saveReducedModel(saveFile, (chr==1 || chr==-1) && fileName.contains("Plasmodium") ? saveFile.substring(saveFile.lastIndexOf("/") + 1, saveFile.lastIndexOf(".")) : null, chr==-1 && i>1, hg19);
				    saveTranscriptProperties(reducedModelRandom[i-1], binSize, peaks!=null ? peaks.firstElement().getFactorNames(false) : null, orchestra, saveFileNameRandom, promoterLocationCategories, reverseStrand, chr, conservation);			
				}
				if (sequence) { //Save the sequences for each transcript
				    Runtime runtime = cpc ? Runtime.getRuntime() : null;
				    extractSequences(reducedModel[i-1], promoterLocationCategories, runtime, saveFileName + "/" + SEQUENCE_DIR_NAME + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")), reverseStrand, chr, minAAs, orchestra);
				    if (rand)
					extractSequences(reducedModelRandom[i-1], promoterLocationCategories, runtime, saveFileNameRandom + "/" + SEQUENCE_DIR_NAME + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")), reverseStrand, chr, minAAs, orchestra);
				}
			    }
			    else if (nReadsMC>0) { //Save info about the number of transcripts found
				nTranscriptsFoundRandomization[l] = reducedModel[i-1].numberOfTranscribedRegions();
				System.out.println("Found " + nTranscriptsFoundRandomization[l] + " transcripts from random model.");
			    }
			}
			if (nDownSample>0) {
			    bwDownSampling.write("\t" + reducedModel[i-1].numberOfTranscribedRegions(TranscribedRegion2.NON_CODING_OTHER_RNA_STR) + "\t" + nNovelCons);
			}
		    }
		    tf[i-1] = null;
		    reducedModel[i-1] = null;
		    geneInfo[i-1] = null;
		    if (nDownSample>0) {
			bwDownSampling.write("\n");
			bwDownSampling.close();
		    }
		    if (nReadsMC>0) {
			BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName + RANDOM_READS_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".txt"));
			for (int l=0; l<nReadsMC; l++)
			    bw.write(nTranscriptsFoundRandomization[l] + "\n");
			bw.close();
		    }
		}
	    }
	} 
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static ABIRead[] randomizeReads(ABIRead[] rnaReads) {
	//Use this method to randomize the location of the reads
	for (int i=0; i<rnaReads.length; i++)
	    rnaReads[i] = new ABIRead(rnaReads[i].getChromosome(), rnaReads[i].getStrand(), random.nextInt(RefSeqGenes.CHROMOSOME_LENGTH[rnaReads[i].getChromosome()]));
	Arrays.sort(rnaReads);
	return rnaReads;
    }

    public static void extractSequences(ReducedModel2 reducedModel, boolean promoterLocationCategories, Runtime runtime, String saveFileName, boolean reverseStrand, int chr, int minAAs, boolean orchestra) throws IOException {
	//Make sure that the directories exist
	File saveDir = new File(saveFileName.substring(0, saveFileName.lastIndexOf("/")));
	System.out.println("Creating directory " + saveDir.toString());
	if (!saveDir.mkdirs())
	    System.err.println("Error, failed to create directories to save: " + saveDir.toString());
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	for (int j=0; j<arr.length; j++) { 
	    //get one file for each category. This will allow us to split up the jobs for CPC speeding the whole thing up significantly
	    String seqSaveFileName = saveFileName + "_transcripts_" + arr[j] + "_" + (reverseStrand ? "r" : "f") + (chr==-1 ? "" : "_" + chr) + ".fa";
	    
	    reducedModel.saveSequences(orchestra, seqSaveFileName, arr[j], minAAs);
	    if (runtime!=null) { //Run CPC to analyze if the sequences are
		String cpcOutDir = seqSaveFileName.substring(0, seqSaveFileName.lastIndexOf("/") + 1) + "CPC/" + (reverseStrand ? "r" : "f") + "_" + chr + "/" + arr[j] + "/";
		(new File(cpcOutDir)).mkdirs();
		String cpcCmd = "bsub -o /dev/null -q all_7d ./run_cpc.sh " + seqSaveFileName + " " + cpcOutDir + seqSaveFileName.substring(seqSaveFileName.lastIndexOf("/") + 1, seqSaveFileName.lastIndexOf(".")) + " " + cpcOutDir;
		System.out.println("Running CPC:\n" + cpcCmd);
		Process cpcProcess = runtime.exec(cpcCmd);
		//cpcProcess.waitFor();
	    }
	}
    }

    public static void saveNovelRegions(ReducedModel2 reducedModel, String saveFileName, boolean reverseStrand, int chr) {

    }
    
    public static void saveGeneLatexTable(String geneLatexTableFileName, int totReads, int nUniqueGenes, int meanUniqueLengths, int totUniqueReads, int nSplitGenes, int meanSplitLengths, int totSplitReads, int nMergedGenes, int meanMergedLengths, int totMergedReads, int nMissedGenes, int meanMissedLength, int totMissedReads) { 
	DecimalFormat format = new DecimalFormat("###.###");
	try {
	    System.out.println("Saving summary statistics for genes to " + geneLatexTableFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(geneLatexTableFileName));
	    bw.write("\\begin{tabular}{lrcrr}\n\t\\textbf{Category} & \\multicolumn{2}{c}{\\textbf{\\# genes}} & \\textbf{$<$Length$>$} & \\textbf{Reads} \\\\\n\t\\hline\\\\\n");
	    double lenSum = 0;
	    int nReads = totUniqueReads + totMergedReads + totSplitReads + totMissedReads;
	    int nGenes = nUniqueGenes + nMergedGenes + nSplitGenes + nMissedGenes;
	    bw.write(RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nUniqueGenes + " & (" + format.format((double)nUniqueGenes/nGenes) + ") & " + meanUniqueLengths + " & " + format.format((double)totUniqueReads/totReads) + "\\\\\n");
	    bw.write(RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nSplitGenes + " & (" + format.format((double)nSplitGenes/nGenes) + ") & " + meanSplitLengths + " & " + format.format((double)totSplitReads/totReads) + "\\\\\n");
	    bw.write(RefSeqGene.MERGED_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nMergedGenes + " & (" + format.format((double)nMergedGenes/nGenes) + ") & " + meanMergedLengths + " & " + format.format((double)totMergedReads/totReads) + "\\\\\n");
	    bw.write(RefSeqGene.MISSED_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nMissedGenes + " & (" + format.format((double)nMissedGenes/nGenes) + ") & " + meanMissedLength + " & " + format.format((double)totMissedReads/totReads) + "\\\\\n");
	    bw.write("\\hline\n");
	    bw.write("Total & " + nGenes + " & & & " + format.format((double)nReads/totReads) + "\n");
	    bw.write("\\end{tabular}\n"); //footer
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving latex table " + e.toString());
	}
    }


    public static void saveTranscriptProperties(ReducedModel2 reducedModel, int binSize, String peakFileNames[], boolean orchestra, String saveFileName, boolean promoterLocationCategories, boolean reverseStrand, int chr, boolean conservation) { 
	saveTranscriptProperty(reducedModel.getTranscribedRegionTypesLengths(), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_LENGTHS_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	saveTranscriptProperty(reducedModel.getTranscribedRegionTypesDensities(), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_DENSITIES_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	saveTranscriptProperty(reducedModel.getTranscribedRegionTypesWithinDensityStdDevs((int)Math.pow(2, binSize)), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_WITHIN_DENISTY_STD_DEV_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	if (peakFileNames!=null) {
	    for (int i=0; i<peakFileNames.length; i++)
		saveTranscriptProperty(reducedModel.getTranscribedRegionTypesPeaks(peakFileNames[i]), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + peakFileNames[i] + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	}
	if (conservation) {
	    reducedModel.calculateConservation(orchestra);
	    saveTranscriptProperty(reducedModel.getTranscribedRegionTypesConservation(), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_CONSERVATION_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	    saveTranscriptProperty(reducedModel.getTranscribedRegionTypesWithinConservationStdDevs(binSize), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_WITHIN_CONSERVATION_STD_DEV_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	}
    }

    public static void saveTranscriptProperty(TreeMap<String, TallyStore> type2data, boolean promoterLocationCategories, String saveFileName, boolean header) {
	//Use this method to save data of a specific transcript
	//property to a file. Each column represents a certain type of
	//transcript and they are padded with NaN to all have the same
	//number of rows which will make it easier to merge.
	try {
	    System.out.println("Saving to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	    if (header) {
		bw.write("#");
		for (int j=0; j<arr.length; j++) 
		    bw.write(arr[j] + "\t");
		bw.write("\n");
	    }
	    int maxCount = 0; //Find the category with the largest no elements
	    for (TallyStore t : type2data.values())
		maxCount = maxCount>t.numberObs() ? maxCount : t.numberObs();
	    for (int i=0; i<maxCount; i++) {
		for (int j=0; j<arr.length; j++) {
		    if (type2data.containsKey(arr[j]) && i<type2data.get(arr[j]).numberObs() && !Double.isNaN(type2data.get(arr[j]).getArray()[i]) && !Double.isInfinite(type2data.get(arr[j]).getArray()[i]))
			bw.write(format.format(type2data.get(arr[j]).getArray()[i]) + "\t");
		    else
			bw.write("NaN\t");
		}
		bw.write("\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving " + e.toString());
	}
    }


    public static String printStats(double fdr, ReducedModel2 reducedModel, TreeMap<TranscribedRegion2, Peak> transcriptsWithPeaks) { return printStats(fdr, reducedModel, transcriptsWithPeaks, null); }

    public static String printStats(double fdr, ReducedModel2 reducedModel, TreeMap<TranscribedRegion2, Peak> transcriptsWithPeaks, Vector<Peak> peaks) {
	//n_regions, proportion transcribed, proportion of reads
	String str = "";//format.format(fdr) + "\t" + reducedModel.numberOfTranscribedRegions() + "\t" + format.format(reducedModel.proportionTranscribed()) + "\t" + format.format(reducedModel.proportionOfReads()) + "\t";
	//PPV sensitivity
	str += format.format(reducedModel.getPositivePredictiveValue()) + "\t";
	//n_unique, n_missed_genes, n_genes_w_multiple
	str += format.format(reducedModel.numberOfGenesWithOneTranscript()) + "\t" + format.format(reducedModel.numberOfGenesWithNTranscripts(0, false, false)) + "\t" + format.format(reducedModel.numberOfGenesWithNTranscripts(2, true, false)) + "\t";
	//n_nc_transcripts, n_transcripts_covering_multiple
	str += format.format(reducedModel.numberOfTranscriptsWithNGenes(0, false, false)) + "\t" + format.format(reducedModel.numberOfTranscriptsWithNGenes(2, true, false)) + "\t";
	//proportion with peaks
	if (peaks!=null)
	    str += format.format((double)transcriptsWithPeaks.size()/(peaks.size() - transcriptsWithPeaks.size() + reducedModel.numberOfTranscribedRegions()));
	else
	    str += format.format((double)transcriptsWithPeaks.size()/reducedModel.numberOfTranscribedRegions());
	System.out.println(str);
	return str;
    }

    public static void createWigFile(String saveFileName, boolean chip, boolean reverseStrand, int chr, boolean bayes, String species, boolean hg19) {
	String saveFile = saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".bed";
	String name = saveFileName.substring(saveFileName.lastIndexOf("mRNA/") + 5) + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f");
	String colorStr = CreateWigFiles.getColorString(saveFileName);
	int priority = CreateWigFiles.getPriority(saveFileName);
	String trackInfoStr = "track type=bigWig name=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" description=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" priority=" + priority + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=" + URL_NAME + name + (chr==-1 ? "" : "_" + chr) + ".bw";
	CreateWigFiles.addTrackInfo(trackInfoStr, TRACK_INFO_FILE_PATH + (saveFileName.contains("MeCP2") ? "MeCP2/" : saveFileName.substring(saveFileName.lastIndexOf("mRNA/") + 4, saveFileName.lastIndexOf("/") + 1)) + TRACK_INFO_FILE_NAME);
	CreateWigFiles.createWWWLink(saveFile, "/www/" + WWW_DIR_NAME + (saveFileName.contains("MeCP2") ? "MeCP2/" : saveFileName.substring(saveFileName.lastIndexOf("mRNA/") + 5, saveFileName.lastIndexOf("/") + 1)), species, hg19);
	//createWigFile(saveFileName, name, chr);
    }

    /*public static void createWigFile(String saveFileName, int chr) { createWigFile(saveFileName, saveFileName.replace(".bed", ""), chr); }

    public static void createWigFile(String saveFileName, String name, int chr) {
	String colorStr = CreateWigFiles.getColorString(saveFileName);
	int priority = CreateWigFiles.getPriority(saveFileName);
	String trackInfoStr = "track type=bigWig name=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" description=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" priority=" + priority + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=" + URL_NAME + name + (chr==-1 ? "" : "_" + chr) + ".bw";
	CreateWigFiles.addTrackInfo(trackInfoStr, TRACK_INFO_FILE_PATH + (saveFileName.contains("MeCP2") ? "MeCP2/" : "") + TRACK_INFO_FILE_NAME);
	CreateWigFiles.createWWWLink(saveFile, "/www/" + WWW_DIR_NAME + (saveFileName.contains("MeCP2") ? "MeCP2/" : ""));
	}*/

    public static int numberOfChromosomes(String species) {
	if (species.equals("hs") || species.equals("human"))
	    return 24;
	else if (species.equals("sc") || species.equals("cerevisiae"))
	    return 17;
	else if (species.equals("pf") || species.equals("plasmodium"))
	    return 15;
	return 21; //default is mouse
    }

    public static int[] getLengthVec(int lMin, int lMax, int deltaL) {
	int lVec[] = new int[(lMax-lMin)/deltaL];
	for (int i=0; i<(lMax-lMin)/deltaL; i++)
	    lVec[i] = lMin + i*deltaL;
	return lVec;
    }

    public static double[] getFDRVec(double fdrMin, double fdrMax, int nFDR) {
	//create nFDR points equally spaced on a log-scale btwn fdrMin and fdrMax
	if (nFDR<=1)
	    return new double[]{fdrMin};
	double fdrVec[] = new double[nFDR], deltaFDR = (Math.log10(fdrMax) - Math.log10(fdrMin))/(nFDR-1);
	for (int i=0; i<nFDR; i++)
	    fdrVec[i] = Math.pow(10, Math.log10(fdrMin) + deltaFDR*i);
	return fdrVec;
    }

    public static double[] getCompareVec(double compareMin, double compareMax, int nCompare) {
	if (nCompare<=1)
	    return new double[]{compareMin};
	double compareVec[] = new double[nCompare], delta = (compareMax - compareMin)/(nCompare - 1);
	for (int i=0; i<nCompare; i++)
	    compareVec[i] = compareMin + delta*i;
	return compareVec;
    }

    ///////////////////////////////////////////////////////



    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName) { return parseEnhancerFile(fileName, "Mm"); }
		  
    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName, String species) { return parseEnhancerFile(fileName, species, -1); }
		  
    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName, int chr) { return parseEnhancerFile(fileName, "Mm", chr); }

    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName, String species, int chr) {
	//Use the subset files by JMG that contain chr, pos and TRUE|FALSE
	TreeSet<Enhancer2> enhancers = new TreeSet<Enhancer2>();
	if (fileName!=null) {
	    try {
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		String line = br.readLine();
		//chr pos TRUE|FALSE
		Pattern enhancerPattern = Pattern.compile("^(\\d+)\\s+(\\d+)\\s+(\\w+)");
		while (line!=null) {
		    Matcher m = enhancerPattern.matcher(line);
		    if (m.find()) {
			int c = Integer.parseInt(m.group(1));
			if ((c==chr || chr==-1) && m.group(3).equals("TRUE")) 
			    enhancers.add(new Enhancer2(c, Integer.parseInt(m.group(2))));
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error reading enhancers from " + fileName + ": " + e.toString());
	    }
	}
	System.out.println("Found " + enhancers.size() + " enhancers.");
	return enhancers;
    }

public static ABIRead[] downSampleReads(ABIRead[] reads, double keepFraction) {
    int nKeep = (int)Math.round((double)reads.length*keepFraction);
	//Permute all the indices, then take the nKeep first ones and
	//retain those reads
	int inds[] = new int[reads.length];
	for (int i=0; i<reads.length; i++)
	    inds[i] = i;
	for (int i=0; i<reads.length*10; i++) {
	    int ind1 = random.nextInt(reads.length), ind2 = random.nextInt(reads.length);
	    int tmp = inds[ind1];
	    inds[ind1] = inds[ind2];
	    inds[ind2] = tmp;
	}
	ABIRead rs[] = new ABIRead[nKeep];
	for (int i=0; i<nKeep; i++)
	    rs[i] = reads[inds[i]];
	Arrays.sort(rs);
	System.out.println("Sampled " + rs.length + " reads.");
	return rs;
    }

}

