//Use this class to keep track of the mappability index. It does so by
//reading from the file. It can be queried efficiently if all queries
//are in terms of ascending coordinate values.

import java.util.*;
import java.util.regex.*;
import java.io.*;

public class Mappability2 {
    
    protected int nBasepairs, nUniqueBasePairs, nWildcards, readLength = 35;
    //Use this window to keep track of the number of reads covering
    //the bases in the window
    protected int mappabilityWindow[] = new int[readLength], windowStart = 0, mappabilityIndex = 0;
    //Keep track of when the distance changes
    protected int segmentEnd, minDist;
    //Reader to get the data
    protected BufferedReader mappabilityReader;
    //Pattern for parsing the file
    protected Pattern mappabilityPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)");


    public Mappability2(String genomeLoc, int chr) {
	//String bedFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr) + "_U_" + AnalyzeControls.READ_LENGTH + ".bed";
	//String statsFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr) + "_stats_" + AnalyzeControls.READ_LENGTH + ".txt";
	String bedFileName = genomeLoc + "chrY_U_" + readLength + ".bed";
	String statsFileName = genomeLoc + "chrY_stats_" + readLength + ".txt";
	try { //Read the data for the summary statistics
	    BufferedReader br = new BufferedReader(new FileReader(statsFileName));
	    Pattern statsPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)");
	    String line = br.readLine();
	    Matcher m = statsPattern.matcher(line);
	    if (m.find()) {
		nBasepairs = Integer.parseInt(m.group(1));
		nUniqueBasePairs = Integer.parseInt(m.group(2));
		nWildcards = Integer.parseInt(m.group(3));
	    }
	    br.close();
	    //Open the file containing the distance info
	    mappabilityReader = new BufferedReader(new FileReader(bedFileName));
	    //Remove the first line since it contains comments
	    mappabilityReader.readLine();
	    //Second line may contain an error
	    line = mappabilityReader.readLine();
	    m = mappabilityPattern.matcher(line);
	    if (m.find()) {
		if (m.group(1)=="0" && m.group(2)=="0") {
		    line = mappabilityReader.readLine();
		    m = mappabilityPattern.matcher(line);
		    if (m.find()) {
			segmentEnd = Integer.parseInt(m.group(2));
			minDist = Integer.parseInt(m.group(3));
		    }
		}
		else { //No error
		    segmentEnd = Integer.parseInt(m.group(2));
		    minDist = Integer.parseInt(m.group(3));
		}
		for (int i=0; i<readLength; i++) //initialize the window
		    mappabilityWindow[i] = minDist>0 ? 1 : 0;
		mappabilityIndex = mappabilityWindow[0];
	    }
	}
	catch (IOException e) {
	    System.err.println("Error reading " + statsFileName + "\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
    }

    //////////////////////////////////////////////////////////////////

    public int getNumberOfUniqueBases() { return nUniqueBasePairs; }
    
    public int getNumberOfBases() { return nBasepairs; }

    //////////////////////////////////////////////////////////////////

    public int numberOfPotentialReadsNextBase() {
	//Find out how many potential reads (considering both strands)
	//that cover the next base
	int retVal = mappabilityIndex;
	//shift everything to the left and add a zero at the end
	for (int i=0; i<readLength-1; i++) { mappabilityWindow[i] = mappabilityWindow[i+1]; }
	mappabilityWindow[readLength-1] = 0;
	if (++windowStart==segmentEnd) {
	    String line;
	    try {
		line = mappabilityReader.readLine();
		if (line==null) { 
		    mappabilityReader.close();
		    return -1; 
		}
	    }
	    catch (IOException e) {
		System.err.println("Error reading from mappabilityReader");
		return -1;
	    }
	    Matcher m = mappabilityPattern.matcher(line);
	    if (m.find()) {
		segmentEnd = Integer.parseInt(m.group(2));
		minDist = Integer.parseInt(m.group(3));
	    }
	}
	for (int i=0; i<readLength; i++)
	    mappabilityWindow[i] += minDist>0 ? 1 : 0;
	mappabilityIndex = mappabilityWindow[0];
	return retVal;
    }

}

