//Use this to merge the .bed-files when we calculate the Mappability
//index. If all files which should be present are present, then the
//files are removed after merging.

import java.util.*;
import java.io.*;
import java.util.regex.*;
import jargs.gnu.CmdLineParser;

public class MergeMappabilityBedFiles {

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    //Use this flag to for the location of the .bed-files
	    CmdLineParser.Option saveDirectoryFlag = argParser.addStringOption('d', "saveDirectory");
	    //Use this flag to decide how many errors to allow
	    CmdLineParser.Option misMatchesFlag = argParser.addIntegerOption('m', "misMatches");
	    //Length of the mers
	    CmdLineParser.Option merLengthFlag = argParser.addIntegerOption('k', "length");
	    //Force removal of files
	    CmdLineParser.Option removeFlag = argParser.addBooleanOption('r', "remove");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag to decide which species to
	    //investigate. Available species are: 
	    //Mouse (Mm) Default
	    //Yeast (Sc)
	    CmdLineParser.Option speciesFlag = argParser.addStringOption('s', "species");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(misMatchesFlag);
	    int nMismatches = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    tmp = argParser.getOptionValue(merLengthFlag);
	    int merLength = tmp!=null ? ((Integer)tmp).intValue() : 35;
	    boolean remove = argParser.getOptionValue(removeFlag)!=null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(saveDirectoryFlag);
	    String saveDir = tmp!=null ? tmp.toString() : "./";
	    //make sure that it ends with '/'
	    if (!saveDir.endsWith("/")) { saveDir += "/"; }
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? tmp.toString() : "Mm";
	    //////////////////////////////////////////////////////////////////
	    String GENOME_LOC = "./";
	    int nChr = 0;
	    if (species.equals("Mm")) {
		GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
		nChr = 21;
	    }
	    else if (species.equals("Sc")) {
		GENOME_LOC = orchestra ? "/home/mh168/Genomes/Sc/sacCer1/" : "/media/exthd/Genomes/Sc/sacCer1/";
		nChr = 17;
	    }
	    for (int i=1; i<=1; i++) {
		//Find out what the last file is supposed to look like
		int start = 0, end = (int)AnalyzeMerDistance.findFileLength(GENOME_LOC + AnalyzeControls.getChromosomeString(i)) - merLength - 1;
		while (start<end) {
		    TreeMap minDists = new TreeMap();
		    minDists = mergeFiles(saveDir, start, i, nChr, species, merLength, nMismatches, minDists);
		    String saveFileName = saveDir + "chr" + AnalyzeControls.getChromosomeNumber(i, species) + "_" + merLength + "_" + nMismatches + "_" + minDists.firstKey().toString() + "_" + minDists.lastKey().toString() + ".bed";
		    if (saveFiles("chr" + AnalyzeControls.getChromosomeNumber(i, species), minDists, saveFileName) && remove)
			removeFiles(saveDir, start, i, nChr, species, merLength, nMismatches); 
		    start = ((Integer)minDists.lastKey()).intValue() + 1;
		}
	    }
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    protected static void removeFiles(String dirName, int start, int c, int nChr, String species, int merLength, int misMatches) {
	TreeMap mergedMinDists = new TreeMap();
	//For the file names
	Pattern chrPattern = Pattern.compile("chr" + AnalyzeControls.getChromosomeNumber(c, species) + "_([\\d|X|Y|M]+)_(\\d+)_(\\d+)_(\\d+)_(\\d+).bed");
	//For the file contents
	Pattern mapPattern = Pattern.compile("chr([\\d|X|Y|M]+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
	//Find all the files
	for (int i=1; i<=nChr; i++) {
	    
	    File dir = new File(dirName);
	    String fileNames[] = dir.list();
	    for (int k=0; k<fileNames.length; k++) {
		Matcher m = chrPattern.matcher(fileNames[k]);
		if (m.find()) {
		    //Make sure that it is the file that we are looking for
		    String chrRef = m.group(1);
		    int merLen = Integer.parseInt(m.group(2));
		    int mis = Integer.parseInt(m.group(3));
		    int s = Integer.parseInt(m.group(4));
		    if (chrRef.equals(AnalyzeControls.getChromosomeNumber(i, species)) && merLen==merLength && mis==misMatches && s==start) 
			(new File(fileNames[k])).delete();
		}
	    }
	}
    }

	
    protected static boolean saveFiles(String chrStr, TreeMap minDists, String saveFileName) {
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    int start = ((Integer)minDists.firstKey()).intValue(), end = ((Integer)minDists.lastKey()).intValue(), prevDist = ((Integer)minDists.get(minDists.firstKey())).intValue(), prevStart = start;
	    for (int i=start + 1; i<=end; i++) {
		int dist = ((Integer)minDists.get(new Integer(i))).intValue();
		if (prevDist!=dist) {
		    //System.out.println(chrStr + "\t" + prevStart + "\t" + (i + 1) + "\t" + prevDist);
		    bw.write(chrStr + "\t" + prevStart + "\t" + (i + 1) + "\t" + prevDist + "\n");
		    prevDist = dist;
		    prevStart = i;
		}
	    }
	    //add the last chunk
	    bw.write(chrStr + "\t" + prevStart + "\t" + (end + 1) + "\t" + prevDist + "\n");
	    bw.close();
	}
	catch (IOException e) { return false; }
	return true;
    }

    protected static TreeMap mergeFiles(String dirName, int start, int c, int nChr, String species, int merLength, int misMatches, TreeMap minDists) {
	TreeMap mergedMinDists = new TreeMap();
	try {
	    //For the file names
	    Pattern chrPattern = Pattern.compile("chr" + AnalyzeControls.getChromosomeNumber(c, species) + "_([\\d|X|Y|M]+)_(\\d+)_(\\d+)_(\\d+)_(\\d+).bed");
	    //For the file contents
	    Pattern mapPattern = Pattern.compile("chr([\\d|X|Y|M]+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
	    //Find all the files
	    for (int i=1; i<=nChr; i++) {
		
		File dir = new File(dirName);
		String fileNames[] = dir.list();
		for (int k=0; k<fileNames.length; k++) {
		    Matcher m = chrPattern.matcher(fileNames[k]);
		    if (m.find()) {
			//Make sure that it is the file that we are looking for
			String chrRef = m.group(1);
			int merLen = Integer.parseInt(m.group(2));
			int mis = Integer.parseInt(m.group(3));
			int s = Integer.parseInt(m.group(4));
			if (chrRef.equals(AnalyzeControls.getChromosomeNumber(i, species)) && merLen==merLength && mis==misMatches && s==start) {
			    BufferedReader br = new BufferedReader(new FileReader(dirName + fileNames[k]));
			    String line = br.readLine();
			    while (line!=null) {
				Matcher m2 = mapPattern.matcher(line);
				if (m2.find()) {
				    int st = Integer.parseInt(m2.group(2));
				    int en = Integer.parseInt(m2.group(3));
				    int dist = Integer.parseInt(m2.group(4));
				    for (int j=st; j<en; j++) { //Update the shortest distances
					Integer key = new Integer(j);
					if (minDists.containsKey(key))
					    dist = ((Integer)minDists.get(key)).intValue()<dist ? ((Integer)minDists.get(key)).intValue() : dist;
					mergedMinDists.put(key, new Integer(dist));
				    }
				}
				line = br.readLine();
			    }
			    br.close();
			}
		    }
		}
	    }
	}
	catch (IOException e) { }
	return mergedMinDists;
    }

}
