//Use this method as a wrapper for the Patronus software which scans
//for consensus sequences and calculates p-values.

import java.util.*;
import java.util.regex.*;
import java.io.*;

public class PatronusRunner {

    protected int markovModelOrder = 1;
    protected boolean orchestra = false;
    protected String seq;
    protected String motif;
    protected TreeMap<String, Integer> nMotifs;
    protected TreeMap<String, Double> pVals;
    protected Runtime runtime;

    public PatronusRunner(boolean o) {
	orchestra = o;
	runtime = Runtime.getRuntime();
    }

    public void setSequence(String s) { seq = s; }

    public void calculateBackgroundModel(int m) {
	markovModelOrder = m;
	calculateBackgroundModel();
    }

    public void calculateBackgroundModel() {
	runtime.exec(locationStr + "patronus -m " + markovModelOrder + " -i " + SEQUENCE_FILE_NAME + " -om " + MODEL_FILE_NAME);
    }

    public void findMotif(String motif) {
	runtime.exec(locationStr + "patronus -i " + SEQUENCE_FILE_NAME + " -im " + MODEL_FILE_NAME + " -p \":iupac_dna:" + motif + "\" -o " + SEQUENCE_FILE_NAME + ".out -os " + SEQUENCE_FILE_NAME + ".1");
    }

    public TreeMap<String, Integer> numberOfMotifsFound() {
	TreeMap<String, Integer> nMotifs = new TreeMap<String, Integer>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(sequenceFileName + ".out"));
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = patronusPattern.matcher(line);
		if (m.find()) {
		    
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error opeing Patronus file.");
	}
	return nMotifs;
    }

    public TreeMap<String, Double> getPValues() {
	TreeMap<String, Double> pVals = new TreeMap<String, Double>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(sequenceFileName + ".out"));
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = patronusPattern.matcher(line);
		if (m.find()) {

		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error opeing Patronus file.");
	}
	return pVals;
    }

}

//Use this class to parse the output file

class PatronusParser {

    protected String fileName;

    public PatronusParser(String f){
	fileName = f;
    }

    //public double[] findMotifTFMScores() { return findMotifTFMScores(""); }
    
    public double[] findMotifTFMScores(String matrixName) {
	Vector motifScores = new Vector();
	matrixName = matrixName.length()==0 ? null : matrixName;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern posPattern = Pattern.compile("^(\\d+)\\s+(\\w+)\\s+([\\d|\\.|-]+)");
	    while (line!=null) {
		Matcher m = posPattern.matcher(line);
		if (m.find()) {
		    if (matrixName==null || matrixName.equals(m.group(2)))
			motifScores.addElement(new Double(m.group(3)));
		    //System.out.println("Found line: " + line + ", with " + m.group(1));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {

	}
	double[] motifs = new double[motifScores.size()];
	for (int i=0; i<motifScores.size(); i++)
	    motifs[i] = ((Double)motifScores.elementAt(i)).intValue();
	return motifs;
    }


    public int[] findMotifLocations(int a) {	return findMotifLocations(""); }

    public int[] findMotifLocations(String matrixName){
	Vector motifInds = new Vector();
	matrixName = matrixName.length()==0 ? null : matrixName;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern posPattern = Pattern.compile("^(\\d+)\\s+(\\w+)\\s+([\\d|\\.|-]+)");
	    while (line!=null) {
		Matcher m = posPattern.matcher(line);
		if (m.find()) {
		    if (matrixName==null || matrixName.equals(m.group(2)))
			motifInds.addElement(new Integer(m.group(1)));
		    //System.out.println("Found line: " + line + ", with " + m.group(1));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {

	}
	int[] motifLocations = new int[motifInds.size()];
	for (int i=0; i<motifInds.size(); i++)
	    motifLocations[i] = ((Integer)motifInds.elementAt(i)).intValue();
	return motifLocations;
    }


}
