//Use this class for parsing the phastCons files for a genome. v2 is
//streamline to find regions of high conservation and it is not
//suitable for finding the amount of conservation for a specific set
//of coordinates.
//

import java.util.*;
import java.util.regex.*;
import java.io.*;

public class PhastConsParser2 {

    protected String fileName, species;
    protected BufferedReader phastConsReader = null;
    //Keep track of where in the genome we are when reading from the file
    protected int index  = 0, bufferStart, blockStart;
    protected double buffer[];
    protected int blockEnd = -1;

    public static final Pattern phastConsHeaderPattern = Pattern.compile("fixedStep\\s+chrom=chr([\\d|X|Y]+)\\s+start=(\\d+)\\s+step=(\\d+)");

    //////////////////////////////////////////////////////////////////

    public PhastConsParser2(int chr, boolean vertebrate, boolean orchestra, String sp, boolean hg19) {
	try {
	    species = sp;
	    fileName = orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/";
	    if (species.equals("mm")) {
		if (chr==20)
		    fileName += "Genomes/Mm/v9/PhastConsScores/chrX";
		else if (chr==21) 
		    fileName += "Genomes/Mm/v9/PhastConsScores/chrY";
		else
		    fileName += "Genomes/Mm/v9/PhastConsScores/chr" + chr;
		fileName += vertebrate ? ".vertebrates.pp.data" : ".euarchontoglires.pp.data";
	    }
	    else if (species.equals("hs")) {
		if (chr==23)
		    fileName += "Genomes/Hs/v36/PhastConsScores/chrX";
		else if (chr==24) 
		    fileName += "Genomes/Hs/v36/PhastConsScores/chrY";
		else
		    fileName += "Genomes/Hs/v36/PhastConsScores/chr" + chr;
		if (hg19) {
		    fileName = fileName.replace("v36", "v37");
		    fileName += ".phastCons46way.wigFix";
		}
		else
		    fileName += ".pp";
	    }
	    phastConsReader = new BufferedReader(new FileReader(fileName));

	}
	catch (IOException e) {
	    System.err.println("Error reading phastConsScores " + e.toString());
	}
    }

    public PhastConsParser2(int chr, boolean vertebrate, boolean orchestra, String sp) {
	try {
	    species = sp;
	    fileName = orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/";
	    if (species.equals("mm")) {
		if (chr==20)
		    fileName += "Genomes/Mm/v9/PhastConsScores/chrX";
		else if (chr==21) 
		    fileName += "Genomes/Mm/v9/PhastConsScores/chrY";
		else
		    fileName += "Genomes/Mm/v9/PhastConsScores/chr" + chr;
		fileName += vertebrate ? ".vertebrates.pp.data" : ".euarchontoglires.pp.data";
	    }
	    else if (species.equals("hs")) {
		if (chr==23)
		    fileName += "Genomes/Hs/v36/PhastConsScores/chrX";
		else if (chr==24) 
		    fileName += "Genomes/Hs/v36/PhastConsScores/chrY";
		else
		    fileName += "Genomes/Hs/v36/PhastConsScores/chr" + chr;
		fileName += ".pp";
	    }
	    phastConsReader = new BufferedReader(new FileReader(fileName));

	}
	catch (IOException e) {
	    System.err.println("Error reading phastConsScores " + e.toString());
	}
    }

    //////////////////////////////////////////////////////////////////

    public boolean hasMoreIslands() { 
	try {
	    return phastConsReader.ready(); 
	}
	catch (IOException e) {
	    //System.err.println("Error reading phastConsScores " + e.toString());
	}
	return false;
    }

    public int getNextIslandStart() { return index; }

    public double[] getNextIsland(int len) {
	double scores[] = new double[len];
	Arrays.fill(scores, 0);
	try {
	    for (int i=0; i<len; i++) {
		String line = phastConsReader.readLine();
		if (line==null) {
		    phastConsReader.close();
		    break;
		}
		Matcher m = phastConsHeaderPattern.matcher(line);
		if (m.find()) { //End of block, check how far to the next one
		    blockStart = Integer.parseInt(m.group(2));
		    int gap = blockStart - index;
		    //System.out.println("Found new block starting " + blockStart + " with gap " + gap);
		    index = blockStart;
		    if (gap>i || i+gap>=scores.length) //Easy, the rest should just be zeros, we're done
			break;
		    else //this is hopefully rare if len is small
			i += gap;
		    line = phastConsReader.readLine();
		    if (line==null) {
			phastConsReader.close();
			break;
		    }
		}
		scores[i] = Double.parseDouble(line);
		index++;
	    }
	}
	catch (IOException e) {
	    System.err.println("Error reading phastConsScores " + e.toString());
	}
	return scores;
    }

    /*public double[] findScores(int start, int end) throws ArrayIndexOutOfBoundsException {
	int len = end - start, i = 0;
	double ret[] = new double[len];
	try {
	    if (phastConsReader==null || len>buffer.length) 
		initializeParser(len);
	    //moveBuffer(start);
	    while (start>bufferStart + buffer.length) { 
		readScore();
		bufferStart++;
	    }
	    //Now we have reached the point where the buffer is
	    //aligned with region of interest.
	    while (start>bufferStart) { 
		buffer = shiftArrayLeft(buffer, readScore());
		bufferStart++;
	    }
	    //Copy the stuff that we need from the buffer to ret
	    for (int j=0; j<len; j++)
		ret[j] = buffer[j];
	}
	catch (IOException e) {
	    System.err.println("Error finding conservation scores from " + fileName + " for " + start + ", " + end);
	    return null;
	}
	//System.out.println("Found scores for " + start);
	return ret;
    }

    protected void initializeParser(int len) throws IOException {
	System.out.println("Initializing PhastConsParser for " + fileName + " with buffer length " + len);
	phastConsReader = new BufferedReader(new FileReader(fileName));
	String line = phastConsReader.readLine();
	Matcher m = phastConsHeaderPattern.matcher(line);
	index = 0;
	bufferStart = 0;
	buffer = new double[len];
	if (m.find()) {
	    //Read all the subsequent scores, they will be
	    //stored in the array scores
	    while (index<buffer.length) 
		buffer[index++] = readScore();
	}
	//System.out.println("Initialized parser, index " + index);
    }

    protected double readScore() throws IOException {
	if (index<blockEnd) {
	    index++;
	    return 0;
	}
	else {
	    String line = phastConsReader.readLine();
	    if (line!=null) {
		Matcher m = phastConsHeaderPattern.matcher(line);
		if (m.find()) {
		    blockEnd = Integer.parseInt(m.group(2));
		    index++;
		    return 0;
		}
		else {
		    index++;
		    return Double.parseDouble(line);
		}
	    }
	}
	return 0;
    }

    public static double[] shiftArrayLeft(double a[], double d) {
	double b[] = new double[a.length];
	for (int i=1; i<a.length; i++)
	    b[i-1] = a[i];
	b[b.length - 1] = d;
	return b;
	}*/


}
