//Use this class to predict the number of peaks as the number of reads
//is increased. It implements the Bayesian algorithm by Lijoi et al
//(2007) and is similar to predict_peaks.m

import jargs.gnu.CmdLineParser;
//import java.io.*;
import java.text.DecimalFormat;
import java.math.BigDecimal;

public class PredictPeaks {

    public static DecimalFormat format = new DecimalFormat("###.####");

    /*public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    //Use this flag to for the location of the
	    //_genes_[\\d+_]E120.txt file containing the peak
	    //information
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('f', "fileName");
	    //Use this flag to determine the number of additional
	    //reads for which we wish to predict the total number of
	    //discovered peaks.
	    CmdLineParser.Option predictReadsFlag = argParser.addIntegerOption('m', "additionalReadsFlag");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? tmp.toString() : null;
	    tmp = argParser.getOptionValue(predictReadsFlag);
	    int predictReads = tmp!=null ? ((Integer)tmp).intValue() : 100000;
	    //////////////////////////////////////////////////////////////////
	    //parse the file containing the peak-info

	    //re-arrange the data to the format required by the
	    //algorithm, that is an array containing the number of
	    //reads detected for each peak

	    //Find the optimal parameters sigma_hat and theta_hat
	    //using a Newton method.

	    //calculate the expected number of peaks based on the
	    //formula from the paper.
	    //int nPredictedPeaks = expectedPeaks(nPeaks, nReads, predictReads, thetaHat, sigmaHat);
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
	}*/

    public PredictPeaks(){ }

    public static int[] expectedPeaks(int nPeaks, int nReads, int predictReads[], double thetaHat, double sigmaHat) {
	int nExpected[] = new int[predictReads.length];
	for (int i=0; i<predictReads.length; i++)
	    nExpected[i] = expectedPeaks(nPeaks, nReads, predictReads[i], thetaHat, sigmaHat);
	return nExpected;
    }

    public static int expectedPeaks(int nPeaks, int nReads, int predictReads, double thetaHat, double sigmaHat) {
	BigDecimal b = new BigDecimal(nPeaks + thetaHat/sigmaHat);
	BigDecimal num = pochhammer(thetaHat + nReads + sigmaHat, predictReads);
	BigDecimal den = pochhammer(thetaHat + nReads, predictReads);
	BigDecimal rat = num.divide(den, java.math.RoundingMode.HALF_EVEN);
	BigDecimal rat2 = rat.subtract(new BigDecimal(1));
	BigDecimal c = b.multiply(rat2);
	//System.out.println("b = " + format.format(b) + "\nnum = " + format.format(num) + "\nden = " + format.format(den) + "\nrat2 = " + format.format(rat2) + "\nc = " + format.format(c));
	return (int)Math.round(c.floatValue());
    }

    public static BigDecimal pochhammer(BigDecimal a, int n) { return pochhammer(a.floatValue(), n); }

    public static BigDecimal pochhammer(BigDecimal a, int n, BigDecimal p, int nPrev) {
	//Use this method when we have already computed the result p
	//for nPrev and we want to extend to n.
	int i = nPrev;
	a = p;
	while (i<n-1) 
	    a = a.multiply(new BigDecimal(i++)); 
	return a;
    }

    public static BigDecimal pochhammer(double a, int n) {
	if (n==0)
	    return new BigDecimal(1);
	int i = 0;
	BigDecimal b = new BigDecimal(a);
	while (++i<n) 
	    b = b.multiply(new BigDecimal(a + i));
	return b;
    }
}
