public class RNARead implements Comparable{

    protected int pos, length, chr, nReads;
    private boolean strand;

    public RNARead(int c, boolean s, int p, int l){
	chr = c;
	strand = s;
	pos = p;
	length = l;
	nReads = 1;
    }

    public RNARead(int c, boolean s, int p, int l, int n){
	chr = c;
	strand = s;
	pos = p;
	length = l;
	nReads = n;
    }

    public RNARead(RNARead r){
	chr = r.chr;
	strand = r.strand;
	pos = r.pos;
	length = r.length;
	nReads = r.nReads;
    }

    public int getPosition(){ return pos; }

    public void setPosition(int p) { pos = p; }

    public boolean getStrand(){ return strand; }
    
    public int getLength(){ return length; }

    public int getChromosome() { return chr; }

    public int numberOfReads() { return nReads; }

    public boolean overlaps(int start, int end) { return pos + length<=start && pos + length>=end; }

    public boolean overlaps(int s) { return pos + length<=s && pos + length>=s; }

    public boolean overlaps(RNARead r) { 
	if (r.chr==chr)
	    return (r.pos + r.length>=pos && r.pos<=pos) || (pos + length>=r.pos && pos<=r.pos);
	return false;
    }

    //////////////////////////////////////////////////////////////////

    public String toString() { return chr + "_" + (strand ? "" : "-") + pos; }

    public int hashCode() { return 5*chr + pos*13 + length*29; }

    public boolean equals(Object o) {
	if (!(o instanceof RNARead))
	    System.err.println("Error comparing reads!");
	RNARead a = (RNARead)o;
	return this.pos==a.pos && this.strand==a.strand;
    }

    public int compareTo(Object o) {
	//Sort according to position
	if (!(o instanceof RNARead))
	    System.err.println("Error comparing reads!");
	RNARead a = (RNARead)o;
	if (this.pos<a.pos)
	    return -1;
	else if (this.pos>a.pos)
	    return 1;
	else {
	    if (this.strand && !a.strand)
		return -1;
	    else if (!this.strand && a.strand)
		return 1;
	    else { //Let the ones with shortest length come first
		if (this.length<a.length)
		    return -1;
		else if (this.length>a.length)
		    return 1;
		else
		    return 0;
	    }
	}
    }
}
