//Use this method to calculate the length distribution for the short RNA fragments.
//
//Example:
//
//java -Xmx1024m RNAReadLengthDistribution -o -I 


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
//import umontreal.iro.lecuyer.stat.Tally;



public class RNAReadLengthDistribution{


    
    public static DecimalFormat format = new DecimalFormat("###.####");

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option rnaSeqFileNameFlag = argParser.addStringOption('R', "RNASeq");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('a', "saveFileName");
	    //use this flag when we only want to focus on features
	    //which are guaranteed to be isolated from one another.
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('I', "isolated");
	    //Use this flag if we only want to consider the average read length for each loci
	    CmdLineParser.Option averageFlag = argParser.addBooleanOption('A', "average");
	    //Use this flag to include a file with genes/peaks to filter against
	    CmdLineParser.Option selectFlag = argParser.addStringOption('s', "select");
	    //Use this flag when we want to center around the TES
	    //instead of the TSS
	    CmdLineParser.Option endFlag = argParser.addBooleanOption('E', "end");
	    //Use this flag when we have other species
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    //Use this flag if we want to center around enhancers instead of TSSs
	    CmdLineParser.Option enhancerFlag = argParser.addStringOption("enhancer");
	    //Different time points
	    CmdLineParser.Option timeFlag = argParser.addIntegerOption('t', "time");
	    //Use the biological replicate
	    CmdLineParser.Option replicateFlag = argParser.addIntegerOption('b', "replicate");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(enhancerFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(timeFlag);
	    int time = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(replicateFlag);
	    int replicate = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    boolean averageLen = argParser.getOptionValue(averageFlag)!=null;
	    boolean ends = argParser.getOptionValue(endFlag)!=null;
	    tmp = argParser.getOptionValue(rnaSeqFileNameFlag);
	    String rnaSeqFileName = tmp!=null ? (String)tmp : "/groups/neuroduo/jesse/SOLiD/Solid_20081111_2_KCl_SmRNA/done/hr0_all/genome/unique_only/hr0_all_unique.tsv";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "/groups/neuroduo/martin/ChIP-Seq/TKData/mRNA/hr0s_sizes.txt";
	    tmp = argParser.getOptionValue(selectFlag);
    	    String selectFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(speciesFlag);
    	    String species = tmp!=null ? (String)tmp : "Mm";
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    //////////////////////////////////////////////////////////////////
	    int nChr = 21;
	    if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
		nChr = 24;
	    //Load the annotation info  and the mRNA
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    Vector<ABIRead[]> forwardReads = new Vector<ABIRead[]>(nChr), reverseReads = new Vector<ABIRead[]>(nChr);
	    int nReadsForward = 0, nReadsReverse = 0;
	    forwardReads.setSize(nChr);
	    reverseReads.setSize(nChr);
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra, species);
		    //Load all the RNA reads
		    forwardReads.setElementAt(loadMRNAReads(rnaSeqFileName, true, i, averageLen), i-1);
		    nReadsForward += forwardReads.elementAt(i-1).length;
		    reverseReads.setElementAt(loadMRNAReads(rnaSeqFileName, false, i, averageLen), i-1);
		    nReadsReverse += reverseReads.elementAt(i-1).length;
		    System.out.println("Loaded reads and annotation for chr " + i);
		}
	    }
	    int isolatedIDs[] = null;
	    if (isolated) {
		if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
		    isolatedIDs = FindReadsAcrossFeature.parseIsolatedFile((orchestra ? "/groups/neuroduo/martin/" : "/media/exthd/") + "Genomes/Hs/v36/isolated_genes.txt");
		else
		    isolatedIDs = FindReadsAcrossFeature.parseIsolatedFile(orchestra);
	    }
	    TreeMap<String, int[]> subsetSizes = new TreeMap<String, int[]>();
	    Vector<String> subsets = EstimatePALRLength.parseSelectFiles(selectFileName);
	    if (subsets.size()>0) {
		for (String s : subsets) 
		    subsetSizes.put(s, getSizes(forwardReads, reverseReads, chr, windowSize, s, isolatedIDs, orchestra, geneInfo, ends, species));
	    }
	    //Look at everything
	    subsetSizes.put("all", getSizes(forwardReads, reverseReads, chr, windowSize, null, isolatedIDs, orchestra, geneInfo, ends, species));
	    //Do the global distribution as well
	    int sizes[] = new int[35];
	    for (int i=0; i<nChr; i++) {
		if (chr==-1 || i+1==chr) {
		    for (int j=0; j<forwardReads.elementAt(i).length; j++)
			sizes[forwardReads.elementAt(i)[j].getLength()]++;
		    for (int j=0; j<reverseReads.elementAt(i).length; j++)
			sizes[reverseReads.elementAt(i)[j].getLength()]++;
		}
	    }
	    subsetSizes.put("global", sizes);
	    //Save all of them in files
	    String saveName = "";
	    saveName = saveFileName + (chr!=-1 ? "_" + chr : "") + ".txt";
	    System.out.println("Saving to " + saveName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	    if (subsetSizes.size()>1) {
		bw.write("#size");
		for (String s : subsetSizes.keySet()) {
		    if (s.lastIndexOf(".")>=0)
			bw.write("\t" + s.substring(s.lastIndexOf("/") + 1, s.lastIndexOf(".")));
		}
		bw.write("\n");
	    }
	    for (int i=0; i<35; i++) {
		bw.write(i + "\t");
		for (String s : subsetSizes.keySet()) 
		    bw.write(subsetSizes.get(s)[i] + "\t");
		bw.write("\n");
	    }
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static int[] getSizes(Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, int chr, int windowSize, String selectFileName, int[] isolatedIDs, boolean orchestra, RefSeqGenes[] geneInfo, boolean ends, String species) {
	int nChr = 21;
	if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
	    nChr = 24;
	int geneIDsKeep[] = FindReadsAcrossFeature.parseSelectFile(selectFileName);
	/////////////////////////////////
	int sizes[] = new int[35];
	int nGenes[] = new int[nChr], nGenesSum = 0;
	Arrays.fill(sizes, 0);
	for (int i=1; i<=nChr; i++) {
	    if (chr==-1 || i==chr) {
		int s[] = getSizes(forwardReads.elementAt(i-1), reverseReads.elementAt(i-1), geneInfo[i-1], isolatedIDs, geneIDsKeep, windowSize, ends);
		for (int j=0; j<sizes.length; j++) 
		    sizes[j] += s[j];
	    }
	}
	return sizes;
    }

    public static int[] getSizes(ABIRead[] forwardReads, ABIRead[] reverseReads, RefSeqGenes geneInfo, int[] isolatedIDs, int[] geneIDsKeep, int windowSize, boolean ends) {
	int sizes[] = new int[35];
	Arrays.fill(sizes, 0);
	for (RefSeqGene g : geneInfo.getGenes()) {
	    if (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) {
		if (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0) {
		    int start, end, tss = ends ? g.getEnd() : g.getStart();
		    ABIRead[] reads;
		    start = tss - windowSize;
		    end = tss + windowSize;
		    ABIRead[] rsf = FindReadsAcrossFeature.getReads(geneInfo.getChromosome(), start, end, forwardReads); //>0 ? 1 : 0;
		    ABIRead[] rsr = FindReadsAcrossFeature.getReads(geneInfo.getChromosome(), start, end, reverseReads); //>0 ? 1 : 0;
		    for (int i=0; i<rsf.length; i++) 
			sizes[rsf[i].getLength()]++;
		    for (int i=0; i<rsr.length; i++) 
			sizes[rsr[i].getLength()]++;
		}
	    }
	}
	return sizes;
    }

    ///////////////////////////////////////////////////////

    public static ABIRead[] loadMRNAReads(String mRNAFileName, boolean forward, int c, boolean average) {
	ABIRead[] reads = null;
	try {
	    int nReads = 0;
	    BufferedReader br = new BufferedReader(new FileReader(mRNAFileName));
	    String line = br.readLine();
	    Pattern mRNAPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([F|R])\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    if (chr==c) { //22 is the splice chromosome
			if ((m.group(3).equals("F") && forward) || (m.group(3).equals("R") && !forward))
			    nReads += average ? 1 : Integer.parseInt(m.group(4));
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    reads = new ABIRead[nReads];
	    br = new BufferedReader(new FileReader(mRNAFileName));
	    line = br.readLine();
	    int i = 0;
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    if (chr==c) { //22 is the splice chromosome
			//Parse the remaining numbers which are the lengths
			String lens[] = (line.substring(m.end())).split(",");
			if (average) { //Calculate the average length
			    int mean = 0;
			    for (int j=0; j<lens.length; j++)
				mean += Integer.parseInt(lens[j].trim());
			    if ((m.group(3).equals("F") && forward)) 
				reads[i++] = new ABIRead(chr, true, Integer.parseInt(m.group(2)), mean/lens.length);
			    else if (m.group(3).equals("R") && !forward)
				reads[i++] = new ABIRead(chr, false, Integer.parseInt(m.group(2)) - mean/lens.length, mean/lens.length);
			}
			else {
			    if ((m.group(3).equals("F") && forward)) {
				for (int j=0; j<Integer.parseInt(m.group(4)); j++)
				    reads[i++] = new ABIRead(chr, true, Integer.parseInt(m.group(2)), Integer.parseInt(lens[j].trim()));
			    }
			    else if (m.group(3).equals("R") && !forward) {
				for (int j=0; j<Integer.parseInt(m.group(4)); j++)
				    reads[i++] = new ABIRead(chr, false, Integer.parseInt(m.group(2)) - Integer.parseInt(lens[j].trim()), Integer.parseInt(lens[j].trim()));
			    }
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading file " + mRNAFileName + " : " + e.toString());
	}
	return reads;
    }



}
