import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class RepeatRegion implements Comparable {
    //Start is always smaller than end

    protected int chr, start, end;
    protected boolean forward;
    protected String familyName, name;

    public RepeatRegion(int c, boolean f, int s, int e, String n, String fn) {
	chr = c;
	forward = f;
	start = s;
	end = e;
	name = n;
	familyName = fn;
    }

    public int getStart() { return start; }

    public int getEnd() { return end; }

    public String getName() { return name; }

    public int getChromosome() { return chr; }
    
    public String getFamilyName() { return familyName; }

    public boolean overlaps(int pos) { return start<=pos && pos<=end; }

    public boolean overlaps(int s, int e) {
	return (start<=s && end>=s) || (start<=e && end>=e) || (start>s && end<e);
    }

    public boolean overlaps(ABIRead r) { return overlaps(r.getPosition(), r.getPosition() + r.getLength()); }

    ////////////////////////////////////////////

    public boolean equals(Object o) {
	if (!(o instanceof RepeatRegion))
	    return false;
	RepeatRegion e = (RepeatRegion)o;
	return e.chr==chr && e.start==start && e.end==end;
    }

    public int compareTo(Object o) {
	if (!(o instanceof RepeatRegion))
	    return 0;
	RepeatRegion e = (RepeatRegion)o;
	if (chr<e.chr)
	    return -1;
	else if (chr>e.chr)
	    return 1;
	else {
	    if (start<e.start)
		return -1;
	    else if (start>e.start)
		return 1;
	    else {
		if (end<e.end)
		    return -1;
		else if (end>e.end)
		    return 1;
		else
		    return 0;
	    }
	}
    }

    public String toString() { return name + ", " + familyName + ", chr" + chr + ":" + start + "-" + end; }

}