//This class is used to represent a peak when parsed from the
//TF_genes_E120.txt list. These lists are generated by
//AnalyzeControls.java and then merged using the merge_output2.sh
//script.

import java.util.regex.*;
import java.util.*;

public class TFPeak implements Comparable {

    //chr start end peakLoc height nReads nInputReads GeneName RefSeq GeneID GeneDist GeneStrand DistCategory
    //We only care about the fields up until GeneName
    public static Pattern peakPattern = Pattern.compile("sjergnsuigchr([\\d|X|Y]+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
    //chr start end nReads
    public static Pattern shortPeakPattern = Pattern.compile("chr([\\d|X|Y]+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");

    protected int chr, start, end, nReads, height, peakLoc, nInputReads;
    protected String tf;

    public TFPeak(String t, String s){
	tf = t;
	Matcher m = peakPattern.matcher(s);
	if (m.find()) {
	    if (m.group(1).equals("X"))
		chr = 21;
	    else if (m.group(1).equals("Y"))
		chr = 22;
	    else
		chr = Integer.parseInt(m.group(1));
	    start = Integer.parseInt(m.group(2));
	    end = Integer.parseInt(m.group(3));
	    peakLoc = Integer.parseInt(m.group(4));
	    height = Integer.parseInt(m.group(5));
	    nReads = Integer.parseInt(m.group(6));
	    nInputReads = Integer.parseInt(m.group(7));
	}
	else {
	    m = shortPeakPattern.matcher(s);
	    if (m.find()) {
		if (m.group(1).equals("X"))
		    chr = 21;
		else if (m.group(1).equals("Y"))
		    chr = 22;
		else
		    chr = Integer.parseInt(m.group(1));
		start = Integer.parseInt(m.group(2));
		end = Integer.parseInt(m.group(3));
		nReads = Integer.parseInt(m.group(4));
	    }
	    else
		System.err.println("Error, could not find any match for line: " + s);
	}
    }

    public int getChromosome() { return chr; }

    public String getTFName() { return tf; }

    public boolean overlaps(TreeSet ps) { return overlaps(ps, 0); }

    public boolean overlaps(TreeSet ps, int d) {
	Iterator it = ps.iterator();
	while (it.hasNext()) {
	    if (overlaps((TFPeak)it.next()))
		return true;
	}
	return false;
    }

    public boolean overlaps(TFPeak p) { return overlaps(p, 0); }

    public boolean overlaps(TFPeak p, int d) {
	//Check if two peaks overlap
	if (start<p.start && end>p.end) 
	    return true;
	else if (p.start<start && p.end>end) 
	    return true;
	else if (p.start<start) 
	    return p.end + d > start;
	else
	    return end + d > p.start;
    }

    public Vector findNearbyPeaks(Vector peaks) { return findNearbyPeaks(0, peaks); }
    
    public Vector findNearbyPeaks(int dist, Vector peaks) {
	Vector nearbyPeaks = new Vector();
	Iterator it = peaks.iterator();
	while (it.hasNext()) {
	    TFPeak p = (TFPeak)it.next();
	    if (overlaps(p, dist))
		nearbyPeaks.addElement(p);
	}
	return nearbyPeaks;
    }


    //////////////////////////////////////////////////////////////////

    public int compareTo(Object o) {
	TFPeak p = (TFPeak)o;
	if (chr<p.chr)
	    return -1;
	else if (chr>p.chr)
	    return 1;
	else {
	    if (start<p.start)
		return -1;
	    else if (start>p.start)
		return 1;
	    else
		return 0;
	}
    }

    public boolean equals(Object o) {
	TFPeak p = (TFPeak)o;
	return p.chr==chr && p.start==start && p.end==end && p.nReads==nReads;
    }

    public String toString() { return start + "\t" + end + "\t" + nReads; }

    public int hashCode() {
	return chr*11 + start + 29*nReads;
    }


}
